:- multifile macro/2.

%% some exercises from Gusfield, Algorithms on Strings, Trees and 
%% Sequences, Computer Science and Computational Biology. Cambridge
%% University Press 1997.
%%
%% page 14, exercise 10.

%% input:  DNA, a sequence of nucleotides (one of {a,t,c,g}).
%% output: protein, a sequence of amino acides (abbreviated with capitals
%%         below)
%%
%% The mapping relates a triple of nucleotides (a `codon') to an amino
%% acid. The actual mapping is given by the `genetic code'.

%% nuc = nucleotide
macro(nuc, {a,t,c,g} ).

%% http://newboree.cines.fr:8104/textes/tables/gencod.html
macro(genetic_code,
      {[t,t,t]:'F', [t,c,t]:'S', [t,a,t]:'Y', [t,g,t]:'C',
       [t,t,c]:'F', [t,c,c]:'S', [t,a,c]:'Y', [t,g,c]:'C',
       [t,t,a]:'L', [t,c,a]:'S', 
       [t,t,g]:'L', [t,c,g]:'S',              [t,g,g]:'W',

       [c,t,t]:'L', [c,c,t]:'P', [c,a,t]:'H', [c,g,t]:'R',
       [c,t,c]:'L', [c,c,c]:'P', [c,a,c]:'H', [c,g,c]:'R',
       [c,t,a]:'L', [c,c,a]:'P', [c,a,a]:'Q', [c,g,a]:'R',
       [c,t,g]:'L', [c,c,g]:'P', [c,a,g]:'Q', [c,g,g]:'R',

       [a,t,t]:'I', [a,c,t]:'T', [a,a,t]:'N', [a,g,t]:'S',
       [a,t,c]:'I', [a,c,c]:'T', [a,a,c]:'N', [a,g,c]:'S',
       [a,t,a]:'I', [a,c,a]:'T', [a,a,a]:'K', [a,g,a]:'R',
       [a,t,g]:'M', [a,c,g]:'T', [a,a,g]:'K', [a,g,g]:'R',

       [g,t,t]:'V', [g,c,t]:'A', [g,a,t]:'D', [g,g,t]:'G',
       [g,t,c]:'V', [g,c,c]:'A', [g,a,c]:'D', [g,g,c]:'G',
       [g,t,a]:'V', [g,c,a]:'A', [g,a,a]:'E', [g,g,a]:'G',
       [g,t,g]:'V', [g,c,g]:'A', [g,a,g]:'E', [g,g,g]:'G'
      }).

macro(frame1,[genetic_code*,nuc:[]^,nuc:[]^]).
macro(frame2,[nuc:[],genetic_code*,nuc:[]^,nuc:[]^]).
macro(frame3,[nuc:[],nuc:[],genetic_code*,nuc:[]^,nuc:[]^]).

macro(translate,{frame1,frame2,frame3}).

macro(test(Input),m(range(Input o translate))).

%% this will give you an automaton of all possible outcomes
macro(ex1, m(range([a,t,g,g,a,c,g,a,a] o translate))).

%% suppose you have a very long sequence of such triples:
macro(ex2, test(length(1000,[a,t,g,g,a,c,g,a,a]))).

%% The resulting transducer *could* be implemented
%% linear in length of input, but currently FSA does not provide such
%% an efficient implementation for non-deterministic and/or non-functional
%% transducers. However, in practice it often is linear:
%%
%% `easy' pattern because 'tga' does not exist, but `gat' does
%% `difficult' pattern `aaa'
%% fsa pred_module=fsa_frozen -aux exs -r 'test(length(N,[a,a,a]))'
%%
%%         
%%    N     msec
%%  100     1800
%%  200     2750
%%  300     3680
%%  400     4680
%%  500     5580
%% 1000    10360
%% 2000    20300
%% 4000    40930


%% some variations

%% this yields all DNA strings of length zeven which map to
%% something containing 'K'
macro(contains_k, m(domain( [?,?,?,?,?,?,?] o translate o $ 'K' ))).

%% is it possible to have a sequence of a,g,t map only to 'A', 'G', 'T'
macro(agt,{a,g,t}* o translate o {'A','G','T'}).

macro(creep,m(domain(translate o ['C','R','E','E','P']))).

%% exercise 13
%%
%% given: proteins S1 and S2, |S2| > |S1|
%% asked: let D1 be a DNA encoding of S1, and D2 of S2
%%        Is there D1 and D2 such that D1 is a substring of D2?
%% 
%% The solution is a set of strings D2, in which D1 is marked by #
%%
%% e.g. chal(['A','E','E','A'],['A','G','G','G','G','A','G'])
%% I'd like to have some (pseudo-)realistic inputs...
macro(chal(S1,S2), m(range(
              domain(translate o S2)
                       o
   $ [ [] : #,domain(translate o S1), [] : #]
			  ))).

%% utility; perhaps this should be builtin (properly generalized)
macro(length(N,Pat),List) :-
    length(List,N),
    fill_pat(List,Pat).

fill_pat([],_).
fill_pat([Pat| T],Pat) :-
    fill_pat(T,Pat).


