%% Lauri Karttunen, The Replace Operator, in: Roche and Schabes (eds),
%% Finite State Language Processing, MIT Press 1997.
%%
%% by Dale Gerdemann and Gertjan van Noord
:- multifile macro/2.

:- op(402,xfy,'uu').
:- op(402,xfy,'ul').
:- op(402,xfy,'lu').
:- op(402,xfy,'ll').
:- op(402,xfy,'=>').
:- op(402,xfy,'<=').
:- op(402,xfy,'--').

%%% This replacement algorithm works by generating and testing. The
%%% first step is to insert brackets {<,>} randomly. Then test that
%%% that the right contexts are marked by a '>', perform the
%%% replacement and then check that the left contexts are followed by
%%% a '<'. 

macro(replace(A,B,C,D),A => B lu C -- D).

%% "rightward" //
macro(Upper => Lower lu Left -- Right,
            insert_brackets
                   o
            constrain_brackets
                   o
            right_context(Right)
                   o
            rep(Upper,Lower)
                   o
            left_context(Left)
                   o
            remove_brackets
     ).

%% "downward" \/
macro(Upper => Lower ll Left -- Right,
            insert_brackets
                   o
            constrain_brackets
                   o
            rep(Upper,Lower)
                   o
            right_context(Right)
                   o
            left_context(Left)
                   o
            remove_brackets
     ).


%% "left-oriented" \\
macro(Upper => Lower ul Left -- Right,
            insert_brackets
                   o
            constrain_brackets
                   o
            left_context(Left)
                   o
            rep(Upper,Lower)
                   o
            right_context(Right)
                   o
            remove_brackets
     ).

%% "upward" ||
macro(Upper => Lower uu Left -- Right,
            insert_brackets
                   o
            constrain_brackets
                   o
            left_context(Left)
                   o
            right_context(Right)
                   o
            rep(Upper,Lower)
                   o
            remove_brackets
     ).

%%% This defines simple (uncontexted) replacement. Note: '$' means
%%% "contains", '~' means "complement" 
%%% Coercion to an identity transducer is implicit.

macro(Upper => Lower,
      [[ ~ $ (Upper - []) , Upper x Lower ] *,~ $ (Upper - [])]). 

macro(rep(Upper,Lower),[<,ignore(Upper,{<,>}),>]
                                 =>
                       [<,ignore(Lower,{<,>}),>] ).

macro(remove_brackets,{<,>} => []).

macro(insert_brackets,inverse(remove_brackets)).

macro(constrain_brackets, ~ $ [<,>]).

/* I'd prefer this style, but it results in even more problems,
%% The bug(s) shows up for instance for the example
%% replace(a,b,x,y)  applied to strings solely consisting of the context,
%% xy
%% (no)
%%
%% or for example replace(a,b,x,x)
%%
%% where  xax -> xbx (as expected)
%%   but  xaxx ->    (fails)

macro(if_p_then_s(L1,L2), ~ [L1,~L2]).

macro(if_s_then_p(L1,L2), ~ [~L1,L2]).

macro(p_iff_s(L1,L2),minimize(
      if_p_then_s(L1,L2)
	    &
      if_s_then_p(L1,L2))
     ).

macro(left_context(Left),
      p_iff_s(ignore([? *,Left],{<,>}) - [? *,{<,>}], [<,? *])
     ).
                                                                   
macro(right_context(Right),                                  
      p_iff_s([? *,>], ignore([Right,? *],{<,>}) - [{<,>},? *])
     ).
*/

macro(dots_left(Left),   [? *,ignore(Left,{<,>})]  - [? *,<] ).

macro(right_dots(Right), [ignore(Right,{<,>}),? *] - [>, ? *]).

macro(left_context(Left),
       ~[~dots_left(Left),  ignore(<,>),? *] &
       ~[ dots_left(Left),~[ignore(<,>),? *]]
	).

macro(right_context(Right),
       ~[  ? *,ignore(>,<), ~right_dots(Right)]   &
       ~[~[? *,ignore(>,<)], right_dots(Right)]
	).

%% examples

macro(ex(4.35),  [a,b] => x uu [c,d^] -- {e,f}    ).

macro(ex(4.42),  [a,b] => x uu  [a,b] -- a         ).


%% BUGS as well ???

macro(bug, a => [] lu b -- [] ).

%% Dale Gerdemann writes:
%%  a => [] lu b -- []
%%
%% The string baaa should be transduced to just b.
%%
