%%% Gertjan van Noord, (c) 1998
%%%
%%% with help from Dale Gerdemann
%%%
%%% Mehryar Mohri and Richard Sproat, An Efficient Compiler for Weighted
%%% Rewrite Rules. 34th Annual Meeting of the ACL, Santa Cruz 1996, pages
%%% 230-238. This only treats the non-weighted case.

:- multifile macro/2.
:- multifile rx/2.

:- discontiguous macro/2.
:- discontiguous ms__mark/4.

macro(replace(L,Phi,Psi,R), ms__replace( L,Phi,Psi, R)).
macro(replace(  Phi,Psi  ), ms__replace([],Phi,Psi,[])).

%%% replace(Left,Phi,Psi,Right) === Phi -> Psi /Left _ Right
%%% i.e. the left-to-right obligatory replacement rule
macro(ms__replace(Left0,Phi0,Psi,Right),  pragma([Phi-Phi0,Left-Left0],
      ms__r(Right)
         o
      ms__f(Phi)
         o
      ms__replace_internal(Phi,Psi)
         o
      ms__l1(Left)
         o
      ms__l2(Left)                     
     )).

macro(ms__r(R),
      reverse(ms__marker(1,
		     [ms__sigma*,reverse(R)],
		     [>]
		    )
	     )
     ).

macro(ms__f(F),
      reverse(ms__marker(1,
			 [{ms__sigma,>}*,reverse([ignore(F,{>}),>])],
			 ['<1','<2']
			)
	     )
     ).

macro(ms__l1(L),
      ms__sloppy_ignore(ms__marker(2,
				   [ms__sigma*,L],
				   '<1'
				   ),
			{'<2':'<2'}
		       )
     ).

macro(ms__l2(L),ms__marker(3,
		   [ms__sigma*,L],
		   '<2'
		  )
     ). 

macro(ms__replace_internal(Phi,Psi), 
      {{ms__sigma,'<2':'<2', > :[]}, 
       ['<1':'<1',
	ignore(Phi, {'<1','<2',> }) x Psi,
	> :[]
       ]
      }*).


macro(ms__sigma,`{'<1','<2',>}).

%%%%%%%%%%%%%%%%
%%% MARKER/3 %%%
%%%%%%%%%%%%%%%%

%% we only implement the markers that we really use, so not in full
%% generality (since I am lazy).
%% marker(1,RegEx,List)
%%                List is a list of symbol(pairs) to be inserted
%% marker(2,RegEx,Sym)
%%                Sym is a symbol(pair) to be deleted
%% marker(2,RegEx,Sym)
%%                Sym is a symbol(pair) to be deleted
rx(ms__marker(Type,Expr,Change),Fa) :-
    fsa_regex:rx(identity(determinize(Expr)),Fa0),
    ms__mark(Type,Change,Fa0,Fa).

%%%%%%%%%%%%%%
%%% type 1 %%%
%%%%%%%%%%%%%%

%% NB: for type 1 we only need insertions ...
ms__mark(1,Insertions0,Fa1,Fa) :-
    fsa_data:fsa_symbols_decl(Fa1,t(Module,ModuleT)),
    ms__regex_notation_to_predicate_list(Insertions0,Insertions,Module),
    fsa_data:fsa_start_states(Fa1,Starts),
    fsa_data:fsa_transitions(Fa1,Trans0),
    fsa_data:fsa_final_states(Fa1,Finals),
    fsa_data:fsa_states_set(Fa1,AllStates),
    ordsets:ord_subtract(AllStates,Finals,NewFinals0),
    ms__add_insertions(Finals,Insertions,NewFinals,NewFinals0,Trans,Trans1),
    ms__replace_transitions_starting_final(Trans0,Trans1,Fa1),
    fsa_data:fsa_construct_rename_states(t(Module,ModuleT),Starts,NewFinals,Trans,[],Fa).

%% this is slow for large automata
%% in such cases it's worthwhile creating a thing for final_states
%% moreover it creates symbolic names first, which are renamed later
%% it would be better to create larger integers for state names here
%% immediately
ms__replace_transitions_starting_final([],[],_).
ms__replace_transitions_starting_final([trans(A0,B,C)|T0],[trans(A,B,C)|T],Fa):-
    (	fsa_data:fsa_final_state(Fa,A0)
    ->	A=q(A0)
    ;	A=A0
    ),
    ms__replace_transitions_starting_final(T0,T,Fa).

ms__add_insertions([],_,F,F) --> [].
ms__add_insertions([F0|Fs],Ins,[q(F0)|NewF0],NewF) -->
    ms__add_ins(Ins,F0),
    ms__add_insertions(Fs,Ins,NewF0,NewF).

ms__add_ins([],_F) --> [].
ms__add_ins([Sym|Syms],F) -->
    [trans(F,[]/Sym,q(F))],
    ms__add_ins(Syms,F).

%%%%%%%%%%%%%%
%%% type 2 %%%
%%%%%%%%%%%%%%

%% NB: for type 2 and 3 we only need a single deletion
ms__mark(2,Deletion0,Fa1,Fa) :-
    fsa_data:fsa_symbols_decl(Fa1,t(Module,_)),
    Module:regex_notation_to_predicate(Deletion0,Deletion),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa2,Trans0,Trans),
    fsa_data:fsa_copy_except(final_states,Fa2,Fa,Finals,AllStates),
    fsa_data:fsa_states_set(Fa1,AllStates),
    ms__add_2_deletions(Finals,Deletion,Trans1,Trans0),
    sort(Trans1,Trans).

ms__add_2_deletions([],_) --> [].
ms__add_2_deletions([Final|Fs],Del) -->
    [trans(Final,Del/[],Final)],
    ms__add_2_deletions(Fs,Del).

%%%%%%%%%%%%%%
%%% type 3 %%%
%%%%%%%%%%%%%%

%% NB: for type 2 and 3 we only need a single deletion
ms__mark(3,Deletion0,Fa1,Fa) :-
    fsa_data:fsa_symbols_decl(Fa1,t(Module,_)),
    Module:regex_notation_to_predicate(Deletion0,Deletion),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa2,Trans0,Trans),
    fsa_data:fsa_copy_except(final_states,Fa2,Fa,Finals,AllStates),
    fsa_data:fsa_states_set(Fa1,AllStates),
    ordsets:ord_subtract(AllStates,Finals,NonFinals),
    ms__add_2_deletions(NonFinals,Deletion,Trans1,Trans0),
    sort(Trans1,Trans).

macro(ms__sloppy_ignore(A,B),ignore0(A,B)).
%% should replace this by Dale's suggestion.

ms__regex_notation_to_predicate_list([],[],_).
ms__regex_notation_to_predicate_list([H0|T0],[H|T],Mod):-
    Mod:regex_notation_to_predicate(H0,H),
    ms__regex_notation_to_predicate_list(T0,T,Mod).
