%% by Gertjan van Noord
%% cf Kager, chapter 3.

%% TODO
%% Dale writes:
%% > One obvious bug in your code is the use of ignore. You really want a
%% > version of ignore that treats sequences of two as a unit (as we used
%% > in the EACL paper).

:- multifile macro/2.

:- op(403, yfx, 'oo').
:- op(400, yfx, '::').

%% positional encoding: every odd position is character
%%                      every even position is special marker:
%%                           marker  0: input *and* output
%%                                   1: input only
%%                                   2: output only
%%                                   3: markup (special output only)

macro(gen, { [letter,     []:0],
             [letter,     []:1],   % deletion
	     [[]:letter,  []:2],   % insertion
	     [[]:bracket, []:3]
	   }*
             o
            ignore(syllable*,[letter,1])
     ).

macro(letter,    a..z).
macro(cons,      {b,c,d,f,g,h,j,k,l,m,n,p,q,r,s,t,v,w,x,y,z}).
macro(vowel,     {a,e,o,u,i}).
macro(o_br,      'O[').
macro(n_br,      'N[').
macro(d_br,      'D[').
macro(r_br,      ']').
macro(bracket,   {o_br,n_br,d_br,r_br}).

%% refer to output only
macro(onset,     [o_br, 3, [cons,{0,2}]+  , r_br, 3]).
macro(nucleus,   [n_br, 3, [vowel,{0,2}]  , r_br, 3]).
macro(coda,      [d_br, 3, [cons,{0,2}]+  , r_br, 3]).

macro(syllable,  [onset^,nucleus,coda^]).

macro(ignore_markers(Expr), Expr o {[?, {0,2,3} : []],
				    [? : [],1 : []]
				   }*).

%% Syllables must not have codas
macro(mark_ot_constraint(no_coda,Mark),
      dots([d_br,3],        [],Mark) ).

macro(mark_ot_constraint(complex_ons,Mark),
      dots(ignore([o_br,3, letter, {0,2}, [letter,{0,2}]+, r_br,3],[letter,1]),
	   [],Mark)).

macro(mark_ot_constraint(complex_coda,Mark),
      dots(ignore([d_br,3, letter, {0,2}, [letter,{0,2}]+, r_br,3],[letter,1]),
	   [],Mark)).

macro(mark_ot_constraint(max_io,Mark),
      dots([letter,1],[],Mark)).

macro(mark_ot_constraint(dep_io,Mark),
      dots([letter,2],[],Mark)).

:- ensure_loaded('../GerdemannVannoord99/replace').
macro(mark_ot_constraint(have_ons,Mark),
      replace([] x Mark,[],n_br)
                 o
      replace(Mark x [],ignore(onset,[letter,1]),[])
     ).

%% this is Xerox' X -> L ... R
%% if X gets complicated you might prefer to use
%% replace([[] x L, X, [] x R]) instead (slower)
macro(dots(X,L,R), [[~ $ X, [[] x L, X, [] x R]]*, ~ $ X]).

macro(add_violation(N),
      % remove stuff
      {[letter,       0 x []],
       [letter,       1 x []],
       [letter x [],  2 x []],
       [bracket x [], 3 x []],
       @
      }*
        o
  [[? *,[]: @ ]+, ? *]
        o
  permute_marker(N)
        o   inverse(
      {[letter,       0 x []],
       [letter,       1 x []],
       [letter x [],  2 x []],
       [bracket x [], 3 x []],
       @
       }*          )
     ).
      
macro(permute_marker(0), ? *).
macro(permute_marker(N), cache(permute_marker) o permute_marker(N1)) :-
    N>0, N1 is N-1.

macro(permute_marker,
      [{[? *,  @ : [], ? *, [] : @ ],
	[? *, [] : @ , ? *,  @ : []]}*, ? *]).

macro(Input oo Precision :: Constraint,oo(Input,Precision,Constraint)).
macro(Input oo              Constraint,oo(Input,0,        Constraint)).

macro(oo(Input,Precision,Constraint),
                          pragma([V1 - (Input o mark_ot_constraint(Constraint,@))],
                  V1 
                    o
             ~ range(V1
		       o
		  add_violation(Precision)
		    )
                    o
             { @ : [],? -  @ }*		 
		    )).


macro(syl(Lang,B), ignore_markers(M)) :-
    conslist(Lang,List),
    create_syl_expr_i(List,B,gen,M).

macro(syls(Lang,[B|Bs]), ignore_markers(M)):-
    conslist(Lang,List),
    create_syl_expr(List,[B|Bs],gen,M).

create_syl_expr([],[],M,M).
create_syl_expr([C|Cs],[N|Ns],M0,M):-
    create_syl_expr(Cs,Ns,M0 oo N :: C, M).

create_syl_expr_i([],_,M,M).
create_syl_expr_i([C|Cs],N,M0,M):-
    create_syl_expr_i(Cs,N,M0 oo N :: C, M).



%%% syls(Lang,ListOfPrecisions) 
%%% syl(Lang,Integer) each constraint has precision Integer

conslist(lenakel,[have_ons,complex_ons,complex_coda,max_io,dep_io,no_coda]).