%% Prince and Smolensky 93, chapter 6.
%% Also see Karttunen '98.
:- multifile macro/2.
:- discontiguous macro/2.

:- ensure_loaded(utils).

macro(onset,     [o_br, cons^  , r_br]).
macro(nucleus,   [n_br, vowel^ , r_br]).
macro(coda,      [d_br, cons^  , r_br]).
macro(unparsed,  [x_br, letter , r_br]).


macro(parse, dots(cons,{o_br, d_br, x_br},r_br)
		 o
             dots(vowel,{n_br, x_br},r_br)).

%%macro(overparse,
%%             introduce_every_pos([{o_br,d_br,n_br},r_br]*)).

macro(overparse,
             introduce_every_pos([{o_br,d_br,n_br},r_br]^)).

macro(syllable_structure,
             ignore([onset^,nucleus,coda^],unparsed)*).

macro(gen, input* o overparse o parse o syllable_structure ).

%% Input segments must be parsed
macro(mark_ot_constraint(parsed,Mark),
      dots(x_br,        [],Mark) ).

%% Syllables must not have codas
macro(mark_ot_constraint(no_coda,Mark),
      dots(d_br,        [],Mark) ).

%% A nucleus position must be filled
macro(mark_ot_constraint(fill_nuc,Mark),
      dots([n_br, r_br],[],Mark) ).

%% An onset position must be filled
macro(mark_ot_constraint(fill_ons,Mark),
      dots([o_br, r_br],[],Mark) ).

%% Syllables must have onsets; i.e. every nucleus
%% must immediately be preceded by an onset
%% see below for a variant that is easier to understand; this one is faster
%% and doesn't require the replacement stuff to be loaded.
macro(mark_ot_constraint(have_ons,Mark),
      {?, [[] x Mark, n_br]}*  % optional marker before a nucleus
              o
      ~ $ [onset,Mark]         % but not if there is a preceding onset
              o                % and obligatory if there is no onset
           ~ [ ( ~[? *,onset] & ~[? *, Mark] ),  n_br,? *]
     ).

%% Gertjan suggests the following alternative:
%%:- ensure_loaded('../GerdemannVannoord99/replace').
%%macro(mark_ot_constraint(have_ons,Mark),
%%      replace([] x Mark,[],n_br)
%%                 o
%%      replace(Mark x [],onset,[])
%%     ).
%%

%% Dale suggests the following alternative:
%%:- ensure_loaded('../GerdemannVannoord99/replace').
%%macro(add_anchors,[[] x '$',? *, [] x '$']).
%%
%%macro(mark_ot_constraint(have_ons,Mark),
%%      add_anchors
%%          o
%%      replace([] x Mark,(['$',? *] - [? *,onset]),n_br)
%%          o
%%     inverse(add_anchors)).

%% conslist(Lang,ListOfConstraints)
%% for each `language' identified by Prince and Smolensky (page 94),
%% give the corresponding list of constraints (most prominent first)
conslist(1,[ have_ons, fill_ons, no_coda , fill_nuc, parsed   ]).
conslist(2,[ have_ons, no_coda , fill_nuc, parsed  , fill_ons ]).
conslist(3,[ no_coda , fill_nuc, parsed  , fill_ons, have_ons ]).
conslist(4,[ have_ons, fill_ons, no_coda , parsed  , fill_nuc ]).
conslist(5,[ have_ons, no_coda , parsed  , fill_nuc, fill_ons ]).
conslist(6,[ no_coda , parsed  , fill_nuc, fill_ons, have_ons ]).
conslist(7,[ have_ons, fill_ons, parsed  , fill_nuc, no_coda  ]).
conslist(8,[ have_ons, parsed  , fill_ons, fill_nuc, no_coda  ]).
conslist(9,[ parsed  , fill_ons, have_ons, fill_nuc, no_coda  ]).

%% Karttunen uses (with counting):
macro(lenient,      syls(2,[0,0,0,2,0])).

%% sligthly better (for counting):
macro(syllabify(N), syls(2,[0,0,1,N,1])).





