%% the operator oo is the `optimality operator'
%% Input oo Constraint
%% Input oo Precision :: Constraint
:- op(403, yfx, 'oo').
:- op(400, yfx, '::').

:- multifile macro/2.
:- discontiguous macro/2.

macro(cons,      {b,c,d,f,g,h,j,k,l,m,n,p,q,r,s,t,v,w,x,y,z}).
macro(vowel,     {a,e,o,u,i}).
macro(letter,    a..z).
macro(o_br,      'O[').
macro(n_br,      'N[').
macro(d_br,      'D[').
macro(x_br,      'X[').
macro(r_br,      ']').
macro(bracket,   {o_br,n_br,d_br,x_br,r_br}).
macro(input,     letter).
macro(output,    bracket).

%% creates transducer which introduces P in every position
%% this is Xerox' [. .] -> P
macro(introduce_every_pos(P), [ [[] x P , ? ]*, [] x P ]).

%% this is Xerox' X -> L ... R
%% if X gets complicated you might prefer to use
%% replace([[] x L, X, [] x R]) instead (slower)
macro(dots(X,L,R), [[~ $ X, [[] x L, X, [] x R]]*, ~ $ X]).

macro(syl(Lang,B), M) :-
    conslist(Lang,List),
    create_syl_expr_i(List,B,gen,M).

macro(syls(Lang,[B|Bs]), M):-
    conslist(Lang,List),
    create_syl_expr(List,[B|Bs],gen,M).

create_syl_expr([],[],M,M).
create_syl_expr([C|Cs],[N|Ns],M0,M):-
    create_syl_expr(Cs,Ns,M0 oo N :: C, M).

create_syl_expr_i([],_,M,M).
create_syl_expr_i([C|Cs],N,M0,M):-
    create_syl_expr_i(Cs,N,M0 oo N :: C, M).

:- user:bb_put(fsa_rx_spy:syls,on).
:- user:bb_put(fsa_rx_spy:syl,on).

:- ensure_loaded('../RocheSchabes97/gvn').     % for is_functional

%% is_functional(E) returns E if E is functional or {} otherwise.
%% Since we want to verify a conjunction of is_functional's we compose
%% is_functional with (? * x ? *). This way we can safely compose
%% is_functional's.  Cf. ../Boolean/ for the origins of
%% this usage of '{}' for false and '? *' for true.
macro(is_f(Expr),is_functional(Expr) o (? * x ? *)).

%% is exact as far as constraint Cons is concerned
macro(exact(X,Cons),
        is_f(X o mark_ot_constraint(Cons, @) o {(? - @) x [], @}*)).

%% creates an automaton for the language {s} where s is either the symbol
%% 'all' to indicate that Expr is fixed point (hence exact) or either s
%% is an integer indicating that Expr is exact for all input strings of
%% length =< s. If L0 is given then we don't try lengths < L0.
macro(find_length(Expr),Macro) :-
    fsa_regex:rx(exact(Expr),Fa0),
    (	Fa0 = fa(_,1,[0],[],[],[])            % if Expr is not exact
    ->	Macro=find_length(0,Expr)
    ;   Macro=all
    ).
    
macro(find_length(L0,Expr),L1) :-
    length(List,L),		% generate `input' list
    L >= L0,
    all_qs(List),		% in increasing order
    fsa_regex:rx(exact(List o Expr),Fa), % check Input o Expr is exact?
    Fa = fa(_,1,[0],[],[],[]),	% for this length, Expr is
    !,				% *not* exact
    L1 is L-1.			% report `previous' length

%% all_qs(X) X is a proper list where each element is the question mark
all_qs([]).
all_qs([input | Qs]):-
    all_qs(Qs).

%% find parameters for language Lang and length Len, i.e. this
%% returns the expression syls(Lang,[A,B,C,D,E]) with appropriate
%% values for A, B, C, D, E such that the implementation is exact
%% for strings of length =< Len.
macro(syls_exact_upto(Lang,Len),fa(Fa)) :-
    conslist(Lang,Cons),
    length(Qs,Len),
    all_qs(Qs),
    find_pars(Cons,Qs o gen, [A,B,C,D,E]),
    fsa_regex:rx(syls(Lang,[A,B,C,D,E]), Fa),
    fsa_regex:rx(exact(fa(Fa)),Fa0),
    (	Fa0 = fa(_,1,[0],[],[],[])
    ->	true
    ;	format(user_error,"Exact!~n",[])
    ).

find_pars([],_,[]).
find_pars([H|T],Pre,[A|Ts]) :-
    find_par(H,Pre,A),
    find_pars(T,Pre oo A :: H,Ts).

find_par(Cons,Expr,Number) :-
    find_par(Cons,Expr,0,Number).

find_par(Cons,Expr,N0,N) :-
    fsa_regex:rx(exact(Expr oo N0::Cons, Cons),Fa0),
    (	Fa0 = fa(_,1,[0],[],[],[]) % if Expr is not exact
    ->	N1 is +(N0,1),             % then increase bound
	find_par(Cons,Expr,N1,N)
    ;   N0=N                       % else we found bound
    ).

%% X is exact if each of the constraints always assign the same number
%% of marks. This is checked by ignoring everything but the marks, and
%% then verifying that the resulting transducer is functional. Note that
%% it is *not* sufficient to check whether X is functional. In many cases
%% an exact OT analysis assigns multiple outputs to an input (contrary to
%% popular belief in OT).
macro(exact(X), Expr) :-
    conslist(_,[Cons|List]),!,
    create_exact_list(List,X,exact(X,Cons),Expr).

create_exact_list([],_,E,E).
create_exact_list([C|Cs],X,E0,E) :-
    create_exact_list(Cs,X,E0 o exact(X,C), E).


macro(find_length_each(N),List):-
    find_length_each(N,List).

find_length_each(N,List) :-
    findall(Name x find_length(syl(Name,N)),conslist(Name,_),List).

macro(ambiguous_output(Expr),E) :-
    fsa_regex:rx(is_functional(Expr),Fa0),
    (	Fa0 = fa(_,1,[0],[],[],[])            % if Expr is not functional
    ->	ambiguous_output(Expr,E)
    ;   E='{}'
    ).

ambiguous_output(Expr,set(Es)):-
    length(List,L),		  % generate `input' list
    all_qs(List),		  % in increasing order
    fsa_regex:rx(is_functional(List o Expr),Fa),
    Fa = fa(_,1,[0],[],[],[]), !, % for this length non functional,
    length(CVs,L),
    findall(CVs,non_functional_cv(CVs,Expr),Es).

non_functional_cv(CVs,Expr) :-
    cvs(CVs),
    fsa_regex:rx(is_functional(CVs o Expr),Fa),
    Fa = fa(_,1,[0],[],[],[]). 

cvs([]).
cvs([CV|CVs]) :-
    cv(CV),
    cvs(CVs).

cv(cons).
cv(vowel).

macro(wrong_output(Expr),E) :-
    fsa_regex:rx(exact(Expr),Fa0),
    (	Fa0 = fa(_,1,[0],[],[],[])            % if Expr is not functional
    ->	wrong_output(Expr,E)
    ;   E='{}'
    ).

wrong_output(Expr,set(Es)):-
    length(List,L),		  % generate `input' list
    all_qs(List),		  % in increasing order
    fsa_regex:rx(exact(List o Expr),Fa),
    Fa = fa(_,1,[0],[],[],[]), !, % for this length not exact,
    length(CVs,L),
    findall(CVs,non_exact_cv(CVs,Expr),Es).

non_exact_cv(CVs,Expr) :-
    cvs(CVs),
    fsa_regex:rx(exact(CVs o Expr),Fa),
    Fa = fa(_,1,[0],[],[],[]). 

