%% spelling rules to be used in conjunction with dutch2level.pl

:- multifile macro/2.

:- ensure_loaded('../twolevel').
:- ensure_loaded('../same_length').

:- bb_put(fsa_rx_spy:spelling,on).
:- bb_put(fsa_rx_spy:cons_doubling,on).
:- bb_put(fsa_rx_spy:long_vowel,on).
:- bb_put(fsa_rx_spy:slaan,on).
:- bb_put(fsa_rx_spy:e_deletion,on).
:- bb_put(fsa_rx_spy:eUmlaut,on).
:- bb_put(fsa_rx_spy:devoicing,on).
:- bb_put(fsa_rx_spy:tkofschip,on).
:- bb_put(fsa_rx_spy:tPresInd,on).

:- fsa_globals:fsa_global_set(regex_cache,on).

macro(spelling,
      twolevel([cons_doubling,long_vowel,e_deletion,eUmlaut,
		slaan,devoicing,tPresInd,tkofschip,gebrand])).


%% rules for +en, +e suffixes
%% kop+en = koppen
%% reken+en = rekenen, hamer+en = hameren handel+en = handelen,
%% wijzig+en = wijzigen, adem+en
%% currently, these cases are accounted for in the lexicon, by encoding stems
%% as reke1n, hame1r, etc. 
macro(cons_doubling,
      var(C, double_cons,
	  comp(pairs, '+':C, [{cons,boundary},vowel, C],[e])
	 )
     ).

%% attempt 3: (ugly but seems to work) (see below)
%% first forget e,
macro(long_vowel,
      var(V, {a,i,o,u},  comp(pairs, V:0,   [V], [cons,plus0, e]))).
%% now delete first e (to eliminate LC) and introduce a disjunctive RC
%% leez+en = lezen, zie+en = zien, doe+en = doen, roze+e = roze
%% zee+en =*zeen, fix: final e goes to e-umlaut (however encoded),
%% and e-deletion rule does not apply here.
macro(e_deletion,
      comp(pairs, e:0,   [], {[e,cons,plus0,e],[plus0,e]})).
%% n.b. dont constrain surface form of e's in LC to prevent
%% e-deletion from applying (which would block this rule)
macro(eUmlaut,
      comp(pairs, e:'E', [e:{0,e},e:{0,e},plus0],[])). 

%% sla+en = slaan ga+en = gaan sta+en = staan
macro(slaan,
      comp(pairs, e:a,   [{[s,{l,t}],g},a,plus0],[])).

%% rules for +t suffix & 0-suffix
%% leez+t = leest, leev+t = leeft
%% leez+# = lees, leev+Te = leefde
macro(devoicing,
      comp(pairs, {z:s,v:f}, [],[plus0,{t,'T':d,boundary}])).
%% zet+t = zet
macro(tPresInd,
      comp(pairs, t:0,  [t,plus0],[])).

%% rules for past tense
%% ramsjte
%%  shakete, racete, bridgede, mimede!
macro(tkofschip,
      comp(pairs, 'T':t, [{[s,j],t,k,f,s,p,[c,h],x},e^ ,plus0],[])).

%% #ge+brand+T# = gebrand
macro(gebrand,
      comp(pairs, 'T':0,  [{t,d},plus0],[boundary])
     ).

%% auxiliary defs

macro(letter, a..z).
macro(vowel, {a,e,i,o,u}).
macro(cons, letter - vowel).
macro(double_cons, cons - {j,w,x,y}).

%% feasible pairs
macro(boundary, '#':0).
macro(plus0, '+':0).
macro(plus, '+' x {0,double_cons}).
macro(diacritics, {'1','2'}:0).
macro(deleted_vowel, vowel x 0).
%%zie+en, sla+en leez+t, leev+t, kop+Te, ren+Te, 
macro(special, {e x {0,a,'E'},z:s,v:f,'T':t,'T':d,'T':0,t:0}). 
macro(pairs,{letter,plus,boundary,diacritics,deleted_vowel,special}).



