:- module(fsa_m_array, [%% updatable arrays using mutable terms
			fsa_m_array_new/1,
			fsa_m_array_new/2,
			fsa_m_array_get/3,
			fsa_m_array_get/4,
			fsa_m_array_put/3,
			fsa_m_array_put/5
		    ]).


#swi# :- style_check([-atom]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.
:- public help_info/4.

user:help_info(module,fsa_m_array,"Mutable Arrays",
"This module provides a mutable array datastructure. The arrays are \
 \n implemented using O'Keefe's N+K trees, with N=127 and K=32.\
 \n \
 \n NB. Array indices start at 0: so 0 refers to the first element of\
 \n the array.\
 \n \
 \n Here's an overview of the predicates provided:  \
 \n \
 \n MutableFsaArray:\
 \n \
 \n * fsa_m_array_new/[1,2]       create a new mutable array\
 \n * fsa_m_array_get/3           lookup a value from a mutable array\
 \n * fsa_m_array_put/[3,5]       update a value in a mutable array\
 \n \
 \n The N+K tree data-structure is described in The Craft of Prolog, by \
 \n Richard A. O'Keefe, MIT Press, 1990, chapters 4.5. and 4.6.").

help_info(class,pred,"List of Predicates",
"This section lists the predicates defined by this module.").

help_info(pred,fsa_m_array_new,"fsa_m_array_new(-MutableFsaArray,[+Size])",
"Initializes MutableFsaArray as a new mutable array."). 

fsa_m_array_new(T) :-
    fsa_array:fsa_array_new(T).

fsa_m_array_new(T,Size) :-
    fsa_array:fsa_array_new(T,Size).

help_info(pred,fsa_m_array_get,
"fsa_m_array_get(+Index,?Val[,?Default],+MutableFsaArray)",
"Val is unified with the Index'th entry in MutableFsaArray. The\
 \n predicate *succeeds* if that entry has not yet been set, without\
 \n binding Val (first form); or it binds Val to Default (second form)."). 
fsa_m_array_get(Index,Val,M_thing) :-
    fsa_array:fsa_array_access(Index,Val0,M_thing),
    (	var(Val0)
    ->	true %% this used to fail!
    ;	get_mutable(Val,Val0)
    ).

fsa_m_array_get(Index,Val,Default,M_thing) :-
    fsa_array:fsa_array_access(Index,Val0,M_thing),
    (	var(Val0)
    ->	Val=Default
    ;	get_mutable(Val,Val0)
    ).

help_info(pred,fsa_m_array_put,
"fsa_m_array_put(+Index,?Val,+MutableFsaArray)\
 \n fsa_m_array_put(+Index,?ValOld,?ValDefault,?Val,+MutableFsaArray)",
"The Index'th entry in MutableFsaArray is updated to Val (using the\
 \n SICStus built-in update_mutable/create_mutable). ValOld will be \
 \n bound to the old value, or to ValDefault if no value existed."). 
fsa_m_array_put(Index,Val,M_thing) :-
    fsa_array:fsa_array_access(Index,Val0,M_thing),
    (	var(Val0)
    ->	create_mutable(Val,Val0)
    ;	update_mutable(Val,Val0)
    ).

fsa_m_array_put(Index,Old,Default,Val,M_thing) :-
    fsa_array:fsa_array_access(Index,Val0,M_thing),
    (	var(Val0)
    ->	Old=Default,
	create_mutable(Val,Val0)
    ;	get_mutable(Old,Val0),
	update_mutable(Val,Val0)
    ).

