:- module(fsa_semiring, []).

#swi# :- style_check([-atom]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.


user:help_info(class,semiring,"Types of transducers",
"The determinization, minimization and minimum path algorithms for\
 \n transducers are implemented in a fully general way, i.e., for various\
 \n types of transducers (`semirings').  \
 \n \
 \n For each supported type, a number of predicates must be defined in a \
 \n corresponding module (these are called `semiring declarations'):\
 \n \
 \n * zero(Val).\
 \n * plus(Val0,Val1,Sum).\
 \n * minus(Val0,Val1,Diff).\
 \n * minimum(Val0,Val1,MinVal).\
 \n * minimum_only(YesNo).\
 \n \
 \n Currently, the following types of transducers are supported:\
 \n \
 \n * fsa_strings (ordinary (string-to-string) transducers)\
 \n * fsa_weights (weighted recognizers, aka string-to-weight transducers)\
 \n * fsa_weighted_strings (weighted transducers, aka\
 \n   string-to-weighted-string transducer).\
 \n \
 \n The examples directory SemiringModules might contain additional \
 \n semiring declarations.  ").


user:help_info(semiring,zero,"zero(?Val).",
"The identity element for addition. For strings, this is the empty\
 \n string; for weights it is 0.").

user:help_info(semiring,plus,"plus(+Val0,+Val1,?Sum).",
"Addition. For weights this is number addition, for strings this is\
 \n concatenation.").

user:help_info(semiring,minus,"minus(+Val0,+Val1,?Diff).",
"Inverse of the addition operator. ").

user:help_info(semiring,minimum,"minimum(+Val0,+Val1,?Min).",
"Minimum value of two given values. For weights this is the minumum of\
 \n two numbers, for strings this is the longest common prefix.").

user:help_info(semiring,minimum_only,"minimum_only(+YesNo).",
"YesNo is one of the atoms\
 \n \
 \n * yes\
 \n * no\
 \n \
 \n indicating whether we are intested in all outputs associated with a\
 \n path or only in the minimal output. For weights this is `yes', for\
 \n strings this is `no'.").

zero(Val,A) :-
    (	Val==fsa_strings -> fsa_strings:zero(A)
    ;   Val==fsa_weights -> fsa_weights:zero(A)
    ;   Val==fsa_weighted_strings -> fsa_weighted_strings:zero(A)
    ;   Val:zero(A)
    ).


plus(Val,A,B,C) :-
    (	Val==fsa_strings -> fsa_strings:plus(A,B,C)
    ;   Val==fsa_weights -> fsa_weights:plus(A,B,C)
    ;   Val==fsa_weighted_strings -> fsa_weighted_strings:plus(A,B,C)
    ;   Val:plus(A,B,C)
    ).

minus(Val,A,B,C) :-
    (	Val==fsa_strings -> fsa_strings:minus(A,B,C)
    ;   Val==fsa_weights -> fsa_weights:minus(A,B,C)
    ;   Val==fsa_weighted_strings -> fsa_weighted_strings:minus(A,B,C)
    ;   Val:minus(A,B,C)
    ).

minimum(Val,A,B,C) :-
    (	Val==fsa_strings -> fsa_strings:minimum(A,B,C)
    ;   Val==fsa_weights -> fsa_weights:minimum(A,B,C)
    ;   Val==fsa_weighted_strings -> fsa_weighted_strings:minimum(A,B,C)
    ;   Val:minimum(A,B,C)
    ).

minimum_only(Val,A) :-
    (	Val==fsa_strings -> fsa_strings:minimum_only(A)
    ;   Val==fsa_weights -> fsa_weights:minimum_only(A)
    ;   Val==fsa_weighted_strings -> fsa_weighted_strings:minimum_only(A)
    ;   Val:minimum_only(A)
    ).


