%%% -*- Module: fsa_t_determinizer; -*-
%% determinize a FST 
%% TODO: implement a check whether input can be determinized.
%% DONE: cf Examples/RocheSchabes97

:- module(fsa_t_determinizer, []).

#swi# :- style_check([-atom]).

:- use_module(fsa_semiring).

#yap# :- use_module(library(lists),   [ append/3, select/3 ]).
#sic# :- use_module(library(lists),   [ select/3 ]).
#sic# :- use_module(library(ordsets), [ ord_union/2 ]).
#yap# :- use_module(library(ordsets), [ ord_union/2 ]).
#swi# :- use_module(ordsets, [ ord_union/2 ]).

t_determinize(Flag,Fa0,Fa):-
    fsa_semiring:minimum_only(Flag,Y),
    fsa_semiring:zero(Flag,Zero),
    fsa_data:fsa_symbols_decl(Fa0,Decl),
    domain_module(Decl,Module),
    prepare_ts_and_js_and_fs(Fa0,JT,JM,TT,FT,Zero),
    fsa_data:fsa_start_states(Fa0,Start0),
    add_zero(Start0,Start1,Zero),
    jump_reachables(Start1,Start,JT,JM,Flag),
    fsa_hash:fsa_hash_new(Done),
    fsa_hash:fsa_hash_access(Start,n,Done),
    add_final(Start,Fins0,Fins1,FT),
    fsa_util:debug_message(2,
	"Eh, I'm starting fsa_t_determinizer. Might loop!~n",[]),
    subset_c([Start],Done,Trans0,Fins1,FT,JT,JM,TT,Flag,Y,Module),
    fsa_util:debug_message(2,
	"Good. I survived fsa_t_determinizer.~n",[]),
    td_finals_nd(Fins0,Fins,Trans,Trans0,Flag,Zero,Y),
    fsa_data:fsa_construct_rename_states(Decl,[Start],Fins,Trans,[],Fa).

domain_module(t(M,_),M).
domain_module(r(M,_),M).
domain_module(t(M,_,_),M).

add_zero([],[],_).
add_zero([State|States0],[q(State,Zero)|States],Zero):-
    add_zero(States0,States,Zero).

%% add transitions for output at final state
td_finals_nd(F0,F,Tr0,Tr,Flag,Zero,Y) :-
    minimum_only(Y,F0,F1,Flag),
    td_finals_nd1(F1,F,Tr0,Tr,Zero).

td_finals_nd1([],[],T,T,_Flag).
td_finals_nd1([q(P0,Output)|FinalTds],[P|Fins],T0,T,Flag) :- 
    td_final_nd(Output,P0,P,T0,T1,Flag),
    td_finals_nd1(FinalTds,Fins,T1,T,Flag).

td_final_nd(Score,P,Final,Ts0,Ts,Zero):-
    (	Score == Zero
    ->  Ts0=Ts, Final=P
    ;   Final=final, Ts0=[trans(P,[]/Score,final)|Ts]
    ).

%%%%%%%%%%%%%%%
%%%% jumps %%%%
%%%%%%%%%%%%%%%

%% jumps are treated as in the per_subset variant of determinizer.pl

prepare_ts_and_js_and_fs(Fa,JumpTable,MemoTable,TransTable,FinTable,Flag) :-
    fsa_data:fsa_states_number(Fa,Size),
    fsa_data:fsa_jumps(Fa,Js),
    fsa_data:fsa_transitions(Fa,Ts),
    fsa_array:fsa_array_new(JumpTable,Size),
    fsa_array:fsa_array_new(TransTable,Size),
    fsa_hash:fsa_hash_new(MemoTable),
    cl_jumps(Js,Jumps0,[],Flag),
    cl_transitions(Ts,Jumps1,Jumps0,Trans1,[]),
    sort(Jumps1,Jumps),
    sort(Trans1,Trans),
    fsa_data:jumps_to_thing(Jumps,JumpTable),
    fsa_data:transitions_to_thing(Trans,TransTable),
    fsa_data:index_finals(Fa,FinTable).

cl_jumps([],J,J,_).
cl_jumps([jump(A,B)|Js],[jump(A,B/Zero)|J0],J,Zero):-
    cl_jumps(Js,J0,J,Zero).

cl_transitions([],J,J,T,T).
cl_transitions([Trans|Trs],J0,J,T0,T) :-
    cl_transition(Trans,J0,J1,T0,T1),
    cl_transitions(Trs,J1,J,T1,T).

cl_transition(trans(State0,E/Out,State),J0,J,T0,T) :-
    (	E == []
    ->	J0 = [jump(State0,State/Out)|J],
	T0=T
    ;	T0=[trans(State0,E,q(State,Out))|T],
	J0=J
    ).

subset_c([],_Done,[],[],_,_,_,_,_,_,_).
subset_c([Subs|Ag0],Done,Trans,Fins0,Fa,JT,JM,TT,Flag,Y,Mod) :-
    collect_ts(Subs,NewT0,TT,Flag),
    fsa_determinizer:merge_transitions(NewT0,NewT10),
    fsa_pred_module:t_determinize_preds(Mod,NewT10,NewT1),
    add_jumps(NewT1,Subs,JT,JM,Flag,Y,Ag0,Ag,Done,Trans0,Trans,Fins0,Fins,Fa,Mod),
    subset_c(Ag,Done,Trans0,Fins,Fa,JT,JM,TT,Flag,Y,Mod).

add_jumps([],_,_,_,_,_,Ag,Ag,_,Trs,Trs,Fs,Fs,_,_).
add_jumps([Sym0-States0|Rest],St,JT,JM,Flag,Y,Ag0,Ag,Done,Trs0,Trs,Fs0,Fs,Fa,Mod):-
    %% add jumps:
    jump_reachables(States0,States1,JT,JM,Flag),
    %% consider_simplifications in output:
    simplify_per_state_if_preds(Flag,Mod,States1,States2),
    sigma_list(States2,Sigma,Flag),
    minus_sigma(States2,Sigma,States3,Flag),
    minimum_only(Y,States3,States,Flag),
    %% add to agenda, hash, transitions, final states:
    add(St,Sym0/Sigma,States,Ag0,Ag1,Done,Trs0,Trs1,Fs0,Fs1,Fa),
    add_jumps(Rest,St,JT,JM,Flag,Y,Ag1,Ag,Done,Trs1,Trs,Fs1,Fs,Fa,Mod).

simplify_per_state_if_preds(fsa_weights,_Mod,States,States).
simplify_per_state_if_preds(fsa_strings,Mod,States0,States) :-
    (	Mod == fsa_frozen
    ->	States0=States
    ;	simplify_per_state(States0,States,Mod,fsa_strings)
    ).
simplify_per_state_if_preds(fsa_weighted_strings,Mod,States0,States) :-
    (	Mod == fsa_frozen
    ->	States0=States
    ;	simplify_per_state(States0,States,Mod,fsa_weighted_strings)
    ).

simplify_per_state([],[],_,_Flag).
simplify_per_state([q(Q,Seq)|Qs0],States,Mod,Flag) :-
    simplify_per_state_(Qs0,Q,[Seq|Tail],Tail,States,Mod,Flag).

simplify_per_state_([],Q,SeqList,[],States,Mod,Flag) :-
    simplify_seqs(SeqList,Q,States,[],Mod,Flag).
simplify_per_state_([q(NQ,Seq)|Qs0],Q,S0,S,States,Mod,Flag) :-
    (	NQ == Q
    ->	S = [Seq|S1],
	simplify_per_state_(Qs0,Q,S0,S1,States,Mod,Flag)
    ;   S=[],
	simplify_seqs(S0,Q,States,States1,Mod,Flag),
	simplify_per_state_(Qs0,NQ,[Seq|Tail],Tail,States1,Mod,Flag)
    ).

simplify_seqs(Seqs0,Q,Pairs,Pairs0,Mod,Flag) :-
    simplify_seqs(Seqs0,Seqs,Mod,Flag),
    to_pairs(Seqs,Q,Pairs,Pairs0).

to_pairs([],_,P,P).
to_pairs([Seq|Seqs],Q,[q(Q,Seq)|P0],P) :-
    to_pairs(Seqs,Q,P0,P).


%% inefficient algorithm for now...
%% and moreover I don't know a more efficient one.
simplify_seqs(Seqs0,Seqs,Mod,Flag) :-
    (	simplify_one(Seqs0,Seqs1,Mod,Flag)
    ->	simplify_seqs(Seqs1,Seqs,Mod,Flag)
    ;   Seqs0=Seqs
    ).

simplify_one([H1|T],[NH|NT],Mod,Flag) :-
    select(H2,T,NT),
    simpl(Flag,H1,H2,NH,Mod).
simplify_one([H|T0],[H|T],Mod,Flag) :-
    simplify_one(T0,T,Mod,Flag).


%% Do we need an interface to proper cleanup, including
%% expand-into-non-overlap? No, not required, since this is
%% single symbols, not pairs. What about identities? You can
%% never get different, simplifyable, identities since identities
%% correlate to input symbols, and if these get merged they
%% merge the corresponding outputs too.
simpl(fsa_weighted_strings,A/W,B/W,C/W,Mod) :-
    simpl(fsa_strings,A,B,C,Mod).

simpl(fsa_strings,[H1|T],[H2|T],[NH|T],Mod) :-
    H1 \== H2,
    !,   % to help spdet
    fsa_pred_module:cleanup(Mod,[H1,H2],[NH]).

simpl(fsa_strings,[H|T1],[H|T2],[H|T],Mod) :-
    simpl(fsa_strings,T1,T2,T,Mod).    

add(Q0,Sym,Q,A0,A,Done,Result0,[trans(Q0,Sym,Q)|Result0],
     Fins0,Fins,Fa) :-
    fsa_hash:fsa_hash_access(Q,Val,Done),
    (	var(Val)
    ->	A = [Q|A0],
	Val = n,
	add_final(Q,Fins0,Fins,Fa)
    ;	A0 = A,
	Fins0 = Fins
    ).

collect_ts([],[],_,_Flag).
collect_ts([q(Begin,His0)|Tail],List0,TT,Flag):-
    fsa_array:fsa_array_access(Begin,L1,TT),
    (   var(L1)
    ->  List0=List1
    ;   add_outputs(L1,His0,L2,Flag),
	List0=[L2|List1]
    ),
    collect_ts(Tail,List1,TT,Flag).

add_outputs([],_,[],_).
add_outputs([Sym-R0|List0],His,[Sym-R|List],Flag):-
    add_output(R0,His,R,Flag),
    add_outputs(List0,His,List,Flag).

add_output([],_,[],_).
add_output([q(Q,His1)|T0],His0,[q(Q,His)|T],Flag):-
    fsa_semiring:plus(Flag,His0,His1,His),
    add_output(T0,His0,T,Flag).

minus_sigma([],_,[],_Flag).
minus_sigma([q(State,W0)|T0],Sigma,[q(State,W)|T],Flag) :-
    fsa_semiring:minus(Flag,W0,Sigma,W),
    minus_sigma(T0,Sigma,T,Flag).

sigma_list([q(_,H)|T],Min,Flag) :-
    sigma_list(T,H,Min,Flag).

sigma_list([],H,H,_).
sigma_list([q(_,H)|T],Min0,Min,Flag) :-
    fsa_semiring:minimum(Flag,H,Min0,Min1),
    sigma_list(T,Min1,Min,Flag).

add_final(List,Fs0,Fs,Fa):-
    add_final0(List,List,Fs0,Fs,Fa).

add_final0([],_List,Fs,Fs,_).
add_final0([q(NdState,Output)|Tail],State,Fs0,Fs,FT):-
    fsa_array:fsa_array_access(NdState,Val,FT),
    (	nonvar(Val)
    ->	Fs0=[q(State,Output)|Fs1]
    ;   Fs0=Fs1
    ),
    add_final0(Tail,State,Fs1,Fs,FT).

jump_reachables(States0,States,JTable,JsTable,Flag) :-
    fsa_hash:fsa_hash_access(States0,States,JsTable),
    (	var(States)
    ->	fsa_hash:fsa_hash_new(Done),
	fsa_hash_access_list(States0,Done),
	jump_reachable(States0,States1,States0,Done,JTable,Flag),
	sort(States1,States)
    ;   true
    ).

fsa_hash_access_list([],_Done).
fsa_hash_access_list([H|T],Done) :-
    fsa_hash:fsa_hash_access(H,[],Done),
    fsa_hash_access_list(T,Done).

jump_reachable([],R,R,_Done,_Table,_Flag).
jump_reachable([q(H,His0)|T],Results0,Results,Done,JTable,Flag) :-
    fsa_array:fsa_array_access(H,Next0,[],JTable),
    add_if_new(Next0,His0,T,Ag,Results0,Results1,Done,Flag),
    jump_reachable(Ag,Results1,Results,Done,JTable,Flag).

add_if_new([],_,Ag,Ag,R,R,_Done,_).
add_if_new([H/His1|T],His0,Ag0,Ag,Res0,Res,Done,Flag) :-
    fsa_semiring:plus(Flag,His0,His1,His),
    fsa_hash:fsa_hash_access(q(H,His),XX,Done),
    (   var(XX)
    ->  Ag1=[q(H,His)|Ag0],
        XX=[],
        Res0=[q(H,His)|Res1]
    ;   Ag1=Ag0,
        Res0=Res1
    ),
    add_if_new(T,His0,Ag1,Ag,Res1,Res,Done,Flag).

minimum_only(no,F,F,_).
minimum_only(yes,Ends0,Ends,Flag):-
    minimum_only00(Ends0,Ends,Flag).

minimum_only00([],[],_).
minimum_only00([q(State,Score)|Tail0],[q(State,Min)|Tail],Flag):-
    minimum_only000(Tail0,State,Score,Min,Tail,Flag).

minimum_only000([],_,S,S,[],_).
minimum_only000([q(State,Score1)|Tail0],State0,Score0,Min,Tail,Flag):-
    (	State==State0
    ->	fsa_semiring:minimum(Flag,Score0,Score1,Score),
	minimum_only000(Tail0,State,Score,Min,Tail,Flag)
    ;   Score0=Min,
	Tail=[q(State,NMin)|Tail1],
	minimum_only000(Tail0,State,Score1,NMin,Tail1,Flag)
    ).

%% add new start states
t_minimize(Flag,Fa0,Fa) :-
    fsa_minimum_path:fsa_minimum_path_array(Fa0,Ps,Flag),
    fsa_data:fsa_components(Syms,States0,Starts0,Finals,Trans0,[],Fa0),
    fsa_semiring:zero(Flag,Zero),
    new_min_transitions(Trans0,Starts0,Trans,Ps,Flag,Zero,Jumps,States0),
    States is States0+1,   %% for new start state
    fsa_data:fsa_construct(Syms,States,[States0],Finals,Trans,Jumps,Fa).


new_min_transitions(Trans0,Starts0,Trans,Ps,Flag,Zero,Jumps,Nr):-
    new_min_transitions0(Trans0,Trans1,Ps,Flag,Zero,Jumps1),
    start_min_transitions(Starts0,Trans,Trans1,Ps,Flag,Zero,Jumps,Jumps1,Nr).

new_min_transitions0([],[],_,_,_,[]).
new_min_transitions0([trans(P0,SymA/SymB,P)|Ts0],Ts,Ps,Flag,Zero,Js):-
    %% if P is not on array, then P is sink-state (hence P0 is sink too)
    %% so it really doesn't matter much what we do with this transition,
    %% since it will be removed by minimizer. Except that we shouldn't let
    %% the whole mechanism fail..
    fsa_m_array:fsa_m_array_get(P,Bs/_,Zero/_,Ps),
    fsa_semiring:plus(Flag,SymB,Bs,Syms0),
    fsa_m_array:fsa_m_array_get(P0,As/_,Zero/_,Ps),
    fsa_semiring:minus(Flag,Syms0,As,Syms),
    trans_or_jump(SymA,Zero,Syms,P0,P,Ts,Ts1,Js,Js1),
    new_min_transitions0(Ts0,Ts1,Ps,Flag,Zero,Js1).

trans_or_jump(Sym,Zero,Syms,P0,P,Ts,Ts1,Js,Js1) :-
    (	Sym==[], Syms==Zero
    ->  Js = [jump(P0,P)|Js1],
	Ts=Ts1
    ;   Ts = [trans(P0,Sym/Syms,P)|Ts1],
	Js=Js1
    ).

%%start_min_transitions(Starts0,Trans,Trans1,Ps,Fa,Flag,Jumps,Jumps1,Nr).
start_min_transitions([],Ts,Ts,_,_,_,Js,Js,_).
start_min_transitions([St10|Sts],Ts0,Ts,Ps,Flag,Zero,Js0,Js,Nr):-
    fsa_m_array:fsa_m_array_get(St10,Syms/_,Zero/_,Ps),
    trans_or_jump([],Zero,Syms,Nr,St10,Ts0,Ts1,Js0,Js1),
    start_min_transitions(Sts,Ts1,Ts,Ps,Flag,Zero,Js1,Js,Nr).

%% trouble starts here..
%% nb the `normal' determinizer uses caching here. We can't do
%% that since in this case the associated values contain predicates
%% too, which moreover might get changed.
t_determinize_preds(Union0,Union,Module):-
    t_determinize_preds_with_targets(Union0,Keylist,Module),
    t_preds_to_union(Keylist,Union,Module).

t_preds_to_union([],[],_).
t_preds_to_union([Pred-Lists|Preds],[Pred-Output|Union],Mod):-
    id(Pred,PredN,_,_),
    add_id_output(Lists,Pred,PredN,Output0,Mod),
    ord_union(Output0,Output),
    t_preds_to_union(Preds,Union,Mod).

add_id_output([],_,_,[],_).
add_id_output([Bool-Val0|Tail],Pred,PredN,[Val|Output],Mod):-
    (	Bool==1
    ->	alter_last_output(Val0,Pred,Val)
    ;   Pred==PredN
    ->	Val0=Val
    ;   add_deq(Val0,Val)
    ),
    add_id_output(Tail,Pred,PredN,Output,Mod).

add_deq([],[]).
add_deq([q(State,List0)|Tail0],[q(State,List)|Tail]):-
    append(List0,['$@dequeue'],List),
    %% push_deq(List1,List),   IS THIS NECCESSARY/ENOUGH?
    add_deq(Tail0,Tail).

alter_last_output([],_,[]).
alter_last_output([q(State,[H|T]/Sc)|Tail0],Pred,[q(State,NList/Sc)|Tail]):-
    !,
    alter_last(T,H,Pred,NList),
    alter_last_output(Tail0,Pred,Tail).
alter_last_output([q(State,[H|T])|Tail0],Pred,[q(State,NList)|Tail]):-
    alter_last(T,H,Pred,NList),
    alter_last_output(Tail0,Pred,Tail).

alter_last([],_,Pred,[Pred]).
alter_last([H|T],P,Pred,[P|List]):-
    alter_last(T,H,Pred,List).
    
t_determinize_preds_with_targets(Union0,Union,M) :-
    findall(Pred-Targets,fsa_t_determinizer:t_exclusive_pred(Union0,Pred,Targets,M),Union).

t_exclusive_pred([First-Target0|Union0],Pred,Targets,M):-
    t_take_or_leave0(First,Pred0,Target0,Targets,Targets2,F1,M),
    t_exclusive_pred(Union0,Pred0,Pred,Targets2,F1,1,M).

t_exclusive_pred([],Pred,Pred,[],F,F,_M).
t_exclusive_pred([Pred1-Target|Rest],Pred0,Pred,Targets,F0,F,M):-
    t_take_or_leave(Pred0,Pred1,Pred2,Target,Targets,Targets0,F0,F1,M),
    t_exclusive_pred(Rest,Pred2,Pred,Targets0,F1,F,M).

%% take
t_take_or_leave0(Pred01,Pred01,Target,[B-Target|Targets],Targets,1,_):-
    id(Pred01,_,0,B).

%% leave
t_take_or_leave0(Pred01,Pred2,_,T,T,0,M):-
    id(Pred01,Pred1,_,_),
    fsa_pred_module:negation(M,Pred1,Pred2).

%% take
t_take_or_leave(Pred00,Pred01,Pred2,Target,[B-Target|Targets],Targets,_,1,M) :-
    id(Pred01,Pred1,0,B),
    id(Pred00,Pred0,B,C),
    fsa_pred_module:conjunction(M,Pred0,Pred1,Pred02),
    identity(C,Pred02,Pred2,M).

%% leave
t_take_or_leave(Pred00,Pred01,Pred2,_,T,T,F,F,M) :-
    id(Pred00,Pred0,0,B),
    id(Pred01,Pred1,_,_),
    fsa_pred_module:negation(M,Pred1,Npred1),
    fsa_pred_module:conjunction(M,Pred0,Npred1,Pred02),
    identity(B,Pred02,Pred2,M).

id(A,B,C,D):-
    (	A='$@'(B)
    ->  D=1
    ;   A=B,
	C=D
    ).
    
identity(0,Pred,Pred,_).
%% identity(1,Pred,'$@'(Pred)).
identity(1,Pred0,Pred,M) :-
    fsa_pred_module:identity(M,Pred0,Pred).

/*
%% not used 
push_deq(L0,L):-
    (	push1(L0,L1)
    ->	push_deq(L1,L)
    ;   L0=L
    ).

push1(L0,L):-
    append(Pref,[Pred,'$@dequeue'|R],L0),
    Pred \== '$@dequeue',
    Pred \= '$@'(_),
    append(Pref,['$@dequeue',Pred|R],L).

*/

