%%% -*- Module: fsa_util; -*-
:- module(fsa_util,[]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

#swi# :- style_check([-atom]).

#swi# :- redefine_system_predicate(fsa_util:time(_)).

%% documentation!
#yap# :- use_module(library(charsio),[ format_to_chars/3,read_from_chars/2 ]).
#!yap# :- use_module(charsio,[ format_to_chars/3,read_from_chars/2 ]).
#yap# :- use_module(library(lists),  [ member/2,append/3 ]).

:- meta_predicate time(:), time(?,:), time(?,:,?), time(?,:,?,?).
:- meta_predicate call_with_ops(?,:).
:- meta_predicate call_with_ops(:).
:- meta_predicate debug_call(?,:).
:- meta_predicate to_file(:,?).

time(Call):-
    time(2,Call,_,_).

%time(D,Call) :-
%    time(D,Call,_,_).
%
%time(D,Call,T) :-
%    time(D,Call,T,_).

time(DebugLevel,Call,Time,Space) :-
    parse_call(Call,Module,Fun,Ar),
#swi#     statistics(cputime,Time00),
#swi#     Time0 is integer(1000*Time00),
#sic#     statistics(runtime,[Time0,_]),
#yap#     statistics(runtime,[Time0,_]),
#sic#     statistics(memory,[Space0,_]),
#yap#     statistics(heap,[A,_]),
#yap#     statistics(global_stack,[B,_]),
#yap#     statistics(local_stack,[C,_]),
#yap#     statistics(trail,[D,_]),
#yap#     Space0 is A+B+C+D,
#swi#     statistics(heapused,A),
#swi#     statistics(globalused,B),
#swi#     statistics(localused,C),
#swi#     statistics(trailused,D),
#swi#     Space0 is A+B+C+D,
    call(Call),
    report_time(DebugLevel,Time0,Space0,Module,Fun,Ar,Time,Space).

report_time(DebugLevel,Time0,Space0,Module,Fun,Ar,Time,Space) :-
#swi#     statistics(cputime,Time10),
#swi#     Time1 is integer(1000*Time10),
#sic#     statistics(runtime,[Time1,_]),
#yap#     statistics(runtime,[Time1,_]),
#sic#     statistics(memory,[Space1,_]),
#yap#     statistics(heap,[A,_]),
#yap#     statistics(global_stack,[B,_]),
#yap#     statistics(local_stack,[C,_]),
#yap#     statistics(trail,[D,_]),
#yap#     Space1 is A+B+C+D,
#swi#     statistics(heapused,A),
#swi#     statistics(globalused,B),
#swi#     statistics(localused,C),
#swi#     statistics(trailused,D),
#swi#     Space1 is A+B+C+D,
    Time is Time1-Time0,
    Space is max(0,integer((Space1-Space0)/1000)),
    debug_message(DebugLevel,"~a: ~w ~w msec; ~w kbytes~n",
                   [Module,Fun/Ar,Time,Space]).

parse_call(Term,Module,Fun,Ar) :-
    strip_modules(Term,fsa_util,Module,Fun,Ar).

strip_modules(Term,Module0,Module,Fun,Ar) :-
    (	Term=Module1:Term1
    ->  strip_modules(Term1,Module1,Module,Fun,Ar)
    ;   functor(Term,Fun,Ar), Module0=Module
    ).

read_with_ops(Term) :-
    fsa_regex:default_ops(Ops),
    read_with_ops(Ops,Term).

%% read_with_ops(+ListOfOperatorDefinitions,?TermRead)
read_with_ops(Ops,Term) :-
    call_with_ops(Ops,read(Term)).
%read_with_ops(Stream,Ops,Term) :-
%    call_with_ops(Ops,read(Stream,Term)).


call_with_ops(Call) :-
    fsa_regex:default_ops(Ops),
    call_with_ops(Ops,Call).

call_with_ops(Ops,Call):-
    get_current_ops(Ops,Old),
    set_ops(Ops),
    user:call_cleanup(Call,fsa_util:set_ops(Old)).

get_current_ops([],[]).
get_current_ops([op(_A,B,C)|T],Later) :-
    get_current_op(B,C,Later,Later0),
    get_current_ops(T,Later0).

get_current_op(B,C,L0,L) :-
    (	relevant_current_op(B,C,A0,B0)
    ->	L0 = [op(A0,B0,C)|L]
    ;	%% op(A,B,C),
	L0 = [op(0,B,C)|L]
    ).

%% set_ops(Operators)
set_ops(Ops0) :-
    sort(Ops0,Ops),   %% unset operators before setting, this is
                      %% required by ISO standard.
    set_ops0(Ops).

set_ops0([]).
set_ops0([op(A,B,C)|T]) :-
    op(A,B,C),
    set_ops0(T).

relevant_current_op(B,C,A0,B0):-
    current_op(A0,B0,C),
    relevant_pattern(B0,B).

relevant_pattern(A,B) :-
    pref(A), pref(B), !.
relevant_pattern(A,B) :-
    post(A), post(B), !.
relevant_pattern(A,B) :-
    in(A), in(B), !.

pref(fx). 
pref(fy).

post(xf).
post(yf).

in(xfx).
in(xfy).
in(yfx).
in(yfy).


%% ord_intersect(+Set1, +Set2, Bool)
%% Bool is yes when the two ordered sets have at least one element in common.
%% It's no otherwise.
ord_intersect([],_,no).
ord_intersect([H|T],L2,Bool) :-
    ord_intersect0(L2,[H|T],Bool).

ord_intersect0([],_,no).
ord_intersect0([Head2|Tail2], [Head1|Tail1], Bool) :-
    compare(Order, Head1, Head2),
    ord_intersect(Order, Head1, Tail1, Head2, Tail2,Bool).

ord_intersect1([],_,_,no).
ord_intersect1([Head1|Tail1], Head2, Tail2,Bool) :-
    compare(Order, Head1, Head2),
    ord_intersect(Order, Head1, Tail1, Head2, Tail2,Bool).

ord_intersect(<, _, List1,H2,T2,Bool) :-
    ord_intersect1(List1,H2,T2,Bool).
ord_intersect(=, _, _, _, _,yes).
ord_intersect(>, Head1, Tail1, _, List2,Bool) :-
    ord_intersect2(List2,Head1,Tail1,Bool).

ord_intersect2([],_,_,no).
ord_intersect2([Head2|Tail2],Head1,Tail1,Bool):-
    compare(Order, Head1, Head2),
    ord_intersect(Order, Head1, Tail1, Head2, Tail2,Bool).


#!swi# %% between(+Lower, +Upper, ?Number)
#!swi# %% is true when Lower, Upper, and Number are integers,
#!swi# %% and Lower =< Number =< Upper.  If Lower and Upper are given,
#!swi# %% Number can be tested or enumerated.  If either Lower or Upper
#!swi# %% is absent, there is not enough information to find it, hence
#!swi# %% failure.
#!swi# %% Numbers are generated in ascending order. If you want descending
#!swi# %% order, use between/4.
#!swi# 
#!swi# between(Edge1,Edge2,I) :-
#!swi#     integer(Edge1),
#!swi#     integer(Edge2),
#!swi#     Edge1 =< Edge2, 
#!swi#     between(Edge1,Edge2,I,+).
#!swi# 
#!swi# %% between/4 as between/3, except that the fourth argument ('-' or
#!swi# %% '+') indicates whether integers are generated in descending or
#!swi# %% ascending order
#!swi# between(Lower,Upper,I,Dir) :-
#!swi#     (	integer(I) 
#!swi#     ->	Lower =< I,
#!swi# 	I =< Upper 
#!swi#     ;	between1(Dir,Lower,Upper,I) 
#!swi#     ).
#!swi# 
#!swi# %%  between1(Dir,Lower, Upper, Point)
#!swi# %%  enumerates values of Point satisfying Lower =< Point =< Upper,
#!swi# %%  where it is already known that Lower =< Upper and Point was a
#!swi# %%  variable.  Dir is an atom ('-' or '+') indicating whether the
#!swi# %%  integers are generated in descending or ascending order.
#!swi# 
#!swi# between1(+,L, U, L) :- 
#!swi#     L =< U.
#!swi# between1(+,L, U, N) :- 
#!swi#     L < U,
#!swi#     M is L+1,
#!swi#     between1(+,M, U, N).
#!swi# between1(-,L, U, U) :- 
#!swi#     L =< U.
#!swi# between1(-,L, U, N) :- 
#!swi#     U > L,
#!swi#     M is U-1,
#!swi#     between1(-,L, M, N).

#!swi# atom_term(Atom0,Term) :-
#!swi#     atom(Atom0),
#!swi#     format_to_chars("~a .",[Atom0],Chars),
#!swi#     %%                ^ space, because Atom may end with funny character
#!swi#     read_from_chars(Chars,Term).

#swi# atom_term(Atom,Term) :-
#swi#     atom_to_term(Atom,Term,[]).

#!swi# term_atom(Term,Atom) :-
#!swi#     (	number(Term)
#!swi#     ->	number__atom(Term,Atom)
#!swi#     ;	format_to_chars('~w',[Term],Chars),
#!swi# 	        atom_codes(Atom,Chars)
#!swi#     ).
#!swi# 

#swi# term_atom(T,A) :-
#swi#     term_to_atom(T,A).

number__atom(Num,Atom) :-
    (	number(Num)
    ->	number_codes(Num,String),
	atom_codes(Atom,String)
    ;	atom_codes(Atom,String),
	number_codes(Num,String)
    ).

debug_call(Level,Call) :-
    fsa_globals:fsa_global_get(debug,Level0),
    (	Level =< Level0
    ->	call(Call)
    ;	true
    ).

debug_message(Level,Format,Args) :-
    debug_call(Level,format(user_error,Format,Args)).

format_to_atom(String,Args,Atom):-
    format_to_chars(String,Args,Chars),
    atom_codes(Atom,Chars).

concat_all([],'').
concat_all([H|T],Atom):-
    concat_all2(T,H,Atom).

concat_all2([],H,H).
concat_all2([H|T],F,Atom) :-
    atom_concat(F,H,N),
    concat_all2(T,N,Atom).

concat_all([],'',_).
concat_all([H|T],Atom,Sep):-
    concat_all2(T,H,Atom,Sep).

concat_all2([],H,H,_).
concat_all2([H|T],F,Atom,Sep) :-
    atom_concat(F,Sep,F2),
    atom_concat(F2,H,N),
    concat_all2(T,N,Atom,Sep).

to_file(Goal,File) :-
    (	File==user
    ->	telling(Current),
	tell(user),
	user:call_cleanup(Goal,tell(Current))
    ;   File==user_error
    ->	telling(Current),
	tell(user_error),
	user:call_cleanup(Goal,tell(Current))
    ;   tell(File),
	user:call_cleanup(Goal,told)
    ).
