%%% -*- Module: fsa_regex; -*-
:- module(fsa_regex,
	  [ fsa_regex_atom_compile/2,
	    fsa_regex_atom_compile_file/2,
	    fsa_regex_compile/2, 
	    %% rx/2,
	    fsa_regex_compile_file/2,
	    fsa_regex_read_compile/1,
	    fsa_regex_read_compile_file/1,
	    fsa_reconsult_aux_file/1,
	    fsa_load_aux_file/1,
	    fsa_regex_clear_cache/0
	  ]).

#swi# :- style_check([-atom]).

:- use_module(fsa_util).

:- meta_predicate ltime(?,:).

:- public fsa_regex_aux:unknown_predicate_handler/2.
fsa_regex_aux:unknown_predicate_handler(macro(_,_),fail).
fsa_regex_aux:unknown_predicate_handler(rx(_,_),fail).

#!swi#:- use_module(library(terms), [ term_hash/2 ]).
#swi#:- use_module(terms, [ term_hash/2 ]).
#!swi# :- use_module(library(ordsets), [ ord_subtract/3, ord_union/3 ]).
#swi# :- use_module(ordsets, [ ord_subtract/3, ord_union/3 ]).
#yap# :- use_module(library(lists),   [ member/2, append/3, select/3,
#yap#					reverse/2, nth0/3, nth0/4 ]).
#sic# :- use_module(library(lists),   [ select/3,reverse/2, nth0/3, nth0/4 ]).
#sic# :- use_module(library(random), [ setrand/1, random/1, random/3 ]).
#sic# :- use_module(library(process), [ process_id/1 ]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous op_help/3.
:- discontiguous rx0/5.
:- discontiguous macro/2.
:- discontiguous prepare_reachable/5.
:- discontiguous prepare_reachable/6.
:- discontiguous is_final_reachable/4.
:- discontiguous trans_reachables/4.
:- discontiguous prepare_co_reachable/5.
:- discontiguous is_start_co_reachable/4.
:- discontiguous trans_co_reachables/4.
:- discontiguous help_info/4.

%%%%%%%%%%%%%%%%%%%%%%%%%
%%% public predicates %%%
%%%%%%%%%%%%%%%%%%%%%%%%%

user:help_info(xpred,fsa_load_aux_file,"fsa_load_aux_file(+File)",
"File is assumed to contain auxiliary regular expression operators. It\
 \n is loaded in module fsa_regex_aux and will be used for compiling\
 \n regular expressions.  \
 \n \
 \n Note that your file with definitions of regular expression operators\
 \n is compiled with the special Prolog-syntax operators for regular\
 \n expression notation loaded. Thus you can use * .. & etc. in your\
 \n definitions. Drawback is that you cannot use operator notation for\
 \n e.g. the is/2 predicate.  \
 \n \
 \n A typical auxiliary definition will be:\
 \n \
 \n     macro(vowel,{a,e,i,o,u}).\
 \n \
 \n A slightly more interesting one:\
 \n \
 \n     macro(free(Expr), ~ $ Expr).\
 \n \
 \n You can also explicitly construct an automaton yourself, e.g.:\
 \n \
 \n     rx(my_operator(Expr),Fa) :-\
 \n         fsa_regex:rx(Expr,Fa0),\
 \n         my_operator_definition(Fa0,Fa).\
 \n \
 \n so you can call fsa_regex:rx/2 for further compilations.").
fsa_load_aux_file(File) :-
    fsa_regex_clear_cache,
    fsa_util:call_with_ops(
#sic# 			   fsa_regex_aux:use_module(File)
#yap# 			   fsa_regex_aux:compile(File)
#swi# 			   fsa_regex_aux:consult(File)
),
    format(user_error,
	   "loaded auxiliary regular operator definitions from ~w~n",
	   [File]).

user:help_info(xpred,fsa_reconsult_aux_file,"fsa_reconsult_aux_file(+File)",
"File is assumed to contain auxiliary regular expression operators. It\
 \n is reconsulted in module fsa_regex_aux and will be used for compiling\
 \n regular expressions. Normally you want to use fsa_load_aux_file\
 \n instead. Use this predicate if you need to debug your Prolog\
 \n definitions in File."). 
fsa_reconsult_aux_file(File) :-
    fsa_regex_clear_cache,
    fsa_util:call_with_ops(
#!swi# 			   fsa_regex_aux:reconsult(File)
#swi# 			   fsa_regex_aux:consult(File)
			  ),
    format(user_error,
	   "reconsulted auxiliary regular operator definitions from ~w~n",
	   [File]).

user:help_info(xpred,fsa_regex_atom_compile_file,
"fsa_regex_atom_compile_file(+RegexAtom,+File)",
"RegexAtom is parsed as a regular expression. This expression is\
 \n compiled to a finite automaton which is written to File."). 
fsa_regex_atom_compile_file(Atom,File) :-
    fsa_regex_atom_compile(Atom,Fa),
    fsa_io:fsa_write_file(File,Fa).

user:help_info(xpred,fsa_regex_atom_compile,
"fsa_regex_atom_compile(+RegexAtom,+Fa)",
"RegexAtom is parsed as a regular expression. This expression is\
 \n compiled to a finite automaton Fa."). 
fsa_regex_atom_compile(Atom,Fa):-
    fsa_util:call_with_ops(fsa_util:atom_term(Atom,Expr)),
    fsa_regex_compile(Expr,Fa).

user:help_info(xpred,fsa_regex_read_compile_file,"fsa_regex_read_compile_file(File)",
"A regular expression is read-in from standard input. The expression is\
 \n compiled and the resulting automaton is saved in file File."). 
fsa_regex_read_compile_file(File) :-
    fsa_regex_read_compile(Fa),
    fsa_io:fsa_write_file(File,Fa).

user:help_info(xpred,fsa_regex_read_compile,"fsa_regex_read_compile(-Fa)",
"A regular expression is read-in from standard input. The expression is\
 \n compiled into an automaton Fa."). 
fsa_regex_read_compile(Fa):-
    prompt(Old,'RegEx: '),
    fsa_util:read_with_ops(Expr),
    prompt(_,Old),
    fsa_regex_compile(Expr,Fa).

user:help_info(xpred,fsa_regex_compile_file,"fsa_regex_compile_file(+Expr,+File)",
"The regular expression Expr (ground Prolog term) is compiled into an\
 \n automaton. The automaton is saved into File."). 
fsa_regex_compile_file(Expr,File):-
    fsa_regex_compile(Expr,Fa),
    fsa_io:fsa_write_file(File,Fa).

user:help_info(xpred,fsa_regex_compile,"fsa_regex_compile(+Term,-Fa) rx(+Term,-Fa)",
"Term is a regular expression. It is compiled into the automaton Fa.\
 \n The first form is typically used for a new regular expression compilation,\
 \n whereas the second form is used for embedded compilations (called from \
 \n user definitions). The only difference is that during debugging the depth\
 \n of recursion is set to zero for the first form. ").

%% top-level call
fsa_regex_compile(Term0,Fa0):-
    fsa_globals:fsa_global_get(debug,Dbg),
    fsa_globals:fsa_global_get(regex_cache,Cache),
    rx(Term0,Fa,Dbg,Cache,0),
    Fa0=Fa.


%% to be called from user rules
rx(Term,Fa0):-
    fsa_globals:fsa_global_get(debug,Dbg),
    fsa_globals:fsa_global_get(regex_cache,Cache),
    (   user:bb_get(fsa_regex:regex_rec,RecLev)  -> true ; RecLev=0 ),
    rx(Term,Fa,Dbg,Cache,RecLev),
    Fa=Fa0.  % instantiate only after the computation. Otherwise
             % errors may be raised if you provide an instantiated Fa
             % which does not unify with the result.

%% ignoring debugging and caching, this reduces to:
%% rx(Term0,Fa,_,_,_) :-
%%    (	  fsa_regex_aux:macro(Term0,Term)
%%    ->  rx(Term,Fa)
%%    ;   fsa_regex_aux:rx(Term0,Fa,0,off,0)
%%    ->  true
%%    ;   macro(Term0,Term)
%%    ->  rx(Term,Fa,0,off,0)
%%    ;   rx0(Term0,Fa)
%%    ->  true
%%    ;	  treat_atomic(Term0,Fa)
%%    ).
%% yap:  48 ->  37
%% sic:  62 ->  55
%% swi: 125 -> 108

rx(Term0,Fa,Dbg,Cache,RecLev0):-
#sic# %  (   current_predicate('$tk_do_one_event',
#sic# %			       tcltk:'$tk_do_one_event'(_,_))
#sic# %  ->  fsa_tk:tk_do_all_events 
#sic# %  ;   write(x), nl
#sic# %  ),
    RecLev is RecLev0,   %% eval Term+1+1..
    (	functor(Term0,Func,Ar)
    ->  true
    ;   user:throw(fsa_error("variable in regular expression~n",[]))
    ),
    debug_rx_start(Dbg,Term0,Func,Ar,RecLev,Time0),
    (   %% case 1: result is in cache
	in_cache(Cache,Func,Term0,Fa) 
    ->	(   Dbg >= 3 ->  format(user_error,"(cached: ~w)~n",[Term0]) ; true )
    ;   %% case 2: there is a user macro for it
	%% user:current_predicate(macro,fsa_regex_aux:macro(_,_)),
	fsa_regex_aux:macro(Term0,Term)
    ->  rx(Term,Fa,Dbg,Cache,RecLev+1),
	add_cache(Cache,Func,Term0,Fa)
    ;	%% case 3: there is a user rule for it
	%% it would be better if we knew whether there were a
	%% definition for _this_ operator
	%% user:current_predicate(rx,fsa_regex_aux:rx(_,_)),
	user:bb_put(fsa_regex:regex_rec,RecLev), %% save it
	fsa_regex_aux:rx(Term0,Fa)
    ->	add_cache(Cache,Func,Term0,Fa)
    ;	%% case 4: there is a built-in macro for it
	macro(Term0,Term)
    ->  rx(Term,Fa,Dbg,Cache,RecLev+1),
	add_cache(Cache,Func,Term0,Fa)
    ;	%% case 5: there is a built-in rule for it
	rx0(Term0,Fa,Dbg,Cache,RecLev+1)
    ->	add_cache(Cache,Func,Term0,Fa)
    ;   %% case 6: remaining atomic expressions represent themselves
	atomic(Term0)
    ->	treat_atomic(Term0,Fa)
    ;	%% case 7: nothing fits: raise an error
#swi# 	user:throw(fsa_error("Error in regular expression:~n~@~n",
#swi# 	  [write_term(Term0,[max_depth(6)])]))
#sic# 	user:throw(fsa_error("Error in regular expression:~n~@~n",
#sic# 	  [write_term(Term0,[max_depth(6)])]))
#yap#   user:throw(fsa_error("Error in regular expression:~n~w~n",
#yap#     [Term0]))
    ),
    debug_rx_end(Dbg,Func,Ar,RecLev,Time0,Fa).

treat_atomic(Atomic,Fa):-
    fsa_globals:fsa_global_get(pred_module,Module),
    fsa_pred_module:regex_notation_to_predicate(Module,Atomic,Type),
    fsa_data:fsa_components(r(Module),2,[0],[1],[trans(0,Type,1)],[],Fa).

%% set spy-points on certain operators
:- user:bb_put(fsa_rx_spy:spy,on).

%% has effect only for `selective' value of regex_cache variable.
:- user:bb_put(fsa_regex_cache:cache,on).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% OPERATOR DEFINITIONS %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

user:help_info(class,regex,"Regular Expressions",
"Regular expressions are the preferred way to specify regular languages and\
 \n regular relations. The regular expression compiler compiles a regular\
 \n expression into an equivalent finite-state automaton. \
 \n \
 \n The syntax of regular expressions in FSA is somewhat non-standard. As usual, \
 \n atomic symbols normally represent themselves. Concatenation is indicated\
 \n using a Prolog list where each of the elements is a regular expression\
 \n itself; union is indicated using curly (`set') brackets, and Kleene\
 \n star uses the *-suffix. For instance:\
 \n \
 \n     [{a,b,c}*,b,{a,b,c}*]\
 \n \
 \n is the set of strings over the alphabet {a,b,c} such that each string\
 \n contains at least one `b'.  Since a list indicates concatenation, the\
 \n empty list `[]' indicates the empty concatenation, i.e. the empty\
 \n string (the language consisting of a single string which is the\
 \n epsilon string). Furthermore, `{}' represents the empty language. \
 \n \
 \n Such regular expressions define regular languages, but regular\
 \n expressions can also be used to define weighted regular languages,\
 \n regular relations and weighted regular relations.  A regular relation\
 \n (transducer) is specified for instance by:\
 \n \
 \n     [{a:a,b:b,c:c}*,b:a,{a:a,b:b,c:c}*]\
 \n \
 \n Where x:y represents a mapping from symbol x to symbol y.  Weights can\
 \n be attached using a double semi-colon:\
 \n \
 \n     {a::0,b::1,c::2}*\
 \n \
 \n This extends to weighted transducers. As an example consider:\
 \n \
 \n     [{a:a::0,b:b::1,c:c::2}*,b:a::0,{a:a::2,b:b::1,c:c::0}*]\
 \n \
 \n btw. this is equivalent to:\
 \n \
 \n     [{a:a:0,b:b:1,c:c:2}*,b:a:0,{a:a:0,b:b:1,c:c:2}*]\
 \n \
 \n Below we provide more detailed documentation on the regular expression\
 \n syntax, the type coercion performed implicitly by the regular\
 \n expression compiler, ways to debug regular expressions, the way in which\
 \n regular expression operators can be added,  and detailed documentation\
 \n on each of the regular expression operators.").

user:help_info(regex,syntax,"Regular expression syntax",
"Regular expressions are defined as Prolog terms, and\
 \n therefore Prolog syntax applies. For detailed information on this,\
 \n cf. the Prolog manual. The brackets () can always be used to express\
 \n the desired grouping. The order of precedence of operators is as\
 \n follows: \
 \n \
 \n           : /         \
 \n           ..\
 \n           + * ^\
 \n           & -         \
 \n           o x xx      \
 \n           ! #\
 \n \
 \n Operators with the same precedence are interpreted left-to-right. For\
 \n example, the expression\
 \n \
 \n     a..z* - b* & c..d*\
 \n \
 \n is interpreted as:\
 \n \
 \n     (((a..z)*) - (b*)) & ((c..d)*)\
 \n \
 \n \
 \n Syntax *restrictions*\
 \n \
 \n These are all due to the use of Prolog syntax. The benefit of using\
 \n Prolog syntax is that I don't need to implement a parser, and you have\
 \n flexibility (by using your own operator definitions). However, a few\
 \n limitations are inherited as well. Here are a few rules of thumb:\
 \n \
 \n Capitals can be used in a regular expression in the Tk entry field, by\
 \n putting them between quotes:\
 \n \
 \n     'A'..'Z'\
 \n \
 \n At the Regex prompt (after fsa -r) you can use:\
 \n \
 \n     'A'..'Z'\
 \n \
 \n At the command-interpreter you can use:\
 \n \
 \n          2 |: -r 'A'..'Z'\
 \n          \
 \n As part of Expr in the fsa -r Expr command, use (this depends on the\
 \n shell you are using. This example works for bourne sh, csh, and bash):\
 \n \
 \n     fsa -r \"'A'..'Z'\"\
 \n \
 \n Use space between operators. Use space before and after a question\
 \n mark (?). Don't use the dot '.' or the vertical bar '|' as (part of) a\
 \n new operator. Similarly, avoid using the comma ',', the ';', and '->'\
 \n as (part of a) regular expression operator. It's neither a good idea\
 \n to use ':-'. Operators can be escaped using (), but hardly ever have\
 \n to (e.g. the following works, even if o is the binary composition\
 \n operator!).\
 \n \
 \n     fsa -r 'o o o'\
 \n \
 \n Brackets can be used for grouping as well:\
 \n \
 \n     fsa -r 'o o o o (o o o)'").

user:help_info(regex,type_coercion,"Type Coercion",
"FSA6 supports four types of automaton:\
 \n \
 \n          * recognizers\
 \n          * weighted recognizers (aka string-to-weight transducers)\
 \n          * transducers (aka string-to-string transducers)\
 \n          * weighted transducers (aka string-to-string-weight transducers)\
 \n \
 \n If automata of different types are combined by a binary regular expression\
 \n operator such as the concatenation operator, then the types of the automata\
 \n are compared, and if neccessary the automata are coerced silently. A\
 \n similar mechanism is used for operations which require a specific type.\
 \n \
 \n Coercion is performed according to the following type hierarchy, where\
 \n coercion is possible in upward direction, using the operator indicated\
 \n within brackets.\
 \n \
 \n \
 \n                 WEIGHTED-                      \
 \n                 TRANSDUCER                     \
 \n                 /       \\                     \
 \n                /          \\                   \
 \n               /             \\                 \
 \n         (identity)       (zero_weights)        \
 \n           /                    \\              \
 \n          /                       \\            \
 \n      WEIGHTED-                 TRANSDUCER      \
 \n      RECOGNIZER                                \
 \n          \\                       /            \
 \n            \\                   /              \
 \n              \\               /                \
 \n        (zero_weights)     (identity)           \
 \n                 \\         /                   \
 \n                   \\     /                     \
 \n                RECOGNIZER                      \
 \n \
 \n Consider the following examples:\
 \n \
 \n     expression:            coerced into:\
 \n \
 \n     { a, b:c     }          { a:a,   b:c   }\
 \n     { a, b::4    }          { a::0,  b::4  }\
 \n     { a, b:c:4   }          { a:a:0, b:c:4 }\
 \n     { a::0,a:b:4 }          { a:a:0, a:b:4 }\
 \n     { a:b, a:b:4 }          { a:b:0, a:b:4 }\
 \n \
 \n Some operators which expect a recognizer will temporarily *freeze*\
 \n their argument automaton if it is of a different type. In that case,\
 \n the symbol pairs or triples are treated as atomic symbols. This may or\
 \n may not be what you want. The following operators work in this way:\
 \n \
 \n complete\
 \n minimize\
 \n determinize\
 \n complement").

user:help_info(regex,spy_points,"Spy Points on Regular expressions",
"The regular expression compiler provides detailed information on the\
 \n computation time and the size of the resulting automata for certain\
 \n regular expression operators, namely for those operators Op for which\
 \n the predicate  \
 \n \
 \n         bb_get(fsa_rx_spy:Op,on).\
 \n \
 \n succeeds. So you can set a spy-point to operator concat by the\
 \n directive: \
 \n \
 \n         ?- bb_put(fsa_rx_spy:concat,on).\
 \n \
 \n The special operator spy(Expr) is equivalent to Expr except that it\
 \n has an associated spy-point.").

user:help_info(regex,extensions,"Extending the regular expression notation",
"Using the -aux[file] command line option, or the AuxFile button of the\
 \n TK Widget, you can load *auxiliary* regular expression operators. The\
 \n file should be a Prolog source file. It will be\
 \n loaded into module *fsa_regex_aux*. The syntax of regular expressions\
 \n can be used in this file (in fact it *must* be used, beware if the\
 \n file also contains ordinary Prolog code!).\
 \n \
 \n Two relations are important:\
 \n   1. macro/2\
 \n   2. rx/2\
 \n \
 \n The first relation is usually defined by unit clauses. It simply states\
 \n that the regular expression in the first argument is an abbreviation\
 \n for the regular expression in the second argument. For example:\
 \n \
 \n          macro(vowels,{a,e,i,u,o}).\
 \n \
 \n Such macro's can be parameterized using Prolog variables; e.g.:\
 \n \
 \n          macro(brz(Expr),determinize(reverse(determinize(reverse(Expr))).\
 \n \
 \n The relation rx/2 can be used for more complicated\
 \n operations (operations that are cumbersome or impossible to define in\
 \n terms of simpler regular expression operators). It defines a relation\
 \n between the regular expression in the first argument and the finite\
 \n automaton in the second argument. It is often useful to be able to\
 \n call the regular expression compiler recursively. This should be done\
 \n through the predicate fsa_regex:rx/2. The following is\
 \n equivalent to the first example of macro/2 above:\
 \n \
 \n          rx( vowels, Fa) :-     \
 \n              fsa_regex:rx({a,e,i,u,o}, Fa).\
 \n \
 \n Consult the Examples directory, for instance in the MohriSproat96,\
 \n Karttunen95, Karttunen96, Karttunen98, GerdemannVannoord, Queens\
 \n directories, for some extensive illustrations.").

user:help_info(regex,several_aux_files,
"Combining several auxiliary regular expression operator files",
"Suppose you want to use the definition of a replace operator in some\
 \n file *replace.pl* in your analysis of Dutch phonology. In the \
 \n latter file you can include the definitions from replace.pl by\
 \n including somewhere at the top of your file the following\
 \n directives:\
 \n \
 \n         :- ensure_loaded(replace).    %% loads replace.pl\
 \n \
 \n         :- multifile macro/2.\
 \n         :- multifile rx/2.\
 \n \
 \n This only works, if the multifile declarations are also present in the \
 \n file you are importing. I.e. in this example the file replace.pl\
 \n should also have the directives\
 \n \
 \n         :- multifile macro/2.\
 \n         :- multifile rx/2.").


user:help_info(class,operator,"Regular Expression Operators",
"This section lists the regular expression operators built-in in FSA.").

user:help_info(operator,Key,Usage,Expl):-
    fsa_regex:op_help(Key,Usage,Expl).

op_help(any_symbol,"?",
"The set of one-symbol strings over the universal alphabet, ie. ?\
 \n can be read as `any symbol whatsoever'. It uses the true/1\
 \n predicate_declaration from the current predicate module (cf the\
 \n chapter Predicates on Symbols). If that\
 \n declaration is not defined then no compilation for this operator is\
 \n possible, and an error occurs. "). 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% the infamous ?/0 %%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rx0(? , Fa, _, _,_):-
    fsa_globals:fsa_global_get(pred_module,Module),
    fsa_pred_module:true(Module,Pred),
    fsa_data:fsa_components(r(Module),2,[0],[1],[trans(0,Pred,1)],[],Fa).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%% DETERMINIZATION MINIMIZATION %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% special cases to ignore *some* peculiarities of Prolog syntax
%% which are frequently encountered

macro('?*',kleene_star(?)).
macro('?+',kleene_plus(?)).
macro('?^',kleene_option(?)).
macro('?:?',pair(?,?)).
macro('?:?*',kleene_star(pair(?,?))).
macro('?:?+',kleene_plus(pair(?,?))).
macro('?:?^',option(pair(?,?))).
macro('?-'(X),difference(?,X)).


op_help(minimize,"\
 \n Expr#   m[inimize](Expr)   mh(Expr)   mb(Expr)",
"Applies minimization to the result of compiling Expr. There are a\
 \n number of related expressions depending on which minimization\
 \n algorithm is to be used.\
 \n \
 \n *mb* uses the algorithm due to Brzozowski, *mh* uses the algorithm by\
 \n Hopcroft (as described in Aho, Hopcroft and Ullman, 1974). \
 \n \
 \n If Expr is a transducer then it is temporarily treated as a \
 \n recognizer over pairs of symbols (using the fsa_frozen predicate\
 \n module)."). 

macro(m(L),       mh(L)).
macro(minimize(L),mh(L)).
macro('#'(L),     mh(L)).
macro(mb(L),
      melt(cleanup(determinize0(reverse(determinize0(reverse(freeze(L,Flag)))))),Flag)).
macro(mh(L),
      melt(cleanup(mh0(coaccessible(determinize0(freeze(L,Flag))))),Flag)).

%% todo: document
%% todo: what must Ex be like? deterministic already? in what sense?
%%       at least co-accessible, I guess
rx0(m_pairs(Ex),Fa,D,C,L) :-
    rx(Ex,Fa0,D,C,L),
    fsa_data:fsa_type(Fa0,Type),
    m_pairs(Type,Fa0,Fa,D,C,L).

m_pairs(recognizer,Fa0,Fa,D,C,L) :-
    rx(m(fa(Fa0)),Fa,D,C,L).
m_pairs(w_recognizer,Fa0,Fa,D,_C,_L) :-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa1,Type,r(pair(Type))),
    ltime(D,fsa_minimizer:minimize_hopcroft(Fa1,Fa2)),
    fsa_data:fsa_copy_except(symbols,Fa2,Fa,r(pair(Type)),Type).

m_pairs(transducer(Sub),Fa0,Fa,D,_C,_L) :-
    to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    fsa_data:fsa_copy_except(symbols,Fa1,Fa2,Type,r(pair(Type))),
    ltime(D,fsa_minimizer:minimize_hopcroft(Fa2,Fa3)),
    fsa_data:fsa_copy_except(symbols,Fa3,Fa,r(pair(Type)),Type).

m_pairs(w_transducer(Sub),Fa0,Fa,D,_C,_L) :-
    w_to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    fsa_data:fsa_copy_except(symbols,Fa1,Fa2,Type,r(pair(Type))),
    ltime(D,fsa_minimizer:minimize_hopcroft(Fa2,Fa3)),
    fsa_data:fsa_copy_except(symbols,Fa3,Fa,r(pair(Type)),Type).

%% note: coaccessible is technically required; but almost always
%% already coaccessible anyway...
%% result: minimal, deterministic, accessible, co-accessible
rx0(mh0(A),Fa,Debug,Cache,Level) :-
    rx(A,FaA,Debug,Cache,Level),
    ltime(Debug,fsa_minimizer:minimize_hopcroft(FaA,Fa)).

op_help(determinize,"A!   determinize(A)   determinize(A,Algorithm)",

"Set of strings denoted by A, but moreover the subset construction\
 \n determinization algorithm is applied to ensure that the automaton is\
 \n deterministic. The algorithm can be specified as the second argument. \
 \n There are several variants\
 \n of the algorithm, which are different with respect to the treatment\
 \n of epsilon transitions:\
 \n \
 \n * per_graph: first construct efree automaton (jumps taken into\
 \n   account on target side of transitions and on start states)\
 \n * per_inverted_graph: first construct efree automaton (jumps taken\
 \n   into account on source side of transitions and and final states)\
 \n * per_reachable_graph: as per_inverted_graph, but maintains accessibility\
 \n * per_co_reachable_graph: as per_graph, but maintains co-accessibility\
 \n * per_subset: compute transitive closure of jumps on the fly for each subset\
 \n * per_state: compute transitive closure of jumps on the fly for each state\
 \n \
 \n These variants and some interesting experimental observations are\
 \n described in a paper I presented at the FSMNLP 98 workshop in Ankara. \
 \n The paper is available from http://www.let.rug.nl/~vannoord/papers/.\
 \n An improved version of the paper has been published in Computational\
 \n Linguistics.\
 \n \
 \n By default the algorithm is chosen by a simple heuristic based on the\
 \n number of states and number of jumps of the input automaton. If A is a\
 \n transducer then it is temporarily treated as a recognizer over atomic\
 \n symbols, which happen to be pairs of predicates.").

macro(determinize(L),melt(determinize0(freeze(L,Flag)),Flag)).
macro('!'(L),        melt(determinize0(freeze(L,Flag)),Flag)).

rx0(determinize0(A),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    (	ftime(Debug,fsa_determinizer:fsa_deterministic(Fa0))
    ->	reachable_fa(Fa0,Fa,Debug)
    ;	fsa_data:fsa_states_number(Fa0,States),
	fsa_data:fsa_jumps(Fa0,Js0), length(Js0,Js),
	(   Js > States
	->  ltime(Debug,fsa_determinizer:determinize(Fa0,Fa,per_state))
	;   ltime(Debug,fsa_determinizer:determinize(Fa0,Fa,per_graph))
	)
    ).

macro(determinize(L,Impl),melt(determinize0(freeze(L,Flag),Impl),Flag)).
rx0(determinize0(A,Impl),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    (	ftime(Debug,fsa_determinizer:fsa_deterministic(Fa0))
    ->	reachable_fa(Fa0,Fa,Debug)
    ;   ltime(Debug,fsa_determinizer:determinize(Fa0,Fa,Impl))
    ).

op_help(efree,"efree(E)   reachable_efree(E)   co_reachable_efree(E)",
"Constructs epsilon-free automaton for the automaton created for\
 \n E. The first variant is faster, the second and third algorithms yield\
 \n smaller automata by only taking into account states reachable from the\
 \n start state, resp. from which a final state is reachable.").

rx0(efree(X),Fa,Debug,Cache,Level) :-
    rx(X,FaA,Debug,Cache,Level),
    ltime(Debug,fsa_determinizer:efree(FaA,Fa)).

rx0(reachable_efree(X),Fa,Debug,Cache,Level) :-
    rx(X,FaA,Debug,Cache,Level),
    ltime(Debug,fsa_determinizer:reachable_efree(FaA,Fa)).

rx0(co_reachable_efree(X),Fa,Debug,Cache,Level) :-
    rx(X,FaA,Debug,Cache,Level),
    ltime(Debug,fsa_determinizer:co_reachable_efree(FaA,Fa)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% COMPLEMENT / DIFFERENCE %%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(term_complement,"term_complement(E)   `E",
"The term complement of E, i.e. the set of all single symbol strings \
 \n minus those in E; this is equivalent to ? - E.").
macro(term_complement(A),   '-'(?,A)).
macro('`'(A),               '-'(?,A)).

op_help(complement,"~E   complement(E)",
"The complement of the language denoted by E. E must be a recognizer.").

%% next 2 are obsolete:
macro(local_complement(A),  swap_final_non_final(complete(determinize(A)))).
macro(global_complement(A), swap_final_non_final(complete(determinize(A)))).

macro(complement(A),        swap_final_non_final(complete(determinize(A)))).
macro('~'(A),               swap_final_non_final(complete(determinize(A)))).

%% undocumented: only used internally (?)
rx0(swap_final_non_final(A),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_copy_except(final_states,Fa0,Fa,Finals0,Finals),
    fsa_data:fsa_states_set(Fa0,States),
    ord_subtract(States,Finals0,Finals).

op_help(difference,"A-B   difference(A,B)",
"Set of strings denoted by A minus those given by B. A and B must be\
 \n recognizers."). 

macro(difference(A,B),      intersection(A,complement(B))).
macro('-'(A,B),             intersection(A,complement(B))).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%% CONTAINMENT %%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(containment,"$E   containment(E)",
"The language consisting of all strings that have an instance of E as a\
 \n sub-string: [? *, E, ? *]. Note that the result is a minimal\
 \n automaton. Since the definition of this operator depends on the ?/0\
 \n operator it is only defined if the current predicate module provides\
 \n a definition of true/1. E can be both a recognizer or a transducer.").  
macro(containment(Expr),[kleene_star(?),Expr,kleene_star(?)]).
macro('$'(Expr),        [kleene_star(?),Expr,kleene_star(?)]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% T-DETERMINIZATION / T-MINIMIZATION %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(t_determinize,"t_determinize(E)",
"The set of pairs denoted by E, but moreover the determinization\
 \n algorithm for transducers by Mohri, cf. also Roche and Schabes, is\
 \n applied to E. NB: this is only guaranteed to terminate if in fact E\
 \n can be determinized in the appropriate sense. The implementation\
 \n currently does not check for this. Refer to the Examples/RocheSchabes97\
 \n directory for an experimental implementation of that check and various\
 \n related algorithms. \
 \n \
 \n *representation* of sequential transducers\
 \n \
 \n Note that in FSA subsequential transducers are represented as ordinary \
 \n transducers. This implies in particular that instead of output symbols \
 \n associated with final states, we have a separate transition over\
 \n epsilon input and final output to a new final state. Similarly,\
 \n automata which require an initial output to be associated with the\
 \n start state will give rise to an extra transition from a new start \
 \n state with epsilon input and the required start output.\
 \n \
 \n The treatment of identity constraints over predicate pairs is\
 \n especially tricky. This is mostly hidden in the predicate module\
 \n declaration of t_determinize/2 preds. Funny things to watch for: \
 \n \
 \n * certain non-functional automata become t_deterministic:\
 \n \
 \n     t_minimize([ a x {b,d}, d*])\
 \n \
 \n This is actually quite useful.\
 \n \
 \n * delayed identity constraints: predicates apply on the input side\
 \n before the transition containing the target of the identity constraint \
 \n is encountered. For example:\
 \n \
 \n         t_minimize({[a:b,?,?,?,?,?,b],[a:c,?,?,?,?,?,c]})\
 \n \
 \n This is especially nasty if the number of question marks do not match\
 \n up: \
 \n \
 \n         t_minimize({[a:b,?,?,?,?,?,b],[a:c,?,?,?,c]})\
 \n \
 \n * the opposite occurs as well: sometimes we have to output symbols\
 \n satisfying a certain predicate which must be identical to an input\
 \n symbol which is yet to be encountered! This currently works. Try for\
 \n instance: \
 \n \
 \n         t_minimize([a:b,a..f])\
 \n \
 \n I think this is quite spectacular."). 
macro(t_determinize(A),
      t_determinize0(sync_sequence_transducer(transducer(A,warn)))).
macro(tdeterminize(A),
      t_determinize0(sync_sequence_transducer(transducer(A,warn)))).

rx0(t_determinize0(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_t_determinizer:t_determinize(fsa_strings,Fa0,Fa)).

op_help(t_minimize,"t_minimize(E)", 
"The set of pairs denoted by E, but moreover the minimization\
 \n algorithm for transducers by Morhi is applied to E. Note that this\
 \n uses the t_determinize operator - for further details check there. ").

%% perhaps have cleanup apply to t_determinize?
%% for cases like {a x[a,c,c],[?x a,[]:c,[]:c]}
%%%%%% macro(tm(A),  m(t_minimize0(t_determinize(A)))).
%% this is not minimal; for instance:
%% 0 --> id({a,b})/[id({a,b})]  --> 1
%% 1 --> id({a,b})/[id({a,b})]  --> 1
%% 1 --> id(`{a,b})/[id(`{a,b})] --> 0
%% 0 --> id(`{a,b})/[id(`{a,b})] --> 0
%% 0,1 both final. 1 start.
%%
%% as you get from t_determinize(replace([a..b,c..d]))

macro(tminimize(A),   t_minimize(A)).

macro(ordinary_t_minimize(A),m(t_minimize0(t_determinize(A)))).

rx0(t_minimize(A),Fa,D,C,L) :-
    rx(A,Fa0,D,C,L),
    fsa_data:fsa_symbols_decl(Fa0,Decl),
    t_min(Decl,A,Fa0,Fa,D,C,L).

t_min(r(_),A,Fa,Fa,_D,_C,_L) :-
    warning(error,transducer,A).
t_min(r(_,_),A,Fa,Fa,_D,_C,_L) :-
    warning(error,transducer,A).
t_min(t(_,_,_),A,Fa,Fa,_D,_C,_L) :-
    warning(error,transducer,A).
t_min(t(fsa_frozen,fsa_frozen),_,Fa0,Fa,D,C,L):-
    !,
    rx(ordinary_t_minimize(fa(Fa0)),Fa,D,C,L).
t_min(t(_,_),_A,Fa0,Fa,D,C,L) :-
    rx(ordinary_t_minimize(
          cleanup(               % get some predicates
	   m_pairs(              % expanded out, and get machine
	    letter_transducer(   % smaller
	     ordinary_t_minimize(
              cleanup(           % for example tdm5!
               fa(Fa0))))))),Fa,D,C,L).

rx0(t_minimize0(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_t_determinizer:t_minimize(fsa_strings,Fa0,Fa)).

op_help(w_determinize,"w_determinize(E)",
"The set of pairs denoted by E, but moreover the determinization\
 \n algorithm for string to weight transducers by Mohri is applied to\
 \n E. Cf. the t_determinize operator for further details."). 
macro(w_determinize(A), w_determinize0(weighted_recognizer(A,error))).

rx0(w_determinize0(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_t_determinizer:t_determinize(fsa_weights,Fa0,Fa)).

op_help(wt_determinize,"wt_determinize(E)",
"Denotes the weighted transducer given by E, but moreover the\
 \n determinization algorithm for weighted transducers by Mohri is applied\
 \n to E. Cf. the t_determinize operator for further details."). 

macro(wt_determinize(A), wt_determinize0(weighted_sequence_transducer(A))).

rx0(wt_determinize0(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_t_determinizer:t_determinize(fsa_weighted_strings,Fa0,Fa)).

op_help(w_minimize,"w_minimize(E)",
"The set of pairs denoted by E, but moreover the minimization algorithm\
 \n for string to weight transducers by Mohri is applied to E. Thus, E\
 \n must denote a weighted recognizer. Cf. the t_determinize operator\
 \n for further details."). 
macro(w_minimize(A), cleanup(m(w_minimize0(w_determinize(A))))).

rx0(w_minimize0(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_t_determinizer:t_minimize(fsa_weights,Fa0,Fa)).

op_help(wt_minimize,"wt_minimize(E)",
"Denotes the weighted transducer given by E, but moreover the\
 \n minimization algorithm for string to weight transducers by Mohri is\
 \n applied to E. Cf. the t_minimize operator for further details. ").
macro(ordinary_wt_minimize(A), m(wt_minimize0(wt_determinize(A)))).

rx0(wt_minimize0(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_t_determinizer:t_minimize(fsa_weighted_strings,Fa0,Fa)).

rx0(wt_minimize(A),Fa,D,C,L) :-
    rx(A,Fa0,D,C,L),
    fsa_data:fsa_symbols_decl(Fa0,Decl),
    wt_min(Decl,A,Fa0,Fa,D,C,L).

wt_min(r(_),A,Fa,Fa,_D,_C,_L) :-
    warning(error,weighted_transducer,A).
wt_min(r(_,_),A,Fa,Fa,_D,_C,_L) :-
    warning(error,weighted_transducer,A).
wt_min(t(_,_),A,Fa,Fa,_D,_C,_L) :-
    warning(error,weighted_transducer,A).
wt_min(t(fsa_frozen,fsa_frozen,_),_,Fa0,Fa,D,C,L):-
    !,
    rx(ordinary_wt_minimize(fa(Fa0)),Fa,D,C,L).
wt_min(t(_,_,_),_A,Fa0,Fa,D,C,L) :-
    rx(ordinary_wt_minimize(
          cleanup(               
	   m_pairs(              
	    weighted_transducer( 
	     ordinary_wt_minimize(
              cleanup(           
               fa(Fa0))))))),Fa,D,C,L).

%% Brzozwski
macro(wb_minimize(A), w_determinize(reverse(w_determinize(reverse(A))))).

%% Brzozwski
macro(tb_minimize(A), t_determinize(reverse(t_determinize(reverse(A))))).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% dictionairies %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(words,"words(ListOfAtoms)",
"Creates minimal automaton for each of the atoms in ListOfAtoms, where\
 \n each atom is expanded out into a concatenation of the atoms corresponding\
 \n to its characters; i.e.\
 \n \
 \n        words([john,peter,mary])\
 \n \
 \n is equivalent to\
 \n \
 \n        { [j,o,h,n],[p,e,t,e,r],[m,a,r,y] }").

rx0(words(List),Fa,_,_,_) :-
    list_atom_chars(List,CharList),
    fsa_dict:fsa_dict_to_fsa(CharList,Fa).

list_atom_chars([],[]).
list_atom_chars([A|T0],[C|T]) :-
    atom_chars(A,C1),
    list_atom_chars_(C1,C),
    list_atom_chars(T0,T).

list_atom_chars_([],[]).
list_atom_chars_([C|Cs],[A|As]) :-
    atom_chars(A,[C]),
    list_atom_chars_(Cs,As).

op_help(perfect_hash,"perfect_hash(ListOfAtoms)",
"Creates weighted recognizer implementing a (minimal) perfect hash \
 \n for the words found in ListOfAtoms. For instance:\
 \n \
 \n        perfect_hash([john,peter,mary])\
 \n \
 \n is equivalent to\
 \n \
 \n        w_minimize( {[j,o,h,n] :: 0,\
 \n         [p,e,t,e,r] :: 2,\
 \n         [m,a,r,y] :: 1 } )").
rx0(perfect_hash(List),Fa,_,_,_) :- 
    list_atom_chars(List,CharList),
    fsa_dict:fsa_dict_to_perfect_hash(CharList,Fa).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%% CROSS PRODUCT %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

macro(triple(A,B,C),cross_product0(A,B,C)).
macro(':'(A,':'(B,C)),L) :-
    !, L = triple(A,B,C).
macro(':'(A,'::'(B,C)),L) :-
    !, L = triple(A,B,C).
macro(':'(':'(A,B),C),L) :-
    !, L = triple(A,B,C).

op_help(pair,"A:B   pair(A,B)   A:B:Weight",
"A and B are symbols; this is a transducer mapping an A to a B. In\
 \n addition, A:B:W defines a weighted transducer mapping A to B with\
 \n associated weight W."). 
macro(pair(A,B),cross_product0(A,B)).
macro(':'(A,B), cross_product0(A,B)).

op_help(symbol_weight_pair,"A::W",
"Defines a weighted recognizer where A is a recognizer, and W its weight;\
 \n alternatively defines a weighted transducer where A is a transducer\
 \n and W its weight.").
macro('::'(':'(A,B),C),triple(A,B,C)).
%macro('::'(A,B),weighted_recognizer(cross_product(A,B),silent)).

rx0('::'(A,B),Fa,D,C,L) :-
    rx(A,Fa0,D,C,L),
    fsa_data:fsa_type(Fa0,Type),
    assign_weight(Type,Fa0,B,Fa,D,C,L).

assign_weight(recognizer,Fa0,B,Fa,D,C,L) :-
    !,
    rx(weighted_recognizer(cross_product(fa(Fa0),B),silent),Fa,D,C,L).
assign_weight(_,Fa0,B,Fa,D,C,L) :-
    rx([fa(Fa0),'::'([],B)],Fa,D,C,L).


op_help(cross_product,"A x B   cross_product(A,B)   A x B x W",
"The set of pairs (A0,B0) such that A0 is in A and B0 is in B.\
 \n Both A and B must describe recognizers.\
 \n \
 \n In addition, A x B x W defines a weighted transducer where strings from A\
 \n are mapped to B, with weight W.").  
macro(cross_product(A,B,C),
      m(cross_product0(recognizer(A),recognizer(B),recognizer(C)))).
macro(cross_product(A,B),
      m(cross_product0(recognizer(A),recognizer(B)))).
macro(x(x(A,B),C),L) :-
    !, L=m(cross_product0(recognizer(A),recognizer(B),recognizer(C))).
macro(x(A,x(B,C)),L) :-
    !, L=m(cross_product0(recognizer(A),recognizer(B),recognizer(C))).
macro(x(A,B),
      m(cross_product0(recognizer(A),recognizer(B)))).

rx0(cross_product0(A,B),Fa,Debug,Cache,Level) :-
    rx(A,FaA,Debug,Cache,Level),
    rx(B,FaB,Debug,Cache,Level),
    ltime(Debug,fsa_regex:cross_product_fa(FaA,FaB,Fa)).

rx0(cross_product0(A,B,C),Fa,Debug,Cache,Level) :-
    rx(A,FaA,Debug,Cache,Level),
    rx(B,FaB,Debug,Cache,Level),
    rx(expand_predicates(C),FaC,Debug,Cache,Level),
    ltime(Debug,fsa_regex:cross_product_fa(FaA,FaB,FaC,Fa)).

cross_product_fa(FaA,FaB,Fa):-
    all_components_reachable(cross_product,FaA,FaB,Fa).

cross_product_fa(FaA,FaB,FaC,Fa):-
    all_components_reachable(cross_product3,FaA,FaB,FaC,Fa).

prepare_reachable(cross_product,FaA,FaB,
		  d(FinalsA,FinalsB,TransA,TransB,JumpsA,JumpsB),
		  StartStates,t(SigmaD,SigmaR)):-
    fsa_data:fsa_symbols_decl(FaA,r(SigmaD)),
    fsa_data:fsa_symbols_decl(FaB,r(SigmaR)),
    pairs_of_start_states(FaA,FaB,StartStates),
    fsa_data:index_finals(FaA,FinalsA),
    fsa_data:index_finals(FaB,FinalsB),
    index_jumps_if_ne(FaA,JumpsA),
    index_jumps_if_ne(FaB,JumpsB),
    fsa_data:index_transitions(FaA,TransA),
    fsa_data:index_transitions(FaB,TransB).

prepare_reachable(cross_product3,FaA,FaB,FaC,
	d(FinalsA,FinalsB,FinalsC,TransA,TransB,TransC,JumpsA,JumpsB,JumpsC),
		  StartStates,t(SigmaD,SigmaR,SigmaW)):-
    fsa_data:fsa_symbols_decl(FaA,r(SigmaD)),
    fsa_data:fsa_symbols_decl(FaB,r(SigmaR)),
    fsa_data:fsa_symbols_decl(FaC,r(SigmaW)),
    triples_of_start_states(FaA,FaB,FaC,StartStates),
    fsa_data:index_finals(FaA,FinalsA),
    fsa_data:index_finals(FaB,FinalsB),
    fsa_data:index_finals(FaC,FinalsC),
    index_jumps_if_ne(FaA,JumpsA),
    index_jumps_if_ne(FaB,JumpsB),
    index_jumps_if_ne(FaC,JumpsC),
    fsa_data:index_transitions(FaA,TransA),
    fsa_data:index_transitions(FaB,TransB),
    fsa_data:index_transitions(FaC,TransC).

index_jumps_if_ne(Fa,JT):-
    fsa_data:fsa_jumps(Fa,Jumps),
    index_jumps_if_ne0(Jumps,Fa,JT).

index_jumps_if_ne0([],_,nojumps).
index_jumps_if_ne0([_|_],Fa,jumps(Table)):-
    fsa_data:index_jumps(Fa,Table).

pairs_of_start_states(FaA,FaB,Pairs):-
    fsa_data:fsa_start_states(FaA,StartsA),
    fsa_data:fsa_start_states(FaB,StartsB),
    pairs_of_start_states_l(StartsA,StartsB,Pairs).

pairs_of_start_states_l([],_,[]).
pairs_of_start_states_l([H|T],StartsB,Pairs):-
    pairs_of_start_state(StartsB,H,Pairs,Pairs0),
    pairs_of_start_states_l(T,StartsB,Pairs0).

pairs_of_start_state([],_,P,P).
pairs_of_start_state([H|T],A,[A/H|Pairs0],Pairs):-
    pairs_of_start_state(T,A,Pairs0,Pairs).

triples_of_start_states(FaA,FaB,FaC,Triples):-
    fsa_data:fsa_start_states(FaA,StartsA),
    fsa_data:fsa_start_states(FaB,StartsB),
    fsa_data:fsa_start_states(FaC,StartsC),
    triples_of_start_states_l(StartsA,StartsB,StartsC,Triples).

triples_of_start_states_l([],_,_,[]).
triples_of_start_states_l([H|T],StartsB,StartsC,Triples):-
    triples_of_start_state(StartsB,StartsC,H,Triples,Triples0),
    triples_of_start_states_l(T,StartsB,StartsC,Triples0).

triples_of_start_state([],_,_,P,P).
triples_of_start_state([H|T],StartsC,A,Triples0,Triples):-
    triples_of_start_st(StartsC,A,H,Triples0,Triples1),
    triples_of_start_state(T,StartsC,A,Triples1,Triples).

triples_of_start_st([],_,_,T,T).
triples_of_start_st([H|T],A,B,[t(A,B,H)|Tr0],Tr) :-
    triples_of_start_st(T,A,B,Tr0,Tr).

is_final_reachable(cross_product,d(FinalsA,FinalsB,_,_,_,_),A/B,Bool):-
    fsa_data:indexed_final(A,FinalsA,Bool1),
    fsa_data:indexed_final(B,FinalsB,Bool2),
    and(Bool1,Bool2,Bool).

is_final_reachable(cross_product3,d(FinalsA,FinalsB,FinalsC,_,_,_,_,_,_),t(A,B,C),Bool):-
    fsa_data:indexed_final(A,FinalsA,Bool1),
    fsa_data:indexed_final(B,FinalsB,Bool2),
    fsa_data:indexed_final(C,FinalsC,Bool3),
    and(Bool1,Bool2,Bool12),
    and(Bool12,Bool3,Bool).

and(yes,B,B).
and(no,_,no).

trans_reachables(cross_product,PA/PB,d(FA,FB,TsA,TsB,JA,JB),List):-
    findall(trans(A,B,C),trans_cross_product1(PA,PB,TsA,TsB,A,B,C),List,List1),
    findall(trans(A,B,C),trans_cross_product2(PA,PB,FB,TsA,A,B,C),List1,List2),
    findall(trans(A,B,C),trans_cross_product3(PA,PB,FA,TsB,A,B,C),List2,List3),
    findall(jump(A,B),trans_cross_product4(JA,PA,PB,A,B),List3,List4),
    findall(jump(A,B),trans_cross_product5(JB,PA,PB,A,B),List4,[]).

trans_cross_product1(PA,PB,TsA,TsB,PA/PB,SymA/SymB,QA/QB):-
    fsa_data:indexed_transition(PA,TsA,SymA,QA),
    fsa_data:indexed_transition(PB,TsB,SymB,QB).

trans_cross_product2(PA,PB,FB,TsA,PA/PB,SymA/[],QA/PB):-
    fsa_data:indexed_final(PB,FB,yes),
    fsa_data:indexed_transition(PA,TsA,SymA,QA).

trans_cross_product3(PA,PB,FA,TsB,PA/PB,[]/SymB,PA/QB):-
    fsa_data:indexed_final(PA,FA,yes),
    fsa_data:indexed_transition(PB,TsB,SymB,QB).

trans_cross_product4(jumps(JA),PA,PB,PA/PB,QA/PB):-
    fsa_data:indexed_jump(PA,JA,QA).

trans_cross_product5(jumps(JB),PA,PB,PA/PB,PA/QB):-
    fsa_data:indexed_jump(PB,JB,QB).


trans_reachables(cross_product3,t(Pa,Pb,Pc),d(Fa,Fb,Fc,Tsa,Tsb,Tsc,Ja,Jb,Jc),List):-
    findall(Trans,tcp3(Pa,Pb,Pc,Fa,Fb,Fc,Tsa,Tsb,Tsc,Ja,Jb,Jc,Trans),List).

tcp3(Pa,Pb,Pc,Fa,Fb,Fc,Tsa,Tsb,Tsc,_,_,_,trans(t(Pa,Pb,Pc),Sa/(Sb/Sc),t(Pa1,Pb1,Pc1))):-
    tcp3_transition(Pa,Fa,Tsa,Sa,[],Pa1),
    tcp3_transition(Pb,Fb,Tsb,Sb,[],Pb1),
    tcp3_transition(Pc,Fc,Tsc,Sc,0,Pc1),
    \+ ( Sa = [], Sb = [], Sc = 0 ).

tcp3(Pa,Pb,Pc,_,_,_,_,_,_,Ja,_,_,jump(t(Pa,Pb,Pc),t(Pa1,Pb,Pc))):-
    fsa_data:indexed_jump(Pa,Ja,Pa1).
tcp3(Pa,Pb,Pc,_,_,_,_,_,_,_,Jb,_,jump(t(Pa,Pb,Pc),t(Pa,Pb1,Pc))):-
    fsa_data:indexed_jump(Pb,Jb,Pb1).
tcp3(Pa,Pb,Pc,_,_,_,_,_,_,_,_,Jc,jump(t(Pa,Pb,Pc),t(Pa,Pb,Pc1))):-
    fsa_data:indexed_jump(Pc,Jc,Pc1).

tcp3_transition(P,_F,Ts,S,_Zero,Q) :-
    fsa_data:indexed_transition(P,Ts,S,Q).
tcp3_transition(P,F,_Ts,Zero,Zero,P) :-
    fsa_data:indexed_final(P,F,yes).


op_help(sl_cross_product,"A xx B   sl_cross_product(A,B)",
"The set of pairs (A0,B0) such that A0 is in A and B0 is in B, moreover\
 \n the strings A0 and B0 are to be of the same length. Both A and B must\
 \n be recognizers.").  
macro(sl_cross_product(A,B),m(sl_cross_product0(m(recognizer(A)),m(recognizer(B))))).
macro(xx(A,B),              m(sl_cross_product0(m(recognizer(A)),m(recognizer(B))))).


rx0(sl_cross_product0(A,B),Fa,Debug,Cache,Level) :-
    rx(A,FaA,Debug,Cache,Level),
    rx(B,FaB,Debug,Cache,Level),
    ltime(Debug,fsa_regex:sl_cross_product_fa(FaA,FaB,Fa)).

sl_cross_product_fa(FaA,FaB,Fa):-
    all_components_reachable(sl_cross_product,FaA,FaB,Fa).

prepare_reachable(sl_cross_product,FaA,FaB,
		  d(FinalsA,FinalsB,TransA,TransB),
		  StartStates,t(SigmaD,SigmaR)):-
    fsa_data:fsa_symbols_decl(FaA,r(SigmaD)),
    fsa_data:fsa_symbols_decl(FaB,r(SigmaR)),
    pairs_of_start_states(FaA,FaB,StartStates),
    fsa_data:index_finals(FaA,FinalsA),
    fsa_data:index_finals(FaB,FinalsB),
    fsa_data:index_transitions(FaA,TransA),
    fsa_data:index_transitions(FaB,TransB).

is_final_reachable(sl_cross_product,d(FinalsA,FinalsB,_,_),A/B,Bool):-
    fsa_data:indexed_final(A,FinalsA,Bool1),
    fsa_data:indexed_final(B,FinalsB,Bool2),
    and(Bool1,Bool2,Bool).

trans_reachables(sl_cross_product,PA/PB,d(_,_,TsA,TsB),Trans):-
    findall(trans(A,B,C),trans_sl_cross_product(PA,PB,TsA,TsB,A,B,C),Trans).

trans_sl_cross_product(PA,PB,TsA,TsB,PA/PB,SymA/SymB,QA/QB):-
    fsa_data:indexed_transition(PA,TsA,SymA,QA),
    fsa_data:indexed_transition(PB,TsB,SymB,QB).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% almost ATOMICS %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(escape,"escape(Sym)",
"Sym is a symbol. This denotes the language consisting of that\
 \n symbol. Can be used to overwrite special meaning of some symbols. For\
 \n instance, escape(?) can be used to denote a literal question mark.\
 \n Sym should be ground Prolog term, and it is passed through the\
 \n predicate *regex_notation_to_predicate* of the current predicate module."). 

rx0(escape(A),Fa,_,_,_):-
    fsa_globals:fsa_global_get(pred_module,Module),
    fsa_pred_module:regex_notation_to_predicate(Module,A,Type),
    fsa_data:fsa_components(r(Module),2,[0],[1],[trans(0,Type,1)],[],Fa).

op_help(interval,"S..T",
"S and T are one-character atoms or integers. In the first case, denotes\
 \n the set of symbols from S up to T in ASCII coding. For instance a..e is\
 \n equivalent to {a,b,c,d,e}. If S and T are integers, represents the set\
 \n of integers in that interval: for instance 8..11 is equivalent to\
 \n {8,9,10,11}."). 

macro(interval(X,Y),m(interval0(X,Y))).
macro('..'(X,Y),    m(interval0(X,Y))).
rx0(interval0(A,B),Fa,Debug,Cache,Level) :-
    (	atomic(A), atomic(B)
    ->	findall(Sym,between_interval(A,B,Sym),Symbols0),
	sort(Symbols0,Symbols),
	fsa_globals:fsa_global_get(pred_module,Module),
	str(Symbols,Trans,Module),
	fsa_data:fsa_components(r(Module),2,[0],[1],Trans,[],Fa)
    ;	pair_to_atom(A,A0),
	pair_to_atom(B,B0),
	rx(identity(interval(A0,B0)),Fa,Debug,Cache,Level)
    ).

str([],[],_).
str([Sym0|Syms],[trans(0,Sym,1)|Trans],Module):-
    fsa_pred_module:regex_notation_to_predicate(Module,Sym0,Sym),
    str(Syms,Trans,Module).

pair_to_atom(In,X):-
    (	var(In)
    ->  user:throw(fsa_error("variable in regular expression~n",[]))
    ;	atomic(In)
    ->	In=X
    ;	In = ':'(X,X)
    ->	true
    ;	user:throw(fsa_error("inappropriate symbol in interval: ~w~n",[In]))
    ).

between_interval(A,Z,X) :-
    (	atom(A), atom(Z)
    ->	atom_codes(A,[NoA]),
	atom_codes(Z,[NoZ]),
	fsa_util:between(NoA,NoZ,NoX),
	atom_codes(X,[NoX])
    ;	integer(A), integer(Z), A =< Z
    ->	fsa_util:between(A,Z,X)
    ;	user:throw(fsa_error("error in interval expression: ~w..~w~n",
				  [A,Z]))
    ).

%% obsolete
macro(class(List),set(List)).
macro(negated_class(List),term_complement(set(List))).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% COACCESSIBLE / ACCESSIBLE %%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% this should be called `coaccessible'
op_help(incomplete,"incomplete(A)",
"Ensures that all states in the automaton for A are co-accessible, i.e.\
 \n for each state there is path to a final state. ").
op_help(coaccessible,"coaccessible(A)",
"Ensures that all states in the automaton for A are co-accessible, i.e.\
 \n for each state there is path to a final state. ").

macro(incomplete(A),coaccessible(A)).
macro(co_reachable(A),coaccessible(A)).
macro(coreachable(A),coaccessible(A)).
macro(co_accessible(A),coaccessible(A)).

rx0(coaccessible(A),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    co_reachable_fa(Fa0,Fa,Debug).

co_reachable_fa(Fa0,Fa,Debug):-
    ltime(Debug,fsa_regex:co_reachable_fa(Fa0,Fa)).

co_reachable_fa(Fa0,Fa):-
    all_components_co_reachable(co_reachable,Fa0,Fa).

prepare_co_reachable(co_reachable,Fa,d(Starts,Table),Finals,Sigma):-
    fsa_data:fsa_final_states(Fa,Finals),
    fsa_data:fsa_symbols_decl(Fa,Sigma),
    fsa_data:index_starts(Fa,Starts),
    inverted_trans_jumps_table(Fa,Table).

inverted_trans_jumps_table(Fa,Table):-
    fsa_data:fsa_transitions(Fa,Trans),
    fsa_data:fsa_jumps(Fa,Jumps),
    fsa_data:fsa_states_number(Fa,Size),
    fsa_array:fsa_array_new(Table,Size),
    inverted_trans_to_keylist(Trans,Keylist0,Keylist1),
    inverted_jumps_to_keylist(Jumps,Keylist1,[]),
    keysort(Keylist0,Keylist),
    mult_keylist_to_thing(Keylist,Table).

inverted_trans_to_keylist([],T,T).
inverted_trans_to_keylist([trans(A,B,C)|Tail],[C-trans(A,B,C)|Trs0],Trs):-
    inverted_trans_to_keylist(Tail,Trs0,Trs).

inverted_jumps_to_keylist([],T,T).
inverted_jumps_to_keylist([jump(A,C)|Tail],[C-jump(A,C)|Trs0],Trs):-
    inverted_jumps_to_keylist(Tail,Trs0,Trs).

mult_keylist_to_thing([],_).
mult_keylist_to_thing([Key-Val|List],Thing):-
    fsa_array:fsa_array_access(Key,[Val|Vals],Thing),
    mult_keylist_to_thing(List,Key,Vals,Thing).

mult_keylist_to_thing([],_,[],_).
mult_keylist_to_thing([Key-Val|List],Key0,Vals0,Thing):-
    (	Key0==Key
    ->	Vals0=[Val|Vals]
    ;   Vals0=[],
	fsa_array:fsa_array_access(Key,[Val|Vals],Thing)
    ),
    mult_keylist_to_thing(List,Key,Vals,Thing).

trans_co_reachables(co_reachable,P,d(_,Trans),List):-
    fsa_array:fsa_array_access(P,List,[],Trans).

is_start_co_reachable(co_reachable,d(Starts,_),P,Bool):-
    fsa_data:indexed_start(P,Starts,Bool).

op_help(reachable,"reachable(A)",
"This operator ensures that for each state s in the automaton of A there\
 \n is a path from a start state to s.").

op_help(accessible,"accessible(A)",
"This operator ensures that for each state s in the automaton of A there\
 \n is a path from a start state to s.").

macro(reachable(A),accessible(A)).
rx0(accessible(A),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    reachable_fa(Fa0,Fa,Debug).

reachable_fa(Fa0,Fa,Debug):-
    ltime(Debug,fsa_regex:reachable_fa(Fa0,Fa)).

reachable_fa(Fa0,Fa):-
    all_components_reachable(reachable,Fa0,Fa).

prepare_reachable(reachable,Fa,d(Finals,Table),Starts,Sigma):-
    fsa_data:fsa_start_states(Fa,Starts),
    fsa_data:fsa_symbols_decl(Fa,Sigma),
    fsa_data:index_finals(Fa,Finals),
    trans_jumps_table(Fa,Table).

trans_jumps_table(Fa,Table):-
    fsa_data:fsa_states_number(Fa,Size),
    fsa_data:fsa_transitions(Fa,Trans),
    fsa_data:fsa_jumps(Fa,Jumps),
    fsa_array:fsa_array_new(Table,Size),
    trans_to_keylist(Trans,Keylist0,Keylist1),
    jumps_to_keylist(Jumps,Keylist1,[]),
    keysort(Keylist0,Keylist),
    mult_keylist_to_thing(Keylist,Table).

trans_to_keylist([],T,T).
trans_to_keylist([trans(A,B,C)|Tail],[A-trans(A,B,C)|Trs0],Trs):-
    trans_to_keylist(Tail,Trs0,Trs).

jumps_to_keylist([],T,T).
jumps_to_keylist([jump(A,C)|Tail],[A-jump(A,C)|Trs0],Trs):-
    jumps_to_keylist(Tail,Trs0,Trs).

trans_reachables(reachable,A,d(_,Table),Trans):-
    fsa_array:fsa_array_access(A,Trans,[],Table).

is_final_reachable(reachable,d(Fins,_),P,Bool):-
    fsa_data:indexed_final(P,Fins,Bool).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%% COMPLETION %%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(complete,"complete(A)",
"Adds transitions and a sink state such that the transition table is\
 \n total, i.e. there is a transition for every symbol from every\
 \n state. If A is a transducer then it is temporarily treated as a\
 \n recognizer over pairs of symbols. If A is a transducer then it is\
 \n temporarily treated as a recognizer.").

macro(complete(A),melt(complete0(freeze(A,Flag)),Flag)).

rx0(complete0(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_regex:complete_fa(Fa0,Fa)).

complete_fa(Fa0,Fa):-
    fsa_data:fsa_copy_except(states,Fa0,Fa1,States0,States),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans,CompletedTrans),
    States is States0+1,
    fsa_data:fsa_symbols_decl(Fa0,r(PredMod)),
    fsa_pred_module:prepare_complement_of_set(PredMod,Fa0,Set),
    complete_trans(0,Trans,CompletedTrans0,[],States0,PredMod,Set),
    sort(CompletedTrans0,CompletedTrans).

complete_trans(State,Trans0,Ts0,Ts,Sink,PredMod,Set) :-
    (	State > Sink
    ->	Ts0=Ts
    ;   complete_trans_states_symbols(State,Trans0,Trans,Ts0,Ts1,
				      Sink,PredMod,Set),
	State2 is State+1,
	complete_trans(State2,Trans,Ts1,Ts,Sink,PredMod,Set)
    ).

complete_trans_states_symbols(State,Trans0,Trans,Ts0,Ts,Sink,PredMod,Set) :-
    get_all_preds(Trans0,Trans,Ts0,Ts1,State,Preds),
    add_trans_for_negation(Preds,State,Sink,Ts1,Ts,PredMod,Set).

get_all_preds([],[],Ts,Ts,_,[]).
get_all_preds([trans(P,Pred,Q)|Trans0],Trans,Ts0,Ts,State,Preds) :-
    (	State==P
    ->	Preds=[Pred|Preds0],
	Ts0=[trans(P,Pred,Q)|Ts1],
	get_all_preds(Trans0,Trans,Ts1,Ts,State,Preds0)
    ;   Trans=[trans(P,Pred,Q)|Trans0],
	Ts0=Ts,
	Preds=[]
    ).

add_trans_for_negation(Preds,State,Sink,Ts0,Ts,PredMod,Set) :-
    fsa_pred_module:complement_of_set(PredMod,Preds,Set,Complements),
    complement_transitions(Complements,State,Sink,Ts0,Ts).

complement_transitions([],_,_,Ts,Ts).
complement_transitions([C|Cs],P,Q,[trans(P,C,Q)|Ts0],Ts):-
    complement_transitions(Cs,P,Q,Ts0,Ts).

%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% IGNORE %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%

op_help(ignore,"ignore(A,B)",
"Strings from A interspersed with strings from B. For instance,\
 \n ignore([a,a,a],c) contains all strings over the alphabet {a,c} which\
 \n contain exactly three a's. Both A and B must be recognizers."). 
macro(ignore(A,B), m(ignore0(recognizer(A),recognizer(B)))).

rx0(ignore0(A,B),Fa,Debug,Cache,Level) :-
    rx(A,FaA,Debug,Cache,Level),
    rx(B,FaB,Debug,Cache,Level),
    ltime(Debug,fsa_regex:ignore_fa(FaA,FaB,Fa)).

ignore_fa(FaA,FaB,Fa):-
    all_components_reachable(ignore,FaA,FaB,Fa).

prepare_reachable(ignore,FaA,FaB,
	     d(StartsB,FinalsA,FinalsB,TransA,TransB,JumpsA,JumpsB),
		  Starts,Sigma):-
    coerce_sigma(FaA,FaB,Sigma),
    findall(a(Start),fsa_data:fsa_start_state(FaA,Start),Starts),
    fsa_data:index_finals(FaA,FinalsA),
    fsa_data:index_finals(FaB,FinalsB),
    index_jumps_if_ne(FaA,JumpsA),
    index_jumps_if_ne(FaB,JumpsB),
    fsa_data:index_transitions(FaA,TransA),
    fsa_data:index_transitions(FaB,TransB),
    fsa_data:fsa_start_states(FaB,StartsB).

is_final_reachable(ignore,d(_,FinalsA,_,_,_,_,_),State,Bool):-
    ignore_final(State,FinalsA,Bool).

ignore_final(b(_,_),_,no).
ignore_final(a(A),FinalsA,Bool):-
    fsa_data:indexed_final(A,FinalsA,Bool).

trans_reachables(ignore,State,Data,News):-
    findall(Trans,ignore_trans(State,Data,Trans),News).

ignore_trans(a(P),d(_,_,_,_,_,jumps(JA),_),jump(a(P),a(Q))):-
    fsa_data:indexed_jump(P,JA,Q).

ignore_trans(a(P),d(_,_,_,TA,_,_,_),trans(a(P),Sym,a(Q))):-
    fsa_data:indexed_transition(P,TA,Sym,Q).

ignore_trans(a(P),d(Starts,_,_,_,_,_,_),jump(a(P),b(Q,P))):-
    member(Q,Starts).

ignore_trans(b(Q,P),d(_,_,Fins,_,_,_,_),jump(b(Q,P),a(P))):-
    fsa_data:indexed_final(Q,Fins,yes).

ignore_trans(b(P,B),d(_,_,_,_,_,_,jumps(Js)),jump(b(P,B),b(Q,B))):-
    fsa_data:indexed_jump(P,Js,Q).

ignore_trans(b(P,B),d(_,_,_,_,Ts,_,_),trans(b(P,B),Sym,b(Q,B))):-
    fsa_data:indexed_transition(P,Ts,Sym,Q).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%% TYPES %%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

macro(transducer(A),transducer(A,silent)).
rx0(transducer(A,Style),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_regex:identity_if_recognizer(Type,Style,A,Fa0,Fa)).

rx0(identity_if_recognizer(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_regex:identity_if_recognizer_silent(Type,Fa0,Fa)).

identity_if_recognizer_silent(recognizer,Fa0,Fa):-
    identity_fa(Fa0,Fa).
identity_if_recognizer_silent(transducer(_),Fa,Fa).
identity_if_recognizer_silent(w_recognizer,Fa,Fa).
identity_if_recognizer_silent(w_transducer,Fa,Fa).

identity_if_recognizer(recognizer,Style,A,Fa0,Fa):-
    warning(Style,transducer,A),
    identity_fa(Fa0,Fa).
identity_if_recognizer(transducer(_),_,_,Fa,Fa).
identity_if_recognizer(w_recognizer,_,A,_,_) :-
    warning(error,transducer,A).
identity_if_recognizer(w_transducer,_,A,_,_) :-
    warning(error,transducer,A).

rx0(recognizer(A),Fa,Debug,Cache,Level):-
    rx(A,Fa,Debug,Cache,Level),
    fsa_data:fsa_type(Fa,Type),
    (	Type == recognizer
    ->	true
    ;	warning(error,recognizer,A)
    ).

warning(silent,_,_).
warning(warn,Expected,Expr):-
#!yap#     format(user_error,"Warning: ~@ is not a ~w~n",
#!yap# 	   [write_term(Expr,[max_depth(6)]),Expected]).
#yap#     format(user_error,"Warning: ~w is not a ~w~n",[Expr,Expected]).
warning(error,Expected,Expr):-
#!yap#     user:throw(fsa_error("~@ is not a ~w~n",
#!yap# 	[write_term(Expr,[max_depth(6)]),Expected])).
#yap#     user:throw(fsa_error("~w is not a ~w~n",[Expr,Expected])).

%% undocumented: only used internally
%% note: in order to create a sequence transducer, we first create
%% a letter transducer (to get rid of any $dequeue's)!

macro(sequence_transducer(A),
      sequence_transducer(letter_transducer(A),silent)).
rx0(sequence_transducer(A,Style),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_regex:sequence_transducer(Fa0,A,Style,Fa)).

sequence_transducer(Fa0,A,Style,Fa):-
    fsa_data:fsa_type(Fa0,Type),
    identity_if_recognizer(Type,Style,A,Fa0,Fa1),
    fsa_data:fsa_type(Fa1,transducer(Sub)),
    single_els_if_letter(Sub,Fa1,Fa).

single_els_if_letter(sequence,Fa,Fa).
single_els_if_letter(letter,Fa1,Fa):-
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    single_el_list(Trans0,Trans).

macro(sync_sequence_transducer(A),
      sync_sequence_transducer(letter_transducer(A),silent)).
rx0(sync_sequence_transducer(A,Style),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_regex:sync_sequence_transducer(Fa0,A,Style,Fa)).

sync_sequence_transducer(Fa0,A,Style,Fa):-
    warning(Style,sequence_transducer,A),
    fsa_data:fsa_copy_except(transitions,Fa0,Fa,Trans0,Trans),
    single_el_list(Trans0,Trans).

single_el_list([],[]).
single_el_list([trans(A,B/C0,D)|Ts0],[trans(A,B/C,D)|Ts]):-
    single_el(C0,C),
    single_el_list(Ts0,Ts).

single_el(E0,E):-
    (   E0==[]
    ->  E0=E
    ;   E=[E0]
    ).


rx0(zero_weights(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    add_zero_weights(Type,Fa0,Fa,A).
   
%% Conversion to weighted recognizer
%%
macro(weighted_recognizer(A),weighted_recognizer(A,silent)).
rx0(weighted_recognizer(A,Style),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level), %% Fa0 is a letter-transducer
                                 %% every [] on range side is replaced by 0
                                 %% other symbols will/must be integers
    fsa_data:fsa_type(Fa0,Type),
    to_weighted_recognizer(Type,Style,A,Fa0,Fa).

to_weighted_recognizer(recognizer,Style,A,Fa0,Fa) :-
    warning(Style,w_recognizer,A),
    add_zero_weights(Fa0,Fa).
to_weighted_recognizer(w_recognizer,_,_,Fa,Fa).
to_weighted_recognizer(transducer(Sub),Style,A,Fa0,Fa) :-
    warning(Style,w_recognizer,A),
    to_letter_if_sequence(Sub,silent,A,Fa0,Fa1),
    fsa_data:fsa_copy_except(symbols,Fa1,Fa2,t(ModA,ModB),r(ModA,fsa_frozen)),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa,Trans0,Trans),
    weighted_transitions(Trans0,Trans1,ModB),
    sort(Trans1,Trans).
to_weighted_recognizer(w_transducer(_),_,A,_,_) :-
    warning(error,weighted_recognizer,A).

add_zero_weights(recognizer,Fa0,Fa,_A) :-
    add_zero_weights(Fa0,Fa).
add_zero_weights(w_recognizer,_Fa0,_Fa,A) :-
    warning(error,recognizer_or_transducer,A).
add_zero_weights(w_transducer(_),_Fa0,_Fa,A) :-
    warning(error,recognizer_or_transducer,A).
add_zero_weights(transducer(_),Fa0,Fa,_A) :-
    add_zero_weights_transducer(Fa0,Fa).

add_zero_weights(Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa1,r(M),r(M,fsa_frozen)),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Tr0,Tr),
    add_zero_weights_tr(Tr0,Tr).

add_zero_weights_tr([],[]).
add_zero_weights_tr([trans(P,Sym,Q)|T0],[trans(P,Sym/0,Q)|T]) :-
    add_zero_weights_tr(T0,T).

weighted_transitions([],[],_).
weighted_transitions([trans(P,Dom/Ran0,Q)|Ts0],Ts,OldMod):-
    findall(trans(P,Dom/Ran,Q),w_output(Ran0,Ran,OldMod),Ts,Ts1),
    weighted_transitions(Ts0,Ts1,OldMod).

w_output(Ran0,Ran,OldMod):-
    (	Ran0==[]
    ->	Ran=0
    ;	fsa_pred_module:evaluate_predicate(OldMod,Ran0,Ran)
    ).

%% conversion to weighted (letter-) transducer
macro(weighted_transducer(A),weighted_transducer(A,silent)).
macro(weighted_letter_transducer(A),weighted_transducer(A,silent)).
rx0(weighted_transducer(A,Style),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    to_weighted_transducer(Type,Style,A,Fa0,Fa).

to_weighted_transducer(recognizer,Style,A,Fa0,Fa) :-
    warning(Style,w_transducer,A),
    add_zero_weights(Fa0,Fa1),
    to_weighted_transducer(w_recognizer,silent,A,Fa1,Fa).
to_weighted_transducer(w_recognizer,Style,A,Fa0,Fa) :-
    warning(Style,w_transducer,A),
    identity_weight_fa(Fa0,Fa).
to_weighted_transducer(transducer(_Sub),Style,A,Fa0,Fa) :-
    warning(Style,w_transducer,A),
    add_zero_weights_transducer(Fa0,Fa1),
    w_to_letter_if_sequence(Fa1,Style,A,Fa).
to_weighted_transducer(w_transducer(Sub),Style,A,Fa0,Fa) :-
    w_to_letter_if_sequence(Sub,Style,A,Fa0,Fa).

w_to_letter_if_sequence(Fa0,Style,A,Fa) :-
    fsa_data:fsa_type(Fa0,w_transducer(Sub)),
    w_to_letter_if_sequence(Sub,Style,A,Fa0,Fa).

w_to_letter_if_sequence(letter,_,_,Fa,Fa).
w_to_letter_if_sequence(sequence,Style,A,Fa0,Fa):-
    warning(Style,weighted_letter_transducer,A),
    fsa_data:fsa_components(Symbols,Number0,Starts,Finals,Trs0,Jumps,Fa0),
    w_td_transs(Trs0,Trs,Number0,Number),
    fsa_data:fsa_construct(Symbols,Number,Starts,Finals,Trs,Jumps,Fa1),
    fsa_sync:sync_w(Fa1,Fa).

w_td_transs(Tr0,Tr,N0,N) :-
    w_td_transs(Tr0,Tr,[],N0,N).
 
w_td_transs([],T,T,N,N).
w_td_transs([trans(P,In/(Outs/W),Q)|T],Tr0,Tr,N0,N) :-
    w_td_trans(Outs,W,P,In,Q,Tr0,Tr1,N0,N1),
    w_td_transs(T,Tr1,Tr,N1,N).

w_td_trans([],W,P0,In,P,[trans(P0,In/([]/W),P)|Tr],Tr,N,N).
w_td_trans([H|T],W,P0,In,P,Tr0,Tr,N0,N):-
    w_td_trans1(T,H,W,P0,In,P,Tr0,Tr,N0,N).

w_td_trans1([],H,W,P0,In,P,[trans(P0,In/(H/W),P)|Tr],Tr,N,N).
w_td_trans1([T0|TT],H,W,P0,In,P,[trans(P0,In/(H/W),P1)|Tr0],Tr,N0,N) :-
    w_td_trans0([T0|TT],P1,P,Tr0,Tr,N0,N).

w_td_trans0([],P,P,Tr,Tr,N,N).
w_td_trans0([H|T],P0,P,[trans(P0,[]/(H/0),P1)|Tr0],Tr,N0,N):-
    P0=N0,
    N1 is N0+1,
    w_td_trans0(T,P1,P,Tr0,Tr,N1,N).



add_zero_weights_transducer(Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa1,t(Ma,Mb),t(Ma,Mb,fsa_frozen)),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Tr0,Tr),
    add_zero_weights_transducer_tr(Tr0,Tr).

add_zero_weights_transducer_tr([],[]).
add_zero_weights_transducer_tr([trans(P,Sym/R,Q)|T0],
			       [trans(P,Sym/(R/0),Q)|T]) :-
    add_zero_weights_transducer_tr(T0,T).

identity_weight_fa(Fa0,Fa):-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa1,r(Sigma,fsa_frozen),
			                     t(Sigma,Sigma,fsa_frozen)),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    identity_weight_transitions(Trans0,Trans1,Sigma),
    sort(Trans1,Trans).  

identity_weight_transitions([],[],_).
identity_weight_transitions([trans(P,Sym0/W,Q)|T0],
		     [trans(P,Sym/(Sym/W),Q)|T],Module):-
    fsa_pred_module:identity(Module,Sym0,Sym),
    identity_weight_transitions(T0,T,Module).

macro(weighted_sequence_transducer(A),weighted_sequence_transducer(A,silent)).
rx0(weighted_sequence_transducer(A,Style),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    to_weighted_seq_transducer(Type,Style,A,Fa0,Fa).

to_weighted_seq_transducer(recognizer,Style,A,Fa0,Fa) :-
    warning(Style,w_transducer,A),
    add_zero_weights(Fa0,Fa1),
    to_weighted_seq_transducer(w_recognizer,silent,A,Fa1,Fa).
to_weighted_seq_transducer(w_recognizer,Style,A,Fa0,Fa) :-
    warning(Style,w_transducer,A),
    identity_seq_weight_fa(Fa0,Fa).
to_weighted_seq_transducer(transducer(_Sub),Style,A,Fa0,Fa) :-
    warning(Style,w_transducer,A),
    add_zero_weights_transducer(Fa0,Fa1),
    w_to_sequence_if_letter(Fa1,Style,A,Fa).
to_weighted_seq_transducer(w_transducer(Sub),Style,A,Fa0,Fa) :-
    w_to_sequence_if_letter_(Sub,Style,A,Fa0,Fa).

w_to_sequence_if_letter(Fa0,Style,A,Fa) :-
    fsa_data:fsa_type(Fa0,w_transducer(Sub)),
    w_to_sequence_if_letter_(Sub,Style,A,Fa0,Fa).

w_to_sequence_if_letter_(sequence,_,_,Fa,Fa).
w_to_sequence_if_letter_(letter,_,_,Fa0,Fa) :-
    ws_add_brackets(Fa0,Fa).

ws_add_brackets(Fa0,Fa) :-
    fsa_data:fsa_copy_except(transitions,Fa0,Fa,Trans0,Trans),
    ws_add_brackets_ts(Trans0,Trans1),
    sort(Trans1,Trans).

ws_add_brackets_ts([],[]).
ws_add_brackets_ts([trans(P,A/(B0/C),Q)|T0],[trans(P,A/(B/C),Q)|T]) :-
    (	B0 == []
    ->	B0=B
    ;   B=[B0]
    ),	
    ws_add_brackets_ts(T0,T).

identity_seq_weight_fa(Fa0,Fa):-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa1,r(Sigma,fsa_frozen),
			                     t(Sigma,Sigma,fsa_frozen)),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    identity_seq_weight_transitions(Trans0,Trans1,Sigma),
    sort(Trans1,Trans).  

identity_seq_weight_transitions([],[],_).
identity_seq_weight_transitions([trans(P,Sym0/W,Q)|T0],
		     [trans(P,Sym/([Sym]/W),Q)|T],Module):-
    fsa_pred_module:identity(Module,Sym0,Sym),
    identity_seq_weight_transitions(T0,T,Module).

%% undocumented: only used internally
%% a transducer with sequences of symbols on the range side,
%% is converted into a transducer where all range symbols are single
%% character (or epsilon)
macro(letter_transducer(A),letter_transducer(A,silent)).
rx0(letter_transducer(A,Style),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_regex:letter_transducer(Fa0,A,Style,Fa)).

letter_transducer(Fa0,A,Style,Fa):-
    fsa_data:fsa_type(Fa0,Type),
    identity_if_recognizer(Type,Style,A,Fa0,Fa1),
    fsa_data:fsa_type(Fa1,transducer(Type1)),
    to_letter_if_sequence(Type1,Style,A,Fa1,Fa).

to_letter_if_sequence(letter,_,_,Fa,Fa).
to_letter_if_sequence(sequence,Style,A,Fa0,Fa):-
    warning(Style,letter_transducer,A),
    fsa_data:fsa_components(Symbols,Number0,Starts,Finals,Trs0,Jumps,Fa0),
    td_transs(Trs0,Trs,Number0,Number),
    fsa_data:fsa_construct(Symbols,Number,Starts,Finals,Trs,Jumps,Fa1),
    fsa_sync:sync(Fa1,Fa).

td_transs(Tr0,Tr,N0,N) :-
    td_transs(Tr0,Tr,[],N0,N).
 
td_transs([],T,T,N,N).
td_transs([trans(P,In/Outs,Q)|T],Tr0,Tr,N0,N) :-
    td_trans(Outs,P,In,Q,Tr0,Tr1,N0,N1),
    td_transs(T,Tr1,Tr,N1,N).

td_trans([],P0,In,P,[trans(P0,In/[],P)|Tr],Tr,N,N).
td_trans([H|T],P0,In,P,Tr0,Tr,N0,N):-
    td_trans1(T,H,P0,In,P,Tr0,Tr,N0,N).

td_trans1([],H,P0,In,P,[trans(P0,In/H,P)|Tr],Tr,N,N).
td_trans1([T0|TT],H,P0,In,P,[trans(P0,In/H,P1)|Tr0],Tr,N0,N) :-
    td_trans0([T0|TT],P1,P,Tr0,Tr,N0,N).

td_trans0([],P,P,Tr,Tr,N,N).
td_trans0([H|T],P0,P,[trans(P0,[]/H,P1)|Tr0],Tr,N0,N):-
    P0=N0,
    N1 is N0+1,
    td_trans0(T,P1,P,Tr0,Tr,N1,N).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%% UNION %%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% 0-ary operators
%% empty {} list denotes empty language
op_help('{}',"{}",
"{} denotes the empty language.").
rx0('{}',Fa,_,_,_) :-
    fsa_globals:fsa_global_get(pred_module,Module),
    fsa_data:fsa_components(r(Module),1,[0],[],[],[],Fa).

op_help(union,"{E1,E2,..,En}   union(E1,E2)   set([E1,E2,..,En])",
"Union of the languages denoted by E1,..,En. As a special case, '{}' is\
 \n the empty language, i.e. a language without any strings. Note that the\
 \n result is a minimal automaton. E1 .. En can be both transducers or\
 \n recognizers. If one of them is a transducer, then all of the others are\
 \n coerced into transducers as well.").  

macro(set(X), union_list(X)).
macro(union(A,B),union_list([A,B])).

macro('{}'(X),union_list(List)):-
    conj_to_list(X,List).

conj_to_list(V,List):-
    (	var(V)
    ->  user:throw(fsa_error("variable in regular expr~n",[]))
    ;   V = (H,T)
    ->	List= [H|List0], conj_to_list(T,List0)
    ;   List=[V]
    ).

rx0(union_list(List),Fa,Debug,Cache,Level):-
    union_list(List,Fa,Debug,Cache,Level).

union_list([],Fa,Debug,Cache,Level):-
    rx('{}',Fa,Debug,Cache,Level).
union_list([H|T],Fa,Debug,Cache,Level):-
    union_list0(T,H,Fa,Debug,Cache,Level).

union_list0([],H,Fa,Debug,Cache,Level):-
    rx(H,Fa,Debug,Cache,Level).
union_list0([Head|Tail],First,Fa,Debug,Cache,Level):-
    rx_list([First,Head|Tail],FaList0,Debug,Cache,Level),
    coerce_list(FaList0,FaList1,Debug,Cache,Level),
    coerce_sigma_list(FaList1,Sigma),
    increment_and_union_list(FaList1,Nr,S,F,T,J),
    fsa_data:fsa_construct(Sigma,Nr,S,F,T,J,Fa0),
    rx(m(fa(Fa0)),Fa,Debug,Cache,Level).

increment_and_union_list([H|Tail],Nr,S,F,T,J) :-
    fsa_data:fsa_components(_,Nr0,S0,F0,T0,J0,H),
    append(S0,S1,S),
    append(F0,F1,F),
    append(T0,T1,T),
    append(J0,J1,J),
    increment_and_union_list(Tail,Nr0,Nr,S1,F1,T1,J1).

increment_and_union_list([],Nr,Nr,[],[],[],[]).
increment_and_union_list([H|Tail],Nr0,Nr,S,F,T,J) :-
    fsa_data:fsa_components(_,Nr1,S0,F0,T0,J0,H),
    incr_states(S0,S,S1,Nr0),
    incr_states(F0,F,F1,Nr0),
    incr_trans(T0,T,T1,Nr0),
    incr_jumps(J0,J,J1,Nr0),
    Nr2 is Nr0+Nr1,
    increment_and_union_list(Tail,Nr2,Nr,S1,F1,T1,J1).

rx_list([],[],_,_,_).
rx_list([H0|T0],[H|T],Debug,Cache,Level):-
    rx(H0,H,Debug,Cache,Level),
    rx_list(T0,T,Debug,Cache,Level).

coerce_sigma_list([H|T],Sigma):-
    fsa_data:fsa_symbols_decl(H,Sigma),
    coerce_sigma_list(T,H,Sigma).

coerce_sigma_list([],_,_).
coerce_sigma_list([H|T],N,Sigma):-
    coerce_sigma(H,N,Sigma),
    coerce_sigma_list(T,H,Sigma).

coerce_list(List0,List,Debug,Cache,Level) :-
    coerce_list(List0,List,recognizer,_,Debug,Cache,Level).

coerce_list(List0,List,Least,Type,Debug,Cache,Level):-
    rec_or_tr(List0,Least,Type),
    coerce_to(List0,Type,List,Debug,Cache,Level).

rec_or_tr([],T,T).
rec_or_tr([H|T],L0,L):-
    fsa_data:fsa_type(H,Type),
    coerce_two(Type,L0,L1),
    rec_or_tr(T,L1,L).

%% current Type, general type so far, general type
%% 
coerce_two(recognizer,L,L).
coerce_two(w_recognizer,L0,L) :-
    coerce_w_recognizer(L0,L).
coerce_two(transducer(_),Prev,Transducer) :-
    coerce_transducer(Prev,Transducer).
coerce_two(w_transducer(_),_,w_transducer).

coerce_transducer(recognizer,transducer).
coerce_transducer(transducer,transducer).
coerce_transducer(w_recognizer,w_transducer).
coerce_transducer(w_transducer,w_transducer).

coerce_w_recognizer(recognizer,   w_recognizer).
coerce_w_recognizer(w_recognizer, w_recognizer).
coerce_w_recognizer(transducer,   w_transducer).
coerce_w_recognizer(w_transducer, w_transducer).

coerce_to([],_,[],_,_,_).
coerce_to([H0|T0],Type,[H|T],Debug,Cache,Level):-
%    fsa_data:fsa_type(H0,Type0),
    coerce1(Type,H0,H,Debug,Cache,Level),
    coerce_to(T0,Type,T,Debug,Cache,Level).

coerce1(recognizer,H,H,_D,_C,_L).
coerce1(transducer,H0,H,D,C,L) :-
    rx(letter_transducer(fa(H0),silent),H,D,C,L).
coerce1(w_transducer,H0,H,D,C,L) :-
    rx(weighted_transducer(fa(H0),silent),H,D,C,L).
coerce1(w_recognizer,H0,H,D,C,L) :-
    rx(weighted_recognizer(fa(H0),silent),H,D,C,L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%% CONCAT %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% [] denotes fsa defining the language consisting only of empty string
%% which we take to be the same as the empty string itself
op_help([],"[]",
"[] denotes the empty string (or equivalently the language\
 \n solely consisting of the empty string)."). 
rx0([],Fa,_,_,_) :-
    fsa_globals:fsa_global_get(pred_module,Module),
    fsa_data:fsa_components(r(Module),1,[0],[0],[],[],Fa).

op_help(concat,"[E1,E2,...En]   concat(E1,E2)",
"The concatenation of the languages denoted by E1, E2, .. En. As a\
 \n special case, [] is the language solely containing of the empty\
 \n string. Note that the result is a minimal automaton. E1 .. En can\
 \n be both recognizers and transducers. If one of them is a transducer\
 \n then all of the others are coerced into transducers as well."). 
macro([H|T],concat_list([H|T])).
macro(concat(A,B),concat_list([A,B])).

rx0(concat_list(List),Fa,Debug,Cache,Level):-
    concat_list(List,Fa0,Debug,Cache,Level),
    rx(m(fa(Fa0)),Fa,Debug,Cache,Level).

rx0(concat_list_nm(List),Fa,Debug,Cache,Level):-
    concat_list(List,Fa,Debug,Cache,Level).

concat_list([],Fa,Debug,Cache,Level):-
    rx([],Fa,Debug,Cache,Level).
concat_list([H|T],Fa,Debug,Cache,Level):-
    concat_list0(T,H,Fa,Debug,Cache,Level).

concat_list0([],H,Fa,Debug,Cache,Level):-
    rx(H,Fa,Debug,Cache,Level).
concat_list0([Head|Tail],First,Fa,Debug,Cache,Level):-
    rx_list([First,Head|Tail],FaList0,Debug,Cache,Level),
    coerce_list(FaList0,FaList1,Debug,Cache,Level),
    coerce_sigma_list(FaList1,Sigma),
    increment_and_concat_list(FaList1,Nr,S,F,T,J),
    fsa_data:fsa_construct(Sigma,Nr,S,F,T,J,Fa).

increment_and_concat_list([H|Tail],Nr,S,F,T,J) :-
    fsa_data:fsa_components(_,Nr0,S,F0,T0,J0,H),
    append(T0,T1,T),
    append(J0,J1,J),
    increment_and_concat_list(Tail,Nr0,Nr,F0,F,T1,J1).

increment_and_concat_list([],Nr,Nr,F,F,[],[]).
increment_and_concat_list([H|Tail],Nr0,Nr,F0,F,T,J) :-
    fsa_data:fsa_components(_,Nr1,S0,LF0,T0,J0,H),
    incr_states(S0,S,[],Nr0),
    incr_states(LF0,LF,[],Nr0),
    incr_trans(T0,T,T1,Nr0),
    add_jumps_concat(F0,S,J,J1),
    incr_jumps(J0,J1,J2,Nr0),
    Nr2 is Nr0+Nr1,
    %% add jumps from previous final to current start
    increment_and_concat_list(Tail,Nr2,Nr,LF,F,T1,J2).

add_jumps_concat([],_,J,J).
add_jumps_concat([Fin|Fins],Starts,J0,J):-
    add_jumps_concat1(Starts,Fin,J0,J1),
    add_jumps_concat(Fins,Starts,J1,J).

add_jumps_concat1([],_,J,J).
add_jumps_concat1([Start|Starts],Fin,[jump(Fin,Start)|Js0],Js):-
    add_jumps_concat1(Starts,Fin,Js0,Js).

incr_states([],X,X,_).
incr_states([S0|T0],[S|T1],T,Incr):-
    S is Incr+S0,
    incr_states(T0,T1,T,Incr).

incr_trans([],Tr,Tr,_).
incr_trans([trans(S0,Sym,Q0)|T0],[trans(S,Sym,Q)|Tr0],Tr,Incr):-
    S is Incr+S0,
    Q is Incr+Q0,
    incr_trans(T0,Tr0,Tr,Incr).

incr_jumps([],Js,Js,_).
incr_jumps([jump(S0,Q0)|T0],[jump(S,Q)|Js0],Js,Incr):-
    S is Incr+S0,
    Q is Incr+Q0,
    incr_jumps(T0,Js0,Js,Incr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%% KLEENE CLOSURE %%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(kleene_star,"E*   kleene_star(E)",
"Kleene closure (zero or more concatenations) of the language denoted\
 \n by E. Note that the result is a minimal automaton. E can be both a \
 \n recognizer or a transducer.").  
macro(kleene(A),     m(kleene_star0(A))). 
macro(kleene_star(A),m(kleene_star0(A))).
macro('*'(A),        m(kleene_star0(A))).

rx0(kleene_star0(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    kleene_star_fa(Fa0,Fa).

kleene_star_fa(Fa0,Fa):-
    fsa_data:fsa_components(Syms,States0,StartStates,Finals0,Trans,Jumps0,Fa0),
    fsa_data:fsa_components(Syms,States, [States0],  Finals, Trans,Jumps,Fa),
    States is States0 + 1,
    append(Finals0,[States0],Finals),
    new_jumps(StartStates,Finals,Jumps1,Jumps0),
    sort(Jumps1,Jumps).

op_help(kleene_plus,"E+   [kleene_]plus(E)",
"Kleene plus (one or more concatentations) of the language denoted by\
 \n E. Note that the result is a minimal automaton. E can be both a recognizer\
 \n or a transducer."). 
macro(plus(A),       m(kleene_plus0(A))).
macro(kleene_plus(A),m(kleene_plus0(A))). 
macro('+'(A),        m(kleene_plus0(A))). 

rx0(kleene_plus0(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    kleene_plus_fa(Fa0,Fa).

kleene_plus_fa(Fa0,Fa):-
    fsa_data:fsa_components(Syms,States0,StartStates,Finals,Trans,Jumps0,Fa0),
    fsa_data:fsa_components(Syms,States, [States0],  Finals,Trans,Jumps,Fa),
    States is States0 + 1,
    new_jumps(StartStates,[States0|Finals],Jumps1,Jumps0),
    sort(Jumps1,Jumps).

op_help(option,"option(E)   E^",
"Union of E with the empty string, i.e. a string from E occurs\
 \n optionally. The result is a minimal automaton. E can be both a recognizer\
 \n or a transducer."). 
macro(option(A),m(option0(A))). 
macro('^'(A),   m(option0(A))). 

rx0(option0(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    option_fa(Fa0,Fa).

option_fa(Fa0,Fa):-
    fsa_data:fsa_components(Syms,States0,StartStates,Finals0,Trans,Jumps0,Fa0),
    fsa_data:fsa_components(Syms,States, [States0],  Finals, Trans,Jumps, Fa),
    States is States0 + 1,
    append(Finals0,[States0],Finals),
    new_jumps(StartStates,[States0],Jumps1,Jumps0),
    sort(Jumps1,Jumps).

%% new_jumps(L1,L2,J0,J)
%% add jump(l2,l1) for each l2 in L2, l1 in L1
new_jumps([],_,J,J).
new_jumps([H|T],Finals,J0,J) :-
    new_jumps0(Finals,H,J0,J1),
    new_jumps(T,Finals,J1,J).

new_jumps0([],_,J,J).
new_jumps0([F|Fs],S,[jump(F,S)|J0],J):-
    new_jumps0(Fs,S,J0,J).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% INTERSECTION / COMPOSITION %%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(intersect,"intersect[ion](A,B)    A & B",
"The intersection of the languages denoted by A and B. Produces a\
 \n minimal automaton. A and B must be recognizers."). 
macro(intersect(A,B),intersection(A,B)).
macro(intersection(A,B), minimize(intersect0(A,B))).
macro('&'(A,B),          minimize(intersect0(A,B))).

op_help(intersect_list,"intersect_list([E0,E1,...,En])",
"This is equivalent to the sequential intersection of\
 \n E0, E1, .., En, i.e., to the expression\
 \n \
 \n     E0 & E1 & ... & En").  

macro(intersect_list(List),Expr) :-
    intersect_list(List,Expr).

%% to find the composition of a list of expressions, first
%% take the composition of the first half of the list, then the
%% second half, and then intersect
intersect_list(List,Expr) :-
    length(List,Length),
    intersect_list(Length,List,[],Expr).

intersect_list(Length,List0,List,Expr) :-
    (	Length =:= 0
    ->	Expr =  kleene_star(?),
	List0=List
    ;	Length =:= 1
    ->	List0 = [Expr|List]
    ;	Length =:= 2
    ->	List0 = [E1,E2|List], Expr=intersection(E1,E2)
    ;	A is Length>>1,
	Z is Length-A,
	Expr=intersection(E1,E2),
	intersect_list(A,List0,List1,E1),
	intersect_list(Z,List1,List,E2)
    ).

/*
intersect_all([],kleene_star(?)).
intersect_all([H|T],E) :-
    intersect_all(T,H,E).

intersect_all([],E,E).
intersect_all([H|T],F,intersection(F,R)) :-
    intersect_all(T,H,R).
*/

rx0(intersect0(A,B),Fa,Debug,Cache,Level) :-
    rx(A,FaA0,Debug,Cache,Level),
    rx(B,FaB0,Debug,Cache,Level),
    coerce_list([FaA0,FaB0],[FaA,FaB],recognizer,Type,Debug,Cache,Level),
    ltime(Debug,fsa_regex:intersect_fa(Type,FaA,FaB,intersection(A,B),Fa)).

intersect_fa(recognizer,FaA,FaB,_,Fa):-
    all_components_reachable(intersect,FaA,FaB,Fa).
intersect_fa(w_recognizer,FaA,FaB,_,Fa):-
    all_components_reachable(intersect,FaA,FaB,Fa).
intersect_fa(transducer,_,_,A,_) :-
    warning(error,recognizer_or_w_recognizer,A).
intersect_fa(w_transducer,_,_,A,_) :-
    warning(error,recognizer_or_w_recognizer,A).

prepare_reachable(intersect,FaA,FaB,
	     d(FinalsA,FinalsB,TransA,TransB,JumpsA,JumpsB,Sigma),
		  StartStates,Sigma):-
    coerce_sigma(FaA,FaB,Sigma),
    pairs_of_start_states(FaA,FaB,StartStates),
    fsa_data:index_finals(FaA,FinalsA),
    fsa_data:index_finals(FaB,FinalsB),
    fsa_data:index_jumps(FaA,JumpsA),
    fsa_data:index_jumps(FaB,JumpsB),
    fsa_data:index_transitions(FaA,TransA),
    fsa_data:index_transitions(FaB,TransB).

trans_reachables(intersect,Pa/Pb,d(_,_,Tra,TrB,JA,JB,Mod),Trans):-
    intersect1_jumps(JA,Pa,Pb,Trans,Trans1),
    intersect2_jumps(JB,Pa,Pb,Trans1,Trans2),
    intersect_trans(Pa,Pb,Tra,TrB,Mod,Trans2,[]).

intersect1_jumps(JT,Pa,Pb,L0,L):-
    fsa_array:fsa_array_access(Pa,List0,[],JT),
    intersect1_list(List0,Pa,Pb,L0,L).

intersect1_list([],_,_,L,L).
intersect1_list([Qa|Tail],Pa,Pb,[jump(Pa/Pb,Qa/Pb)|L0],L):-
    intersect1_list(Tail,Pa,Pb,L0,L).

intersect2_jumps(JT,Pa,Pb,L0,L):-
    fsa_array:fsa_array_access(Pb,List0,[],JT),
    intersect2_list(List0,Pa,Pb,L0,L).

intersect2_list([],_,_,L,L).
intersect2_list([Qb|Tail],Pa,Pb,[jump(Pa/Pb,Pa/Qb)|L0],L):-
    intersect2_list(Tail,Pa,Pb,L0,L).

intersect_trans(Pa,Pb,TrA0,TrB0,Mod,Trans0,Trans) :-
    fsa_array:fsa_array_access(Pa,TrA,[],TrA0), 
    fsa_array:fsa_array_access(Pb,TrB,[],TrB0),
    trans_intersect(TrA,TrB,Mod,Pa/Pb,Trans0,Trans).

trans_intersect([],_,_,_,T,T).
trans_intersect([Sym-Qs|Tail],TrB,Mod,Src,Trans0,Trans):-
    trans_intersect0(TrB,Mod,Src,Sym,Qs,Trans0,Trans1),
    trans_intersect(Tail,TrB,Mod,Src,Trans1,Trans).

trans_intersect0([],_,_,_,_,T,T).
trans_intersect0([SymB-Qbs|Tail],Mod,Src,SymA,Qas,Trans0,Trans):-
    (	trans_intersect_sym(Mod,SymA,SymB,Sym)
    ->	trans_intersect1(Qas,Src,Sym,Qbs,Trans0,Trans1)
    ;   Trans0=Trans1
    ),
    trans_intersect0(Tail,Mod,Src,SymA,Qas,Trans1,Trans).

trans_intersect_sym(r(Mod),SymA,SymB,Sym) :-
    fsa_pred_module:conjunction(Mod,SymA,SymB,Sym).
trans_intersect_sym(r(Mod,fsa_frozen),SymA/W0,SymB/W1,Sym/W) :-
    fsa_pred_module:conjunction(Mod,SymA,SymB,Sym),
    fsa_semiring:plus(fsa_weights,W0,W1,W).

trans_intersect1([],_,_,_,T,T).
trans_intersect1([Qa|Qas],Src,Sym,Qbs,Trans0,Trans):-
    trans_intersect2(Qbs,Src,Sym,Qa,Trans0,Trans1),
    trans_intersect1(Qas,Src,Sym,Qbs,Trans1,Trans).

trans_intersect2([],_,_,_,T,T).
trans_intersect2([Qb|Qbs],Src,Sym,Qa,[trans(Src,Sym,Qa/Qb)|Trans0],Trans):-
    trans_intersect2(Qbs,Src,Sym,Qa,Trans0,Trans).

is_final_reachable(intersect,d(FinalsA,FinalsB,_,_,_,_,_),A/B,Bool):-
    fsa_data:indexed_final(A,FinalsA,Bool1),
    fsa_data:indexed_final(B,FinalsB,Bool2),
    and(Bool1,Bool2,Bool).

op_help(compose,"E0 o E1   compose(E0,E1)",
" \
 \n The set of pairs (A,C) such that (A,B) is in E0 and (B,C) is in\
 \n E1. Both E0 and E1 are (coerced into) transducers. Note that\
 \n the result is a minimal automaton.\
 \n \
 \n Note that in case both E0 and E1 are not same-length transducers, then\
 \n often the resulting transducer will give rise to `spurious' results in\
 \n the sense that for a given input the same output is produced several\
 \n times. See the paper by Pereira and Riley, 1996, for some suggestions\
 \n to repair this. Obviously, in cases where you can determinize the\
 \n transducer (with t_determinize) the spurious ambiguities will disappear\
 \n as well.").  

macro(compose(A,B),m(compose0(A,B))).
macro(o(A,B),      m(compose0(A,B))).

op_help(compose_list,"compose_list([E0,E1,...,En])",
"This is equivalent to the sequential composition of\
 \n E0, E1, .., En, i.e., to the expression\
 \n \
 \n     E0 o E1 o ... o En").  

macro(compose_list(List),Expr) :-
    compose_list(List,Expr).

%% to find the composition of a list of expressions, first
%% take the composition of the first half of the list, then the
%% second half, and then compose
compose_list(List,Expr) :-
    length(List,Length),
    compose_list(Length,List,[],Expr).

compose_list(Length,List0,List,Expr) :-
    (	Length =:= 0
    ->	Expr =  kleene_star(?),
	List0=List
    ;	Length =:= 1
    ->	List0 = [Expr|List]
    ;	Length =:= 2
    ->	List0 = [E1,E2|List], Expr=compose(E1,E2)
    ;	A is Length>>1,
	Z is Length-A,
	Expr=compose(E1,E2),
	compose_list(A,List0,List1,E1),
	compose_list(Z,List1,List,E2)
    ).

/*
compose_list([],kleene_star(?)).
compose_list([H|T],E) :-
    compose_list(T,H,E).

compose_list([],E,E).
compose_list([H|T],F,compose(F,R)) :-
    compose_list(T,H,R).
*/


rx0(compose0(A,B),Fa,Debug,Cache,Level) :-
    rx(A,FaA0,Debug,Cache,Level),
    rx(B,FaB0,Debug,Cache,Level),
    coerce_list([FaA0,FaB0],[FaA,FaB],recognizer,Type,Debug,Cache,Level),
    ltime(Debug,fsa_regex:compose_fa(Type,FaA,FaB,Fa)).

compose_fa(transducer,FaA,FaB,Fa) :-
    all_components_reachable(compose,FaA,FaB,Fa).
compose_fa(w_transducer,FaA,FaB,Fa) :-
    all_components_reachable(w_compose,FaA,FaB,Fa).
compose_fa(recognizer,FaA,FaB,Fa) :-
    all_components_reachable(intersect,FaA,FaB,Fa).
compose_fa(w_recognizer,FaA,FaB,Fa) :-
    all_components_reachable(intersect,FaA,FaB,Fa).

prepare_reachable(compose,FaA,FaB,
     d(FsA,FsB,TransA,TransB,ETransA,ETransB,JumpsA,JumpsB,Mod),
		  StartStates,t(Dom,Ran)):-
    fsa_data:fsa_symbols_decl(FaA,t(Dom,Mod0)),
    fsa_data:fsa_symbols_decl(FaB,t(Mod,Ran)),
    (	Mod0 == Mod
    ->	true
    ;	user:throw(fsa_error(
		"composition: predicate modules not equal: ~w ~w~n",
		[Mod0,Mod]))
    ),
    pairs_of_start_states(FaA,FaB,StartStates),
    fsa_data:index_finals(FaA,FsA),
    fsa_data:index_finals(FaB,FsB),
    fsa_data:index_jumps(FaA,JumpsA),
    fsa_data:index_jumps(FaB,JumpsB),
    index_compose_transitions(l,FaA,TransA,ETransA),
    index_compose_transitions(r,FaB,TransB,ETransB).


prepare_reachable(w_compose,FaA,FaB,
     d(FsA,FsB,TransA,TransB,ETransA,ETransB,JumpsA,JumpsB,Mod),
		  StartStates,t(Dom,Ran,fsa_frozen)):-
    fsa_data:fsa_symbols_decl(FaA,t(Dom,Mod,fsa_frozen)),
    fsa_data:fsa_symbols_decl(FaB,t(Mod,Ran,fsa_frozen)),
    pairs_of_start_states(FaA,FaB,StartStates),
    fsa_data:index_finals(FaA,FsA),
    fsa_data:index_finals(FaB,FsB),
    fsa_data:index_jumps(FaA,JumpsA),
    fsa_data:index_jumps(FaB,JumpsB),
    index_w_compose_transitions(l,FaA,TransA,ETransA),
    index_w_compose_transitions(r,FaB,TransB,ETransB).


trans_reachables(compose,Pa/Pb,d(_,_,TrA,TrB,Etra,Etrb,JA,JB,Mod),Trans):-
    compose_trans(Pa,Pb,TrA,TrB,Mod,Trans,Trans1),
    trans_compose_left_jumps(Pa,Pb,Etrb,Trans1,Trans2),
    trans_compose_right_jumps(Pa,Pb,Etra,Trans2,Trans3),
    intersect1_jumps(JA,Pa,Pb,Trans3,Trans4),
    intersect2_jumps(JB,Pa,Pb,Trans4,[]).

compose_trans(Pa,Pb,TrA0,TrB0,Mod,Trans0,Trans) :-
    fsa_array:fsa_array_access(Pa,TrA,[],TrA0), 
    fsa_array:fsa_array_access(Pb,TrB,[],TrB0),
    trans_compose(TrA,TrB,Mod,Pa/Pb,Trans0,Trans).

trans_compose([],_,_,_,T,T).
trans_compose([Sym-Qs|Tail],TrB,Mod,Src,Trans0,Trans):-
    trans_compose0(TrB,Mod,Src,Sym,Qs,Trans0,Trans1),
    trans_compose(Tail,TrB,Mod,Src,Trans1,Trans).

trans_compose0([],_,_,_,_,T,T).
trans_compose0([SymB-Qbs|Tail],Mod,Src,SymA,Qas,Trans0,Trans):-
    (	compose_conjunction(SymA,SymB,X0,X2,Mod)
    ->	(   X0/X2 == []/[]
	->  jumps_compose1(Qas,Src,Qbs,Trans0,Trans1)
	;   trans_compose1(Qas,Src,X0/X2,Qbs,Trans0,Trans1)
	)
    ;   Trans0=Trans1
    ),
    trans_compose0(Tail,Mod,Src,SymA,Qas,Trans1,Trans).

trans_compose1([],_,_,_,T,T).
trans_compose1([Qa|Qas],Src,Sym,Qbs,Trans0,Trans):-
    trans_compose2(Qbs,Src,Sym,Qa,Trans0,Trans1),
    trans_compose1(Qas,Src,Sym,Qbs,Trans1,Trans).

trans_compose2([],_,_,_,T,T).
trans_compose2([Qb|Qbs],Src,Sym,Qa,[trans(Src,Sym,Qa/Qb)|Trans0],Trans):-
    trans_compose2(Qbs,Src,Sym,Qa,Trans0,Trans).

jumps_compose1([],_,_,T,T).
jumps_compose1([Qa|Qas],Src,Qbs,Jumps0,Jumps):-
    jumps_compose2(Qbs,Src,Qa,Jumps0,Jumps1),
    jumps_compose1(Qas,Src,Qbs,Jumps1,Jumps).

jumps_compose2([],_,_,T,T).
jumps_compose2([Qb|Qbs],Src,Qa,Jumps0,Jumps):-
    (	Src==Qa/Qb
    ->	Jumps0=Jumps1
    ;   Jumps0=[jump(Src,Qa/Qb)|Jumps1]
    ),
    jumps_compose2(Qbs,Src,Qa,Jumps1,Jumps).

trans_compose_left_jumps(Pa,Pb,Etrb,Trans0,Trans):-
    fsa_array:fsa_array_access(Pb,Keylist,[],Etrb),
    trans_compose_left_jumps0(Keylist,Pa,Pb,Trans0,Trans).

trans_compose_left_jumps0([],_,_,T,T).
trans_compose_left_jumps0([Sym-Qb|Tail],Pa,Pb,
			  [trans(Pa/Pb,[]/Sym,Pa/Qb)|Trans0],Trans):-
    trans_compose_left_jumps0(Tail,Pa,Pb,Trans0,Trans).

trans_compose_right_jumps(Pa,Pb,Etrb,Trans0,Trans):-
    fsa_array:fsa_array_access(Pa,Keylist,[],Etrb),
    trans_compose_right_jumps0(Keylist,Pa,Pb,Trans0,Trans).

trans_compose_right_jumps0([],_,_,T,T).
trans_compose_right_jumps0([Sym-Qa|Tail],Pa,Pb,
			  [trans(Pa/Pb,Sym/[],Qa/Pb)|Trans0],Trans):-
    trans_compose_right_jumps0(Tail,Pa,Pb,Trans0,Trans).

compose_conjunction($@(X1a),PairB,L,R,Mod):-
    compose_id(PairB,X1a,L,R,Mod).
compose_conjunction(X0/X1a,PairB,X0,R,Mod):-
    compose_non_id(PairB,X1a,R,Mod).

compose_non_id($@(X1b),X1a,X2,Mod):-
    fsa_pred_module:conjunction(Mod,X1a,X1b,X2).
compose_non_id(X1b/X2,X1a,X2,Mod):-
    fsa_pred_module:conjunction(Mod,X1a,X1b,_).

compose_id($@(X1b),X1a,X1pred,X1pred,Mod):-
    fsa_pred_module:conjunction(Mod,X1a,X1b,X1),
    fsa_pred_module:identity(Mod,X1,X1pred).
compose_id(X1b/X2,X1a,X0,X2,Mod):-
    fsa_pred_module:conjunction(Mod,X1a,X1b,X0).

trans_reachables(w_compose,Pa/Pb,d(_,_,TrA,TrB,Etra,Etrb,JA,JB,Mod),Trans):-
    w_compose_trans(Pa,Pb,TrA,TrB,Mod,Trans,Trans1),
    trans_w_compose_left_jumps(Pa,Pb,Etrb,Trans1,Trans2),
    trans_w_compose_right_jumps(Pa,Pb,Etra,Trans2,Trans3),
    intersect1_jumps(JA,Pa,Pb,Trans3,Trans4),
    intersect2_jumps(JB,Pa,Pb,Trans4,[]).

w_compose_trans(Pa,Pb,TrA0,TrB0,Mod,Trans0,Trans) :-
    fsa_array:fsa_array_access(Pa,TrA,[],TrA0), 
    fsa_array:fsa_array_access(Pb,TrB,[],TrB0),
    trans_w_compose(TrA,TrB,Mod,Pa/Pb,Trans0,Trans).

trans_w_compose([],_,_,_,T,T).
trans_w_compose([Sym-Qs|Tail],TrB,Mod,Src,Trans0,Trans):-
    trans_w_compose0(TrB,Mod,Src,Sym,Qs,Trans0,Trans1),
    trans_w_compose(Tail,TrB,Mod,Src,Trans1,Trans).

trans_w_compose0([],_,_,_,_,T,T).
trans_w_compose0([SymB-Qbs|Tail],Mod,Src,SymA,Qas,Trans0,Trans):-
    (	w_compose_conjunction(SymA,SymB,X0,X2,W,Mod)
    ->	(   X0/(X2/W) == []/([]/0)
	->  jumps_compose1(Qas,Src,Qbs,Trans0,Trans1)
	;   trans_compose1(Qas,Src,X0/(X2/W),Qbs,Trans0,Trans1)
	)
    ;   Trans0=Trans1
    ),
    trans_w_compose0(Tail,Mod,Src,SymA,Qas,Trans1,Trans).

w_compose_conjunction(SymA/W0,SymB/W1,X0,X2,W,Mod) :-
    compose_conjunction(SymA,SymB,X0,X2,Mod),
    fsa_semiring:plus(fsa_weights,W0,W1,W).

trans_w_compose_left_jumps(Pa,Pb,Etrb,Trans0,Trans):-
    fsa_array:fsa_array_access(Pb,Keylist,[],Etrb),
    trans_w_compose_left_jumps0(Keylist,Pa,Pb,Trans0,Trans).

trans_w_compose_left_jumps0([],_,_,T,T).
trans_w_compose_left_jumps0([(Sym/W)-Qb|Tail],Pa,Pb,
			  [trans(Pa/Pb,[]/(Sym/W),Pa/Qb)|Trans0],Trans):-
    trans_w_compose_left_jumps0(Tail,Pa,Pb,Trans0,Trans).

trans_w_compose_right_jumps(Pa,Pb,Etrb,Trans0,Trans):-
    fsa_array:fsa_array_access(Pa,Keylist,[],Etrb),
    trans_w_compose_right_jumps0(Keylist,Pa,Pb,Trans0,Trans).

trans_w_compose_right_jumps0([],_,_,T,T).
trans_w_compose_right_jumps0([(Sym/W)-Qa|Tail],Pa,Pb,
			  [trans(Pa/Pb,Sym/([]/W),Qa/Pb)|Trans0],Trans):-
    trans_w_compose_right_jumps0(Tail,Pa,Pb,Trans0,Trans).

is_final_reachable(compose,d(FinalsA,FinalsB,_,_,_,_,_,_,_),A/B,Bool):-
    fsa_data:indexed_final(A,FinalsA,Bool1),
    fsa_data:indexed_final(B,FinalsB,Bool2),
    and(Bool1,Bool2,Bool).

is_final_reachable(w_compose,d(FinalsA,FinalsB,_,_,_,_,_,_,_),A/B,Bool):-
    fsa_data:indexed_final(A,FinalsA,Bool1),
    fsa_data:indexed_final(B,FinalsB,Bool2),
    and(Bool1,Bool2,Bool).

%% index_compose_transitions(+Side,+Fa,?TransThing,?EtransThing)
%% as index_transitions, except that
%% if Side=r, collect all transitions trans(P,[]/Sym,Q) into EtransThing)
%% which maps P to list of Sym-Q pairs
%% if Side=l, collect all transitions trans(P,Sym/[],Q) into EtransThing)
%% which maps P to list of Sym-Q pairs
%% futhermore, normal transition pairs $@(X)/$@(X) are replaced by $@(X)
index_compose_transitions(Dir,Fa,TT,ET):-
    fsa_data:fsa_transitions(Fa,Trans0),
    fsa_data:fsa_states_number(Fa,Size),
    collect_compose_transitions(Trans0,Dir,Trans1,E1),
    sort(E1,E2),
    fsa_array:fsa_array_new(TT,Size),
    fsa_array:fsa_array_new(ET,Size),
    fsa_data:transitions_to_thing(Trans1,TT),
    fsa_data:jumps_to_thing(E2,ET).

collect_compose_transitions([],_,[],[]).
collect_compose_transitions([Trans|TransT],Dir,T,E):-
    collect_compose_trans(Dir,Trans,T,T0,E,E0),
    collect_compose_transitions(TransT,Dir,T0,E0).

collect_compose_trans(r,trans(P,SymL/SymR,Q),T,T0,E,E0):-
    (	SymL==[]
    ->	E=[jump(P,SymR-Q)|E0],
	T=T0
    ;	sym_pair(SymL/SymR,SymPair),
	T=[trans(P,SymPair,Q)|T0],
	E=E0
    ).

collect_compose_trans(l,trans(P,SymL/SymR,Q),T,T0,E,E0):-
    (	SymR==[]
    ->	E=[jump(P,SymL-Q)|E0],
	T=T0
    ;	sym_pair(SymL/SymR,SymPair),
	T=[trans(P,SymPair,Q)|T0],
	E=E0
    ).

sym_pair(SymA/SymB,Pair):-
    (	SymA = $@(A)
    ->	Pair = $@(A)
    ;   Pair = SymA/SymB
    ).

index_w_compose_transitions(Dir,Fa,TT,ET):-
    fsa_data:fsa_transitions(Fa,Trans0),
    fsa_data:fsa_states_number(Fa,Size),
    collect_w_compose_transitions(Trans0,Dir,Trans1,E1),
    sort(E1,E2),
    fsa_array:fsa_array_new(TT,Size),
    fsa_array:fsa_array_new(ET,Size),
    fsa_data:transitions_to_thing(Trans1,TT),
    fsa_data:jumps_to_thing(E2,ET).

collect_w_compose_transitions([],_,[],[]).
collect_w_compose_transitions([Trans|TransT],Dir,T,E):-
    collect_w_compose_trans(Dir,Trans,T,T0,E,E0),
    collect_w_compose_transitions(TransT,Dir,T0,E0).

collect_w_compose_trans(r,trans(P,SymL/(SymR/W),Q),T,T0,E,E0):-
    (	SymL==[]
    ->	E=[jump(P,(SymR/W)-Q)|E0],
	T=T0
    ;	sym_triple(SymL,SymR,W,SymPair),
	T=[trans(P,SymPair,Q)|T0],
	E=E0
    ).

collect_w_compose_trans(l,trans(P,SymL/(SymR/W),Q),T,T0,E,E0):-
    (	SymR==[]
    ->	E=[jump(P,(SymL/W)-Q)|E0],
	T=T0
    ;	sym_triple(SymL,SymR,W,SymPair),
	T=[trans(P,SymPair,Q)|T0],
	E=E0
    ).

sym_triple(SymA,SymB,W,Triple):-
    (	SymA = $@(A)
    ->	Triple = $@(A)/W
    ;   Triple = (SymA/SymB)/W
    ).

macro(sigma(Set,Expr),          intersection(kleene(Set),Expr)).
macro(sigma(Domain,Range,Expr), o(o(kleene(Domain),Expr),kleene(Range))).
macro(domain_sigma(Set,Expr),   o(kleene(Set),Expr)).
macro(range_sigma(Set,Expr),    o(Expr,kleene(Set))).

%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% REVERSAL %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%

op_help(reverse,"reverse(E)",
"set of strings F such that the reversal of F is in E."). 
rx0(reverse(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_regex:reverse_fa(Fa0,Fa)).

reverse_fa(Fa0,Fa):-
    fsa_data:fsa_type(Fa0,Type),
    fsa_data:fsa_components(Syms,States,FinalStates,StartStates,Trans,Jumps,Fa0),
    reverse_or_empty(StartStates,Syms,States,FinalStates,Trans,Jumps,Type,Fa).

reverse_or_empty([],Syms,_,_,_,_,_,Fa):-
    fsa_data:fsa_components(Syms,1,[0],[],[],[],Fa).

reverse_or_empty([H|T],Syms,States,FinalStates,Trans,Jumps,Type,Fa):-
    fsa_data:fsa_components(Syms,States,[H|T],FinalStates,RevTrans,RevJumps,Fa),
    reverse_trans(Trans,RevTrans0,Type),
    sort(RevTrans0,RevTrans),
    reverse_jumps(Jumps,RevJumps0),
    sort(RevJumps0,RevJumps).

reverse_trans([],[],_Type).
reverse_trans([trans(A,B0,C)|T0],[trans(C,B,A)|T],Type) :-
    reverse_sym(Type,B0,B),
    reverse_trans(T0,T,Type).

reverse_sym(recognizer,B,B).
reverse_sym(transducer(Sub),L/B0,L/B):-
    reverse_sym0(Sub,B0,B).

reverse_sym0(letter,B,B).
reverse_sym0(sequence,B0,B):-
    reverse(B0,B).

reverse_jumps([],[]).
reverse_jumps([jump(A,C)|T0],[jump(C,A)|T]) :-
    reverse_jumps(T0,T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% INVERSE DOMAIN RANGE IDENT %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(inversion,"inversion(E)   inverse(E)   invert(E)",
"The set of pairs B:A such that A:B is in E. If E is a recognizer, then\
 \n it is converted to its identity transducer."). 
macro(inversion(A),inverse(A)).
macro(invert(A),   inverse(A)).

rx0(inverse(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_regex:inverse_fa(Type,A,Fa0,Fa)).

inverse_fa(recognizer,A,Fa,Fa) :-
    warning(warn,transducer,A).
inverse_fa(w_recognizer,A,Fa,Fa) :-
    warning(error,transducer,A).
inverse_fa(transducer(Sub),_,Fa0,Fa):-
    to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    fsa_data:fsa_copy_except(symbols,    Fa1,Fa2,t(Dl,Rl),t(Rl,Dl)),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa, Trans0,Trans),
    inverse_trans(Trans0,Trans1),
    sort(Trans1,Trans).
inverse_fa(w_transducer(Sub),_,Fa0,Fa):-
    w_to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    fsa_data:fsa_copy_except(symbols,    Fa1,Fa2,t(Dl,Rl,fsa_frozen),
			                         t(Rl,Dl,fsa_frozen)),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa, Trans0,Trans),
    inverse_w_trans(Trans0,Trans1),
    sort(Trans1,Trans).

inverse_trans([],[]).
inverse_trans([trans(P,A/B,Q)|Tail0],[trans(P,B/A,Q)|Tail]):-
    inverse_trans(Tail0,Tail).

inverse_w_trans([],[]).
inverse_w_trans([trans(P,A/(B/W),Q)|Tail0],[trans(P,B/(A/W),Q)|Tail]):-
    inverse_w_trans(Tail0,Tail).

op_help(identity,"id(E)   identity(E)",
"The set of pairs A:A such that A is in E. "). 
macro(identity(A),identity0(recognizer(A))).
macro(      id(A),identity0(recognizer(A))).

%% check recognizer?
rx0(identity0(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_regex:identity_fa(Fa0,Fa)).

identity_fa(Fa0,Fa):-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa1,r(Sigma),t(Sigma,Sigma)),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    identity_transitions(Trans0,Trans1,Sigma),
    sort(Trans1,Trans).  %% sorting is required because $@ might alter order

identity_transitions([],[],_).
identity_transitions([trans(P,Sym0,Q)|T0],
		     [trans(P,Sym/Sym,Q)|T],Module):-
    fsa_pred_module:identity(Module,Sym0,Sym),
    identity_transitions(T0,T,Module).
    
op_help(domain,"domain(E)",
"The set of strings A such A:B is in E."). 

rx0(domain(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_regex:domain_fa(Type,A,Fa0,Fa)).

domain_fa(Fa0,Fa) :-
    fsa_data:fsa_type(Fa0,Type),
    domain_fa(Type,fa(Fa0),Fa0,Fa).

domain_fa(recognizer,A,Fa,Fa) :-
    warning(warn,transducer,A).
domain_fa(transducer(_),_,Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,    Fa0,Fa1,t(Sigma,_),r(Sigma)),
    fsa_data:fsa_copy_except(jumps,      Fa1,Fa2,Jumps0,Jumps),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa, Trans0,Trans),
    domain_transitions(Trans0,Trans1,Jumps1,Jumps0),
    sort(Trans1,Trans),
    sort(Jumps1,Jumps).
domain_fa(w_recognizer,_,Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,    Fa0,Fa1,r(Sigma,_),r(Sigma)),
    fsa_data:fsa_copy_except(jumps,      Fa1,Fa2,Jumps0,Jumps),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa, Trans0,Trans),
    domain_transitions(Trans0,Trans1,Jumps1,Jumps0),
    sort(Trans1,Trans),
    sort(Jumps1,Jumps).
domain_fa(w_transducer(_),_,Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,    Fa0,Fa1,t(Sigma,_,_),r(Sigma)),
    fsa_data:fsa_copy_except(jumps,      Fa1,Fa2,Jumps0,Jumps),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa, Trans0,Trans),
    domain_transitions(Trans0,Trans1,Jumps1,Jumps0),
    sort(Trans1,Trans),
    sort(Jumps1,Jumps).

domain_transitions([],[],J,J).
domain_transitions([trans(P,Left/_,Q)|T0],T,J0,J):-
    (	Left==[]
    ->	J0=[jump(P,Q)|J1],
	T=T1
    ;   Left = '$@'(Left0)
    ->  J0=J1,
	T=[trans(P,Left0,Q)|T1]
    ;	J0=J1,
	T=[trans(P,Left,Q)|T1]
    ),
    domain_transitions(T0,T1,J1,J).

op_help(weighted_domain,"weighted_domain(E)",
"The set of weighted strings A::B such A:C:B is in E."). 

rx0(weighted_domain(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_regex:weighted_domain_fa(Type,A,Fa0,Fa)).

weighted_domain_fa(recognizer,A,Fa,Fa) :-
    warning(warn,weighted_transducer,A).
weighted_domain_fa(transducer(_),A,Fa0,Fa) :-
    warning(warn,weighted_transducer,A),
    domain_fa(Fa0,Fa).
weighted_domain_fa(w_recognizer,A,Fa,Fa) :-
    warning(warn,weighted_transducer,A).
weighted_domain_fa(w_transducer(_),_,Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,    Fa0,Fa1,t(Sigma,_,SW),r(Sigma,SW)),
    fsa_data:fsa_copy_except(jumps,      Fa1,Fa2,Jumps0,Jumps),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa, Trans0,Trans),
    weighted_domain_transitions(Trans0,Trans1,Jumps1,Jumps0),
    sort(Trans1,Trans),
    sort(Jumps1,Jumps).

weighted_domain_transitions([],[],J,J).
weighted_domain_transitions([trans(P,Left/(_/W),Q)|T0],T,J0,J):-
    (	Left==[], W =:= 0
    ->	J0=[jump(P,Q)|J1],
	T=T1
    ;   Left = '$@'(Left0)
    ->  J0=J1,
	T=[trans(P,Left0/W,Q)|T1]
    ;	J0=J1,
	T=[trans(P,Left/W,Q)|T1]
    ),
    weighted_domain_transitions(T0,T1,J1,J).

op_help(range,"range(E)",
"The set of strings B such that A:B is in E; if E is a weighted transducer\
 \n then we obtain the set of strings B such that A:B:W in E."). 

op_help(weighted_range,"weighted_range(E)",
"The set of weighted strings B:W such that A:B:W is in E."). 

rx0(range(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_regex:range_fa(Type,A,Fa0,Fa)).

range_fa(recognizer,A,Fa,Fa) :-
    warning(warn,transducer,A).
range_fa(w_recognizer,A,_,_) :-
    warning(error,transducer,A).
range_fa(w_transducer(Sub),_,Fa0,Fa):-
    fsa_data:fsa_copy_except(states,     Fa0,Fa1,N0,N),
    fsa_data:fsa_copy_except(symbols,    Fa1,Fa2,t(_,Sigma,_),r(Sigma)),
    fsa_data:fsa_copy_except(jumps,      Fa2,Fa3,Jumps0,Jumps),
    fsa_data:fsa_copy_except(transitions,Fa3,Fa, Trans0,Trans),
    w_range_transitions(Trans0,Sub,Trans1,Jumps1,Jumps0,N0,N),
    sort(Trans1,Trans),
    sort(Jumps1,Jumps).
range_fa(transducer(Sub),_,Fa0,Fa):-
    fsa_data:fsa_copy_except(states,     Fa0,Fa1,N0,N),
    fsa_data:fsa_copy_except(symbols,    Fa1,Fa2,t(_,Sigma),r(Sigma)),
    fsa_data:fsa_copy_except(jumps,      Fa2,Fa3,Jumps0,Jumps),
    fsa_data:fsa_copy_except(transitions,Fa3,Fa, Trans0,Trans),
    range_transitions(Trans0,Sub,Trans1,Jumps1,Jumps0,N0,N),
    sort(Trans1,Trans),
    sort(Jumps1,Jumps).

range_transitions([],_,[],J,J,N,N).
range_transitions([trans(P,_/Right,Q)|T0],Sub,T,J0,J,N0,N):-
    range_transition(Sub,Right,P,Q,T,T1,J0,J1,N0,N1),
    range_transitions(T0,Sub,T1,J1,J,N1,N).

range_transition(letter,Right,P,Q,T,T1,J0,J1,N,N) :-
    (	Right==[]
    ->	J0=[jump(P,Q)|J1],
	T=T1
    ;   Right = '$@'(Right0)
    ->	J0=J1,
	T=[trans(P,Right0,Q)|T1]
    ;	J0=J1,
	T=[trans(P,Right,Q)|T1]
    ).

range_transition(sequence,Right,P,Q,T,T1,J0,J1,N0,N) :-
    range_transition_seq(Right,P,Q,T,T1,J0,J1,N0,N).

range_transition_seq([],P,Q,T,T,[jump(P,Q)|J],J,N,N).
range_transition_seq([H|Tail],P,Q,T0,T,J0,J,N0,N) :-
    range_transition_seq(Tail,H,P,Q,T0,T,J0,J,N0,N).

range_transition_seq([],Sym0,P,Q,[trans(P,Sym,Q)|T],T,J,J,N,N) :-
    (	Sym0 = '$@'(Sym1)
    ->	Sym = Sym1
    ;   Sym = Sym0
    ).
range_transition_seq([H|Tail],Sym0,P,Q,[trans(P,Sym,N0)|T0],T,J0,J,N0,N):-
    N1 is N0+1,
    (	Sym0 = '$@'(Sym1)
    ->	Sym1 = Sym
    ;	Sym0 = Sym
    ),
    range_transition_seq(Tail,H,N0,Q,T0,T,J0,J,N1,N).
    
w_range_transitions([],_,[],J,J,N,N).
w_range_transitions([trans(P,_/(Right/_),Q)|T0],Sub,T,J0,J,N0,N):-
    range_transition(Sub,Right,P,Q,T,T1,J0,J1,N0,N1),
    w_range_transitions(T0,Sub,T1,J1,J,N1,N).

rx0(weighted_range(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_regex:weighted_range_fa(Type,A,Fa0,Fa)).

weighted_range_fa(recognizer,A,Fa,Fa) :-
    warning(error,weighted_transducer,A).
weighted_range_fa(w_recognizer,A,_,_) :-
    warning(error,weighted_transducer,A).
weighted_range_fa(transducer(_),A,_,_) :-
    warning(error,weighted_transducer,A).
weighted_range_fa(w_transducer(letter),_,Fa0,Fa):-
    fsa_data:fsa_copy_except(states,     Fa0,Fa1,N0,N),
    fsa_data:fsa_copy_except(symbols,    Fa1,Fa2,t(_,Sigma,SW),r(Sigma,SW)),
    fsa_data:fsa_copy_except(jumps,      Fa2,Fa3,Jumps0,Jumps),
    fsa_data:fsa_copy_except(transitions,Fa3,Fa, Trans0,Trans),
    weighted_range_transitions(Trans0,letter,Trans1,Jumps1,Jumps0,N0,N),
    sort(Trans1,Trans),
    sort(Jumps1,Jumps).

weighted_range_transitions([],_,[],J,J,N,N).
weighted_range_transitions([trans(P,_/(Right/W),Q)|T0],Sub,T,J0,J,N0,N):-
    weighted_range_transition(Sub,Right,W,P,Q,T,T1,J0,J1,N0,N1),
    weighted_range_transitions(T0,Sub,T1,J1,J,N1,N).

weighted_range_transition(letter,Right,W,P,Q,T,T1,J0,J1,N,N) :-
    (	Right==[],
	W =:= 0
    ->	J0=[jump(P,Q)|J1],
	T=T1
    ;   Right = '$@'(Right0)
    ->	J0=J1,
	T=[trans(P,Right0/W,Q)|T1]
    ;	J0=J1,
	T=[trans(P,Right/W,Q)|T1]
    ).

op_help(weights,"weights(E)",
"A recognizer for the weights as defined in the weighed recognizer or\
 \n weighted transducer E.").

rx0(weights(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_regex:weights_fa(Type,A,Fa0,Fa)).

weights_fa(recognizer,A,_,_) :-
    warning(error,weighted_automaton,A).
weights_fa(transducer(_),A,_,_) :-
    warning(error,weighted_automaton,A).
weights_fa(w_transducer(_Sub),_,Fa0,Fa):-
    fsa_data:fsa_copy_except(symbols, Fa0,Fa1,t(_,_,fsa_frozen),r(fsa_frozen)),
    fsa_data:fsa_copy_except(jumps,      Fa1,Fa2,Jumps0,Jumps),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa, Trans0,Trans),
    w_weights_transitions(Trans0,Trans1,Jumps1,Jumps0),
    sort(Trans1,Trans),
    sort(Jumps1,Jumps).
weights_fa(w_recognizer,_,Fa0,Fa):-
    fsa_data:fsa_copy_except(symbols, Fa0,Fa1,r(_,fsa_frozen),r(fsa_frozen)),
    fsa_data:fsa_copy_except(jumps,      Fa1,Fa2,Jumps0,Jumps),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa, Trans0,Trans),
    weights_transitions(Trans0,Trans1,Jumps1,Jumps0),
    sort(Trans1,Trans),
    sort(Jumps1,Jumps).

weights_transitions([],[],J,J).
weights_transitions([trans(P,_/Right,Q)|T0],T,J0,J):-
    (	Right=:=0
    ->	J0=[jump(P,Q)|J1],
	T=T1
    ;	J0=J1,
	T=[trans(P,Right,Q)|T1]
    ),
    weights_transitions(T0,T1,J1,J).

w_weights_transitions([],[],J,J).
w_weights_transitions([trans(P,_/(_/Right),Q)|T0],T,J0,J):-
    (	Right=:=0
    ->	J0=[jump(P,Q)|J1],
	T=T1
    ;	J0=J1,
	T=[trans(P,Right,Q)|T1]
    ),
    w_weights_transitions(T0,T1,J1,J).


op_help(no_weights,"no_weights(E)",
"If E is a weighted recognizer, then this defines the recognizer obtained\
 \n by ignoring all weights. If E is a weighted transducer, then this defines\
 \n a transducer obtained by ignoring all weights.").

rx0(no_weights(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_regex:no_weights_fa(Type,A,Fa0,Fa)).

no_weights_fa(recognizer,A,_,_) :-
    warning(error,weighted_automaton,A).
no_weights_fa(transducer(_),A,_,_) :-
    warning(error,weighted_automaton,A).
no_weights_fa(w_transducer(_Sub),_,Fa0,Fa):-
    fsa_data:fsa_copy_except(symbols, Fa0,Fa1,t(Ma,Mb,fsa_frozen),t(Ma,Mb)),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa, Trans0,Trans),
    w_no_weights_transitions(Trans0,Trans1),
    sort(Trans1,Trans).
no_weights_fa(w_recognizer,_,Fa0,Fa):-
    fsa_data:fsa_copy_except(symbols, Fa0,Fa1,r(Ma,fsa_frozen),r(Ma)),
    fsa_data:fsa_copy_except(jumps,      Fa1,Fa2,Jumps0,Jumps),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa, Trans0,Trans),
    no_weights_transitions(Trans0,Trans1,Jumps1,Jumps0),
    sort(Trans1,Trans),
    sort(Jumps1,Jumps).

no_weights_transitions([],[],J,J).
no_weights_transitions([trans(P,Right/_,Q)|T0],T,J0,J):-
    (	Right==[]
    ->	J0=[jump(P,Q)|J1],
	T=T1
    ;	J0=J1,
	T=[trans(P,Right,Q)|T1]
    ),
    no_weights_transitions(T0,T1,J1,J).

w_no_weights_transitions([],[]).
w_no_weights_transitions([trans(P,L/(Right/_),Q)|T0],[trans(P,L/Right,Q)|T]):-
    w_no_weights_transitions(T0,T).


%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% CLEAN UP %%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(cleanup,"cleanup(E)",
"The cleanup operator attempts to pack several transitions into\
 \n one. For instance, assume there are two transitions from state p to q\
 \n over the predicates p1 and p2 respectively. If p3 is a predicate which\
 \n is true just in case either p1 or p2 is true, then we replace the two\
 \n transitions by one transition over predicate p3. Note that if E is\
 \n a sequence transducer, then cleanup does not attempt to create identities\
 \n or combine one or more transitions involving identities (since that\
 \n would require global analysis in order to know how identities on input\
 \n and output side line up)."). 

rx0(cleanup(Expr),Fa,Debug,Cache,Level) :-
    rx(Expr,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_cleanup:cleanup_fa(Fa0,Fa)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% EXPAND_PREDICATES %%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(expand_non_overlapping_predicates,
"expand_non_overlapping_predicates(E)",
"This operator constructs an automaton M for the expression E in such\
 \n a way that all predicates which occur in M have an empty intersection,\
 \n i.e. the resulting predicates in M are non-overlapping. Note that\
 \n in the case of a sequence transducer, M is first coerced into a\
 \n (synchronized) letter transducer.").

rx0(expand_non_overlapping_predicates(A),Fa,Debug,Cache,Level) :-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_cleanup:expand_non_overlapping_fa(Type,Fa0,Fa)).

%% a mess...; only used in fsa_fl; perhaps move it there?

%% letter_transducer / recognizer.
%% result is of pred_module fsa_frozen
rx0(expand_predicates(A),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_regex:expand_predicates_fa(Fa0,Fa,_)).

rx0(expand_predicates(Symbol,A),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_regex:expand_predicates_fa(Fa0,Fa,Symbol)).

expand_predicates_fa(Fa0,Fa,DefaultSymbol):-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa1,Mods0,Mods),
    fsa_data:fsa_type(Fa0,Type),
    new_symbols_decl(Type,Mods,Fa1,Fa2),
    (	Mods0==Mods
    ->	Fa2=Fa
    ;	fsa_data:fsa_copy_except(transitions,Fa2,Fa,Trans0,Trans),
	expand_transitions(Trans0,Trans1,Mods0,DefaultSymbol),
	sort(Trans1,Trans)
    ).

new_symbols_decl(recognizer,r(fsa_frozen),Fa,Fa).
new_symbols_decl(transducer(Sub),t(fsa_frozen,fsa_frozen),Fa0,Fa) :-
    to_letter_if_sequence(Sub,silent,_,Fa0,Fa).
new_symbols_decl(w_recognizer,r(fsa_frozen,fsa_frozen),Fa,Fa).
new_symbols_decl(w_transducer(Sub),t(fsa_frozen,fsa_frozen,fsa_frozen),F0,F) :-
    w_to_letter_if_sequence(Sub,silent,_,F0,F).

expand_transitions([],[],_,_Ds).
expand_transitions([H|T],Tr,Mods,Ds):-
    expand_transition(Mods,H,Tr,Tr1,Ds),
    expand_transitions(T,Tr1,Mods,Ds).

expand_transition(r(Mod),trans(P,Pred,Q),Trans,Tail,Ds):-
    findall(trans(P,Sym,Q),eval_pred(Mod,Pred,Sym,Ds),Trans,Tail).

expand_transition(r(Mod,fsa_frozen),trans(P,Pred/W,Q),Trans,Tail,Ds):-
    findall(trans(P,Sym/W,Q),eval_pred(Mod,Pred,Sym,Ds),Trans,Tail).

expand_transition(t(ModD,ModR),trans(P,PredD/PredR,Q),Trans,Tail,Ds):-
    findall(trans(P,SymD/SymR,Q),
	    eval_pred_pair(ModD,ModR,PredD,PredR,SymD,SymR,Ds),
	    Trans,Tail).

expand_transition(t(ModD,ModR,fsa_frozen),trans(P,PredD/(PredR/W),Q),Trans,Tail,Ds):-
    findall(trans(P,SymD/(SymR/W),Q),
	    eval_pred_pair(ModD,ModR,PredD,PredR,SymD,SymR,Ds),
	    Trans,Tail).

eval_pred_pair(Mod,Mod,'$@'(Pred),'$@'(Pred),Sym,Sym,Ds):-
    !,
    fsa_pred_module:evaluate_predicate(Mod,Pred,Sym),
    (	var(Sym)
    ->	Sym=Ds
    ;	true
    ).

eval_pred_pair(ModD,ModR,PredD,PredR,SymD,SymR,Ds):-
    (	PredD==[]
    ->  SymD=[]
    ;   eval_pred(ModD,PredD,SymD,Ds)
    ),
    (	PredR==[]
    ->  SymR=[]
    ;   eval_pred(ModR,PredR,SymR,Ds)
    ).

eval_pred(Mod,Pred,Sym,Ds) :-
    fsa_pred_module:evaluate_predicate(Mod,Pred,Sym),
    (	var(Sym)
    ->	Sym=Ds
    ;	true
    ),
    warn_if_variable_symbol(Sym,Pred).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%% SUBS DEL INS %%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(subs,"subs(E)",
"E is supposed to be a transducer, or weighted recognizer. The result will be \
 \n all pairs allowed by E and furthermore all pairs (x,y) such that (x',y) is \
 \n in E and x' can be formed by substituting one symbol in x.").
op_help(del,"del(E)",
"E is supposed to be a transducer, or weighted recognizer. The result will be\
 \n all pairs allowed by E and furthermore all pairs (x,y) such that (x',y) is \
 \n in E and x' can be formed by deleting one symbol in x.").
op_help(ins,"ins(E)",
"E is supposed to be a transducer, or weighted recognizer. The result will be\
 \n all pairs allowed by E and furthermore all pairs (x,y) such that (x',y) is \
 \n in E and x' can be formed by inserting one symbol in x.").

rx0(subs(A),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_regex:subs_fa(Fa0,Fa)).

rx0(del(A),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_regex:del_fa(Fa0,Fa)).

rx0(ins(A),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    ltime(Debug,fsa_regex:ins_fa(Fa0,Fa)).

ins_fa(Fa0,Fa) :-
    fsa_data:fsa_type(Fa0,Type),
    ins_fa(Type,Fa0,Fa).

ins_fa(recognizer,Fa0,Fa):-
    identity_fa(Fa0,Fa1),
    ins_fa(Fa1,Fa).
ins_fa(w_recognizer,Fa0,Fa) :-
    all_components_reachable(ins,Fa0,Fa).
ins_fa(transducer(Sub),Fa0,Fa) :-
    to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    all_components_reachable(ins,Fa1,Fa).
ins_fa(w_transducer(Sub),Fa0,Fa) :-
    w_to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    all_components_reachable(ins,Fa1,Fa).

del_fa(Fa0,Fa) :-
    fsa_data:fsa_type(Fa0,Type),
    del_fa(Type,Fa0,Fa).

del_fa(recognizer,Fa0,Fa):-
    identity_fa(Fa0,Fa1),
    del_fa(Fa1,Fa).
del_fa(w_recognizer,Fa0,Fa) :-
    all_components_reachable(del,Fa0,Fa).
del_fa(transducer(Sub),Fa0,Fa) :-
    to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    all_components_reachable(del,Fa1,Fa).
del_fa(w_transducer(Sub),Fa0,Fa) :-
    w_to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    all_components_reachable(del,Fa1,Fa).

subs_fa(Fa0,Fa) :-
    fsa_data:fsa_type(Fa0,Type),
    subs_fa(Type,Fa0,Fa).

subs_fa(recognizer,Fa0,Fa):-
    identity_fa(Fa0,Fa1),
    subs_fa(Fa1,Fa).
subs_fa(w_recognizer,Fa0,Fa) :-
    all_components_reachable(subs,Fa0,Fa).
subs_fa(transducer(Sub),Fa0,Fa) :-
    to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    all_components_reachable(subs,Fa1,Fa).
subs_fa(w_transducer(Sub),Fa0,Fa) :-
    w_to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    all_components_reachable(subs,Fa1,Fa).

prepare_reachable(subs,Fa,d(DSigma,Finals,Trans,Jumps,Decl),Starts,Decl):-
    fsa_data:fsa_symbols_decl(Fa,Decl),
    domain_sigma(Decl,DSigma),
    findall(q(Start,0),fsa_data:fsa_start_state(Fa,Start),Starts),
    fsa_data:index_finals(Fa,Finals),
    fsa_data:index_transitions(Fa,Trans),
    index_jumps_if_ne(Fa,Jumps).

is_final_reachable(subs,d(_,Finals,_,_,_),q(State,_),Bool):-
    fsa_data:indexed_final(State,Finals,Bool).

trans_reachables(subs,q(State,Int),D,New):-
    findall(Trans,subs_trans(Int,State,D,Trans),New).

subs_trans(Int,P,d(_,_,Trans,_,_),trans(q(P,Int),Sym,q(Q,Int))):-
    fsa_data:indexed_transition(P,Trans,Sym,Q).

subs_trans(0,P,d(Mod,_,Trans,_,Decl),trans(q(P,0),Qm/Out,q(Q,1))):-
    fsa_pred_module:true(Mod,Qm),
    fsa_data:indexed_transition(P,Trans,_/Out0,Q),
    rem_id(Decl,Out0,Out).

subs_trans(Int,P,d(_,_,_,jumps(Js),_),jump(q(P,Int),q(Q,Int))):-
    fsa_data:indexed_jump(P,Js,Q).
    
rem_id(t(_,_),Sym0,Sym):-
    (	Sym0 = '$@'(Sym1)
    ->	Sym1=Sym
    ;   Sym0=Sym
    ).

rem_id(t(_,_,_),Sym0/W,Sym/W):-
    (	Sym0 = '$@'(Sym1)
    ->	Sym1=Sym
    ;   Sym0=Sym
    ).

rem_id(r(_,_),Sym,Sym).


prepare_reachable(del,Fa,d(DSigma,Finals,Trans,Jumps,Decl),Starts,Decl):-
    fsa_data:fsa_symbols_decl(Fa,Decl),
    domain_sigma(Decl,DSigma),
    findall(q(Start,0),fsa_data:fsa_start_state(Fa,Start),Starts),
    fsa_data:index_finals(Fa,Finals),
    fsa_data:index_transitions(Fa,Trans),
    index_jumps_if_ne(Fa,Jumps).

is_final_reachable(del,d(_,Finals,_,_,_),q(State,_),Bool):-
    fsa_data:indexed_final(State,Finals,Bool).

trans_reachables(del,q(State,Int),D,New):-
    findall(Trans,del_trans(Int,State,D,Trans),New).

del_trans(Int,P,d(_,_,Trans,_,_),trans(q(P,Int),Sym,q(Q,Int))):-
    fsa_data:indexed_transition(P,Trans,Sym,Q).

del_trans(Int,P,d(_,_,_,jumps(Js)),jump(q(P,Int),q(Q,Int))):-
    fsa_data:indexed_jump(P,Js,Q).

del_trans(0,P,d(_,_,Trans,_,Decl),Edge):-
    fsa_data:indexed_transition(P,Trans,_/Out0,Q),
    (	Out0==[]
    ->	Edge=jump(q(P,0),q(Q,1))
    ;   rem_id(Decl,Out0,Out),
	Edge=trans(q(P,0),[]/Out,q(Q,1))
    ).
    
prepare_reachable(ins,Fa,d(DSigma,Finals,Trans,Jumps,Decl),Starts,Decl):-
    fsa_data:fsa_symbols_decl(Fa,Decl),
    domain_sigma(Decl,DSigma),
    findall(q(Start,0),fsa_data:fsa_start_state(Fa,Start),Starts),
    fsa_data:index_finals(Fa,Finals),
    fsa_data:index_transitions(Fa,Trans),
    index_jumps_if_ne(Fa,Jumps).

is_final_reachable(ins,d(_,Finals,_,_,_),State,Bool):-
    ins_final(State,Finals,Bool).

ins_final(extra(State),Finals,Bool):-
    fsa_data:indexed_final(State,Finals,Bool).
ins_final(q(State,_),Finals,Bool):-
    fsa_data:indexed_final(State,Finals,Bool).

trans_reachables(ins,State,D,New):-
    findall(Trans,ins_trans(State,D,Trans),New).

ins_trans(q(State,Int),D,Trans):-
    ins_trans0(Int,State,D,Trans).
ins_trans(extra(State),D,Trans):-
    ins_trans1(State,D,Trans).

ins_trans0(Int,P,d(_,_,Trans,_,_),trans(q(P,Int),Sym,q(Q,Int))):-
    fsa_data:indexed_transition(P,Trans,Sym,Q).

ins_trans0(Int,P,d(_,_,_,jumps(Js),_),jump(q(P,Int),q(Q,Int))):-
    fsa_data:indexed_jump(P,Js,Q).

ins_trans0(0,P,d(Mod,_,_,_,Decl),trans(q(P,0),Qm/Output,extra(P))):-
    ins_output(Decl,Output),
    fsa_pred_module:true(Mod,Qm).

%% no need to have jump here as well, we'll get these anyway
ins_trans1(P,d(_,_,Ts,_,_),trans(extra(P),Sym,q(Q,1))):-
    fsa_data:indexed_transition(P,Ts,Sym,Q).

ins_output(t(_,_),[]).
ins_output(r(_,_),0).
ins_output(t(_,_,_),[]/0).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% MISC OPERATORS %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

op_help(word,"word(Atom)",
"Denotes the string Atom, as a concatenation of its individual\
 \n characters. For instance word(regular) is equivalent to\
 \n [r,e,g,u,l,a,r]."). 
rx0(word(Atom),Fa,_Debug,_Cache,_Level):-
    atom(Atom),
    atom_codes(Atom,String),
    string_to_atoms(String,Atoms),
    fsa_globals:fsa_global_get(pred_module,Module),
    word_transitions(Atoms,Trans,0,Len0,Module),
    Len is Len0+1,
    fsa_data:fsa_construct(r(Module),Len,[0],[Len0],Trans,[],Fa).

word_transitions([],[],L,L,_).
word_transitions([H|T],[trans(L0,Pred,L1)|T2],L0,L,Module):-
    fsa_pred_module:regex_notation_to_predicate(Module,H,Pred),
    L1 is L0+1,
    word_transitions(T,T2,L1,L,Module).

string_to_atoms([],[]).
string_to_atoms([Int|Ints],[Atom|Atoms]):-
    atom_codes(Atom,[Int]),
    string_to_atoms(Ints,Atoms).

%% convert_pred_module(NewModule,Expr)
op_help(convert_pred_module,"convert_pred_module(NewModule,Expr)\
 \n convert_pred_module(NewDomainMod,NewRangeMod,Expr)",
"Converts the automaton defined by Expr into an automaton using the \
 \n pred_module declarations found in NewModule. Note that this is\
 \n possible only in case the newer module is at least as expressive as\
 \n the old one. For instance, you can convert an automaton with the\
 \n *fsa_frozen* predicate module into an automaton with the *fsa_preds*\
 \n predicate module, but not vice versa. The binary operator is for\
 \n recognizers, the ternary operator for letter transducers.").

rx0(converse_fl(Expr),Fa,Debug,Cache,Level) :-
    rx(Expr,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    fsa_fl:converse_fa(on,Type,Fa0,Fa,_).

rx0(convert_pred_module(New,Expr),Fa,Debug,Cache,Level):-
    rx(Expr,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_regex:convert_pred_module(Type,Expr,New,Fa0,Fa)).

convert_pred_module(recognizer,_,New,Fa0,Fa):-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa1,r(Old),r(New)),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    convert_pred_trans(Trans0,Trans1,Old,New),
    sort(Trans1,Trans).

convert_pred_module(w_recognizer,_,New,Fa0,Fa):-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa1,r(Old,FF),r(New,FF)),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    convert_pred_trans_w(Trans0,Trans1,Old,New),
    sort(Trans1,Trans).

convert_pred_module(transducer(Sub),Expr,New,Fa0,Fa) :-
    convert_pred_module(transducer(Sub),Expr,New,New,Fa0,Fa).

convert_pred_module(w_transducer(Sub),Expr,New,Fa0,Fa) :-
    convert_pred_module(w_transducer(Sub),Expr,New,New,Fa0,Fa).

rx0(convert_pred_module(NewD,NewR,Expr),Fa,Debug,Cache,Level):-
    rx(Expr,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    ltime(Debug,fsa_regex:convert_pred_module(Type,Expr,NewD,NewR,Fa0,Fa)).

convert_pred_module(recognizer,A,_,_,_,_):-
    warning(error,transducer,A).
convert_pred_module(w_recognizer,A,_,_,_,_):-
    warning(error,transducer,A).
convert_pred_module(transducer(Sub),_,NewD,NewR,Fa0,Fa):-
    to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    fsa_data:fsa_copy_except(symbols,Fa1,Fa2,t(OldD,OldR),t(NewD,NewR)),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa,Trans0,Trans),
    convert_pred_trans_t(Trans0,Trans1,OldD,OldR,NewD,NewR),
    sort(Trans1,Trans).
convert_pred_module(w_transducer(Sub),_,NewD,NewR,Fa0,Fa):-
    w_to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    fsa_data:fsa_copy_except(symbols,Fa1,Fa2,t(OldD,OldR,FF),t(NewD,NewR,FF)),
    fsa_data:fsa_copy_except(transitions,Fa2,Fa,Trans0,Trans),
    convert_pred_trans_wt(Trans0,Trans1,OldD,OldR,NewD,NewR),
    sort(Trans1,Trans).

convert_pred_trans([],[],_,_).
convert_pred_trans([trans(P,Sym0,Q)|T0],Tr,Old,New):-
    findall(trans(P,Sym,Q),
	    convert_pred_sym(Sym0,Sym,Old,New),Tr,Tr0),
    convert_pred_trans(T0,Tr0,Old,New).

convert_pred_trans_w([],[],_,_).
convert_pred_trans_w([trans(P,Sym0/W,Q)|T0],Tr,Old,New):-
    findall(trans(P,Sym/W,Q),
	    convert_pred_sym(Sym0,Sym,Old,New),Tr,Tr0),
    convert_pred_trans_w(T0,Tr0,Old,New).

convert_pred_trans_t([],[],_,_,_,_).
convert_pred_trans_t([trans(P,SymD00/SymR00,Q)|T0],Tr,OldD,OldR,
		     NewD,NewR):-
    (	SymD00 = '$@'(SymD0),
	SymR00 = '$@'(SymR0)
    ->	SymD=SymR
    ;   SymD00 = SymD0,
	SymR00 = SymR0
    ), 
    findall(trans(P,SymD/SymR,Q),
	    (	convert_pred_sym(SymD0,SymD,OldD,NewD),
		convert_pred_sym(SymR0,SymR,OldR,NewR)
	    ), Tr,Tr0),
    convert_pred_trans_t(T0,Tr0,OldD,OldR,NewD,NewR).

convert_pred_trans_wt([],[],_,_,_,_).
convert_pred_trans_wt([trans(P,SymD00/(SymR00/W),Q)|T0],Tr,OldD,OldR,
		     NewD,NewR):-
    (	SymD00 = '$@'(SymD0),
	SymR00 = '$@'(SymR0)
    ->	SymD=SymR
    ;   SymD00 = SymD0,
	SymR00 = SymR0
    ), 
    findall(trans(P,SymD/(SymR/W),Q),
	    (	convert_pred_sym(SymD0,SymD,OldD,NewD),
		convert_pred_sym(SymR0,SymR,OldR,NewR)
	    ), Tr,Tr0),
    convert_pred_trans_t(T0,Tr0,OldD,OldR,NewD,NewR).

convert_pred_sym(Sym0,Sym,Old,New):-
    (	Sym0 == []
    ->	Sym = []
    ;	fsa_pred_module:evaluate_predicate(Old,Sym0,Sym1),
	fsa_pred_module:regex_notation_to_predicate(New,Sym1,Sym)
    ),
    warn_if_variable_symbol(Sym,Sym0).

op_help(fa,"fa(Fa)",
"Fa is already a finite automaton in appropriate format."). 
rx0(fa(Fa),Fa,_Debug,_Cache,_Level).

op_help(file,"file(X)",
"This denotes the finite-automaton read from file X."). 
rx0(file(X),Fa,_,_,_) :-
    fsa_io:fsa_read_file(X,Fa).

rx0(flag(_),Fa,_,_,_):-
    fsa_globals:fsa_global_get(fa,Fa).

op_help(spy,"spy(Expr)",
"Equivalent to Expr, but sets spy-point on compilation of Expr. This\
 \n implies that for debug level 1 or higher the CPU-time is reported\
 \n required to compile Expr, as well as the size of the resulting \
 \n automaton.").
macro(spy(X),X).

op_help(cache,"cache(Expr)",
"Equivalent to Expr, but caches result of compiling Expr, if the flag\
 \n *regex_cache* is set to *selective*. If that flag has value *off* then\
 \n there is no caching. If the value is *on* then the regular expression \
 \n compiler caches every sub-computation."). 
macro(cache(X),X).

%% only used internally; not documented
macro(freeze(A),freeze(A,_)).
rx0(freeze(A,Sigma0),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_type(Fa0,Type),
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,Sigma0,Sigma),
    freeze_if_transducer(Type,Sigma0,Sigma).

freeze_if_transducer(recognizer,S,S).
freeze_if_transducer(w_recognizer,_,r(fsa_frozen)).
freeze_if_transducer(transducer(_),_,r(fsa_frozen)).
freeze_if_transducer(w_transducer(_),_,r(fsa_frozen)).

%% only used internally; not documented
rx0(melt(A,Sigma1),Fa,Debug,Cache,Level):-
    rx(A,Fa0,Debug,Cache,Level),
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,_,Sigma1).

op_help(random,"random(NrStates,NrSymbols,Den,JDens[,FDens])",
"A random non-deterministic automaton is constructed consisting of\
 \n the number of states specified in the first argument, number of\
 \n symbols in the second argument. The desired density of the automaton \
 \n is given in the third argument, whereas the fourth argument is the \
 \n jump density. The final argument is a number between 0 and 1 indicating \
 \n the likelihood that a state is final. If no fifth argument is specified, \
 \n then all states are final.\
 \n \
 \n For example, random(20,10,0.1,0.1) will be an automaton with 20 states,\
 \n 10 symbols, approximately 400 transitions and 40 jumps. Automata generated\
 \n this way always contain a single start state. They use the fsa_frozen\
 \n predicate module; transition labels are integers counting from 0 upward.\
 \n \
 \n If you need deterministic automata, consider the det_random/[4,5]\
 \n regular expression operators.").

op_help(det_random,"det_random(NrStates,NrSymbols,Den,FDens)",
"A deterministic random automaton is constructed consisting of the number of\
 \n states specified in the first argument, number of symbols in the\
 \n second argument. The desired density of the automaton is given in the \
 \n third argument, whereas the final argument is a number between\
 \n 0 and 1 indicating the likelihood that a state is final.\
 \n For instance, det_random(20,10,0.1,0.1) will be an automaton with 20 states,\
 \n (at most) 10 symbols, approximately 20 transitions, and approximately\
 \n 2 final states.  Note that the generated automaton may contain states \
 \n which are unreachable from the start state. Automata generated\
 \n this way always contain a single start state. They use the fsa_frozen\
 \n predicate module; transition labels are integers counting from 0 upward.\
 \n \
 \n If you need automata in which all states are reachable, consider the\
 \n reachable/1 regular expression operator.").

macro(random(IntStates,IntSymbols,Dens,JDens),
      random(IntStates,IntSymbols,Dens,JDens,1)).

rx0(random(IntStates,IntSymbols,Dens,JDens,FDens),Fa,_,_,_):-
    fsa_random(IntStates,IntSymbols,Dens,JDens,FDens,Fa).

rx0(random,Fa,_,_,_):-
    fsa_random(Fa).

rx0(det_random(IntStates,IntSymbols,Tdens,Fdens),Fa,_,_,_):-
    det_random(IntStates,IntSymbols,Tdens,Fdens,Fa).

%% not documented
rx0(det_random,Fa,_,_,_):-
    det_random(Fa).


%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% pragma %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% pragma(X,Y) is equivalent to Y, except it's supposed to be more efficient.
%% ListOfVars is a list of terms occurring in Macro0.
%% Each of these is evaluated with rx/2. Macro is the same as Macro0,
%% except that each occurrence of a term in ListOfTerms is replaced
%% by fa(X) where X is the corresponding result of rx/2.

op_help(pragma,"pragma(ListOfExpressions,E)",
"Equivalent to E, but this expression instructs the regular expression\
 \n compiler that it should first compile each of the expressions in the list\
 \n ListOfExpressions, which are supposed to occur repetitively in E. This is\
 \n typically used in macros which use sub-expressions more than once. For\
 \n example, consider the following macro (adapted from Kaplan and Kay):\
 \n \
 \n     macro(p_iff_s(L1,L2),\
 \n          if_p_then_s(L1,L2) & if_s_then_p(L1,L2)). \
 \n \
 \n The pragma operator can be used in order to ensure that L1 and L2 are\
 \n compiled only once:\
 \n \
 \n     macro(p_iff_s(L1,L2), pragma([V1-L1,V2-L2],\
 \n          if_p_then_s(V1,V2) & if_s_then_p(V1,V2))). \
 \n \
 \n The first argument of pragma thus is a list of pairs V-Term where V is\
 \n a variable which occurs in the second argument (in typical cases it occurs\
 \n at least twice). The Term associated with V is compiled only once. For\
 \n every occurrence of V in the second argument, the result of that compilation\
 \n is used.       ").


macro(bind(_Var,Term),Term).

rx0(pragma(ListOfVars, Macro0), Fa, D,C,L) :-
    (	list_of_pairs(ListOfVars)
    ->	true
    ;   user:throw(fsa_error(
	 "First argument of pragma/2 should be a key-value list (~w)~n",
			     [ListOfVars]))
    ),
    rx_pair_list(ListOfVars,ListOfFa,D,C,L),
    replace_term(Macro0,Macro,ListOfFa),
    rx(Macro,Fa,D,C,L).


rx_pair_list([],[],_,_,_).
rx_pair_list([Var-H0|T0],[Var-H|T],Debug,Cache,Level):-
    rx(H0,H,Debug,Cache,Level),
    rx_pair_list(T0,T,Debug,Cache,Level).

list_of_pairs([A-_B|T]) :-
    var(A),
    list_of_pairs0(T).

list_of_pairs0([]).
list_of_pairs0([A-_B|T]) :-
    var(A),
    list_of_pairs0(T).

is_in_pair_list(M0,[M-Fa|Tail],Fa0) :-
    (	M0 == M
    ->  Fa0=Fa
    ;	is_in_pair_list(M0,Tail,Fa0)
    ).

replace_term(M0,M,List) :-
    (	var(M0)
    ->	(   is_in_pair_list(M0,List,Fa)
	->  M=fa(Fa)
	;   M0=M
	)
    ;	functor(M0,F,A),
	functor(M, F,A),
	replace_term(A,M0,M,List)
    ).

replace_term(N0,Term0,Term,Fas) :-
    (	N0 =< 0
    ->	true
    ;   arg(N0,Term0,Arg0),
	arg(N0,Term,Arg),
	replace_term(Arg0,Arg,Fas),
	N is N0-1,
	replace_term(N,Term0,Term,Fas)
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%% UTILS %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%

coerce_sigma(FaA,FaB,Sigma):-
    fsa_data:fsa_symbols_decl(FaA,Sigma0),
    fsa_data:fsa_symbols_decl(FaB,Sigma),
    (	Sigma0==Sigma
    ->	true
    ;   user:throw(fsa_error(
   "Cannot coerce automata with different predicate modules (~w and ~w)~n",
				  [Sigma0,Sigma])
		       )
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% All components accessible %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

all_components_reachable(Op,Fa0,Fa1,Fa2,Fa):-
    prepare_reachable(Op,Fa0,Fa1,Fa2,Data,Starts,Sigma),
    all_components_reachable0(Op,Data,Starts,Sigma,Fa).

all_components_reachable(Op,Fa0,Fa1,Fa):-
    prepare_reachable(Op,Fa0,Fa1,Data,Starts,Sigma),
    all_components_reachable0(Op,Data,Starts,Sigma,Fa).

all_components_reachable(Op,Fa0,Fa) :- 
    prepare_reachable(Op,Fa0,Data,Starts,Sigma),
    all_components_reachable0(Op,Data,Starts,Sigma,Fa).

all_components_reachable0(Op,Data,Starts0,Sigma,Fa) :-
    fsa_hash:fsa_hash_new(Hash),
    add_starts(Starts0,Starts,Agenda,Hash,0,Size0),
    reachable(Agenda,Hash,Op,Data,Finals,Transs,Jumps,Size0,Size),
    fsa_data:fsa_construct(Sigma,Size,Starts,Finals,Transs,Jumps,Fa).

reachable([],_Hash,_Op,_Data,[],[],[],S,S).
reachable([St/StNr|T],Hash,Op,Data,Fin0,Trn0,Jmp0,Nr0,Nr):-
    is_final_reachable(Op,Data,St,Bool),
    add_final_reachable(Bool,StNr,Fin0,Fin1),
    sorted_trans_reachables(Op,St,Data,New1),
    add_news(New1,Ag,T,Hash,Nr0,Nr1,Trn0,Trn1,Jmp0,Jmp1,StNr),
    reachable(Ag,Hash,Op,Data,Fin1,Trn1,Jmp1,Nr1,Nr).

sorted_trans_reachables(Op,St,Data,New) :-
    trans_reachables(Op,St,Data,New0),
    %% sorting these should ensure result is canonical
    %% at least for reachability of deterministic automaton
    remove_source_states(New0,New1),
    sort(New1,New).

remove_source_states([],[]).
remove_source_states([H|T],[NH|NT]) :-
    remove_source_state(H,NH),
    remove_source_states(T,NT).

remove_source_state(jump(_,Q),jump(Q)).
remove_source_state(trans(_,Sym,Q),trans(Sym,Q)).

add_news([],Ag,Ag,_Hash,Nr,Nr,Tr,Tr,Js,Js,_).
add_news([H|T],Ag0,Ag,Hash,Nr0,Nr,Tr0,Tr,Js0,Js,StNr):-
    add_new(H,Ag0,Ag1,Hash,Nr0,Nr1,Tr0,Tr1,Js0,Js1,StNr),
    add_news(T,Ag1,Ag,Hash,Nr1,Nr,Tr1,Tr,Js1,Js,StNr).

add_final_reachable(yes,S,[S|Fin],Fin).
add_final_reachable(no,_,Fin,Fin).

add_new(jump(H),Ag0,Ag,Hash,Nr0,Nr,Tr,Tr,[jump(StNr,Var)|Js],Js,StNr):-
    fsa_hash:fsa_hash_access(H,Var,Hash),
    (	var(Var)
    ->  Ag0 = [H/Nr0|Ag],
	Var=Nr0,
	Nr is Nr0+1
    ;   Ag0=Ag,
	Nr0=Nr
    ).

add_new(trans(Sym,H),Ag0,Ag,Hash,Nr0,Nr,[trans(StNr,Sym,Var)|Tr],Tr,
	Js,Js,StNr):-
    fsa_hash:fsa_hash_access(H,Var,Hash),
    (	var(Var)
    ->  Ag0 = [H/Nr0|Ag],
	Var=Nr0,
	Nr is Nr0+1
    ;   Ag0=Ag,
	Nr0=Nr
    ).

%% add_starts(Starts,Agenda,Hash,0,Size0) 
add_starts([],[],[],_,S,S).
add_starts([H|T],[S0|Starts],[H/S0|Ag],Hash,S0,S):-
    fsa_hash:fsa_hash_access(H,S0,Hash),
    S1 is S0+1,
    add_starts(T,Starts,Ag,Hash,S1,S).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% All components co-accessible %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%currently not used
%%all_components_co_reachable(Op,Fa0,Fa1,Fa):-
%%    prepare_co_reachable(Op,Fa0,Fa1,Data,Finals,Sigma),
%%    all_components_co_reachable0(Op,Data,Finals,Sigma,Fa).

all_components_co_reachable(Op,Fa0,Fa) :- 
    prepare_co_reachable(Op,Fa0,Data,Finals,Sigma),
    all_components_co_reachable0(Op,Data,Finals,Sigma,Fa).

all_components_co_reachable0(Op,Data,Finals0,Sigma,Fa) :-
    fsa_hash:fsa_hash_new(Hash),
    add_finals(Finals0,Finals,Agenda,Hash,0,Size0),
    co_reachable(Agenda,Hash,Op,Data,Starts,Transs,Jumps,Size0,Size),
    (	Starts==[]
    ->	fsa_data:fsa_construct(Sigma,1,[0],[],[],[],Fa)
    ;   fsa_data:fsa_construct(Sigma,Size,Starts,Finals,Transs,Jumps,Fa)
    ).

co_reachable([],_Hash,_Op,_Data,[],[],[],S,S).
co_reachable([St/StNr|T],Hash,Op,Data,Start0,Trn0,Jmp0,Nr0,Nr):-
    is_start_co_reachable(Op,Data,St,Bool),
    add_start_co_reachable(Bool,StNr,Start0,Start1),
    trans_co_reachables(Op,St,Data,New1),
    co_add_news(New1,Ag,T,Hash,Nr0,Nr1,Trn0,Trn1,Jmp0,Jmp1,StNr),
    co_reachable(Ag,Hash,Op,Data,Start1,Trn1,Jmp1,Nr1,Nr).

co_add_news([],Ag,Ag,_Hash,Nr,Nr,Tr,Tr,Js,Js,_).
co_add_news([H|T],Ag0,Ag,Hash,Nr0,Nr,Tr0,Tr,Js0,Js,StNr):-
    co_add_new(H,Ag0,Ag1,Hash,Nr0,Nr1,Tr0,Tr1,Js0,Js1,StNr),
    co_add_news(T,Ag1,Ag,Hash,Nr1,Nr,Tr1,Tr,Js1,Js,StNr).

add_start_co_reachable(yes,S,[S|Start],Start).
add_start_co_reachable(no,_,Start,Start).

co_add_new(jump(H,_P),Ag0,Ag,Hash,Nr0,Nr,Tr,Tr,[jump(Var,StNr)|Js],Js,StNr):-
    fsa_hash:fsa_hash_access(H,Var,Hash),
    (	var(Var)
    ->  Ag0 = [H/Nr0|Ag],
	Var=Nr0,
	Nr is Nr0+1
    ;   Ag0=Ag,
	Nr0=Nr
    ).

co_add_new(trans(H,Sym,_P),Ag0,Ag,Hash,Nr0,Nr,[trans(Var,Sym,StNr)|Tr],Tr,
	Js,Js,StNr):-
    fsa_hash:fsa_hash_access(H,Var,Hash),
    (	var(Var)
    ->  Ag0 = [H/Nr0|Ag],
	Var=Nr0,
	Nr is Nr0+1
    ;   Ag0=Ag,
	Nr0=Nr
    ).

%% add_finals(Finals,Agenda,Hash,0,Size0) 
add_finals([],[],[],_,S,S).
add_finals([H|T],[S0|Finals],[H/S0|Ag],Hash,S0,S):-
    fsa_hash:fsa_hash_access(H,S0,Hash),
    S1 is S0+1,
    add_finals(T,Finals,Ag,Hash,S1,S).

%%%%%%%%%%%%%%%
%%%% CACHE %%%%
%%%%%%%%%%%%%%%

#!yap# :- dynamic fsa_regex_cache:cache/3.

%% in_cache(off,_,_,_).
in_cache(on,_,Expr,Val):-
    ground(Expr),
    term_hash(Expr,Index),   %'
#!yap#     fsa_regex_cache:cache(Index,Expr,Val),
#yap#     recorded(fsa_regex_cache,cache(Index,Expr,Val),_),
    !.
in_cache(selective,Func,Expr,Val):-
    atom(Func),   %% bb_valid_key(Func),
    user:bb_get(fsa_regex_cache:Func,on),
    ground(Expr),
    term_hash(Expr,Index),   %'
#!yap#     fsa_regex_cache:cache(Index,Expr,Val),
#yap#     recorded(fsa_regex_cache,cache(Index,Expr,Val),_),
    !.

%% from Bips/intrins2.pl
%% used here to ignore functors that can never be defined to
%% be cached (e.g. floats)
%%bb_valid_key(X) :- atom(X), !.
%%bb_valid_key(X) :- integer(X), X>>25 =:= X>>26.

add_cache(off,_,_,_).
add_cache(on,_,Expr,Val):-
    (	ground(Expr)
    ->	term_hash(Expr,Index),   %'
#!yap#        assertz(fsa_regex_cache:cache(Index,Expr,Val))
#yap#        catch(recordz(fsa_regex_cache,cache(Index,Expr,Val),_),P,format(user_error,"Caching disabled for large term (~w)~n",[P]))
    ;   true    % ignore non-ground here.
    ).
add_cache(selective,Func,Expr,Val):-
    (	atom(Func)
    ->	add_cache_selective(Func,Expr,Val)
    ;   true
    ).

add_cache_selective(Func,Expr,Val) :-
    (	user:bb_get(fsa_regex_cache:Func,on)
    ->	add_cache_selective_if_ground(Expr,Val)
    ;	true			% ignore non-cached functors
    ).


add_cache_selective_if_ground(Expr,Val) :-
    (	ground(Expr)
    ->	term_hash(Expr,Index),	%'
#!yap#        assertz(fsa_regex_cache:cache(Index,Expr,Val))
#yap#        catch(recordz(fsa_regex_cache,cache(Index,Expr,Val),_),P,format(user_error,"Caching disabled for large term (~w)~n",[P]))
    ;	true			% ignore non-ground here.
    ).


fsa_regex_clear_cache :-
    fsa_util:debug_call(1,fsa_regex:verbose_clear_cache),
#!yap#     (   fsa_regex_cache:predicate_property(Term,dynamic),
#!yap#         retractall(fsa_regex_cache:Term),
#!yap#         fail
#!yap#     ;   true
#!yap#     ).
#yap#     (   recorded(fsa_regex_cache,_,Ref),
#yap#         erase(Ref),
#yap# 	      fail
#yap#     ;   true
#yap#     ).

verbose_clear_cache :-
#!swi#     statistics(program,[Used0,_]),
#swi#     statistics(heapused,Used0),
#!yap#     (   fsa_regex_cache:predicate_property(Term,dynamic),
#!yap#         retractall(fsa_regex_cache:Term),
#!yap# 	       fail
#!yap#     ;   true
#!yap#     ),
#yap#     (   recorded(fsa_regex_cache,_,Ref),
#yap#         erase(Ref),
#yap# 	      fail
#yap#     ;   true
#yap#     ),
#!swi#     statistics(program,[Used,_]),
#swi#     statistics(heapused,Used),
    Freed is Used0-Used,
    format(user_error,"Cache cleared (~w bytes freed)~n",[Freed]).

%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%% OPERATORS %%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%

default_ops( [op(0,fx,(-)),
	      op(0,yfx,(*)),
	      op(0,yfx,(+)),
	      op(0,yfx,(#)),
	      op(0,xfy,(^)),
#swi#              op(0,fx,(?)),
 	      op(395,xfy,(:)),
	      op(396,xfy,(::)),
	      op(397,xfx,(..)),
	      op(400,xf,(*)),
	      op(400,xf,(+)),
	      op(400,xf,(^)),
	      op(401,yfx,(-)),
	      op(401,yfx,(&)),
	      op(403,yfx,o),
	      op(403,yfx,x),
	      op(403,yfx,xx),
	      op(405,xf,(!)),
	      op(405,xf,(#)),
	      op(398,fy,(~)),
	      op(398,fy,(`)),
	      op(398,fy,($))
	      ]).

%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% DEBUGGING %%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%

debug_rx_start(Dbg,Term0,Func,Ar,RecLev,Time0) :-
    (	Dbg > 0
    ->	%% tell the world what's going on
	(   user:bb_get(fsa_rx_spy:Func,on)
	->  (   Dbg >= 1
	    ->
#swi#               statistics(cputime,Time00),
#swi#               Time0 is integer(1000*Time00),
#!swi#               statistics(runtime,[Time0,_]),
		format(user_error,"***fsa_regex: ~w treating ",[RecLev]),
#!yap# 		format(user_error,"~@~n",
#!yap# 		       [write_term(Term0,[max_depth(6)])])
#yap#           format(user_error,"~w~n",[Term0])
	    ;	true
	    )
	;   (   Dbg >= 3
	    ->
#swi#           statistics(cputime,Time00),
#swi#           Time0 is integer(1000*Time00),
#!swi#           statistics(runtime,[Time0,_]),
		format(user_error,"fsa_regex: ~w treating ~w/~w~n",
		       [RecLev,Func,Ar])
	    ;	true
	    )
	),
	(   Dbg >= 4
	->  
#!yap#       format(user_error,"~@~n",[write_term(Term0,[max_depth(6)])])
#yap#       format(user_error,"~w~n",[Term0])
	;   true
	)
    ;	true
    ).


debug_rx_end(Dbg,Func,Ar,RecLev,Time0,Fa) :-
    (	Dbg > 0
    ->	%% once more inform the world what we did
	(   user:bb_get(fsa_rx_spy:Func,on)
	->  (   Dbg >= 1
	    ->
#swi#           statistics(cputime,Time10),
#swi#           Time1 is integer(1000*Time10),
#!swi#           statistics(runtime,[Time1,_]),
		Time is Time1-Time0,
		fsa_data:fsa_states_number(Fa,States),
		format(user_error,
		    "***fsa_regex: ~w treated ~w/~w (~w states; ~w msec) ~n",
		    [RecLev,Func,Ar,States,Time])
	;   true
	    )
	;   (   Dbg >= 3
	    ->
#swi#           statistics(cputime,Time10),
#swi#           Time1 is integer(1000*Time10),
#!swi#           statistics(runtime,[Time1,_]),
		Time is Time1-Time0,
		fsa_data:fsa_states_number(Fa,States),
		format(user_error,
		    "fsa_regex: ~w treated ~w/~w (~w states; ~w msec) ~n",
		    [RecLev,Func,Ar,States,Time])
	    ;	true
	    )
	)
    ;	true
    ).

ltime(Debug,Call):-
    (	Debug >= 2
    ->	fsa_util:parse_call(Call,Module,Fun,Ar),
#swi#     statistics(cputime,T00),
#swi#     T0 is integer(1000*T00),
#!swi#     statistics(runtime,[T0,_]),
	call(Call),
#swi#     statistics(cputime,T10),
#swi#     T1 is integer(1000*T10),
#!swi#     statistics(runtime,[T1,_]),
	T is T1-T0,
	format_time(T,Module,Fun,Ar)
    ;	call(Call)
    ).

format_time(T,Module,Fun,Ar):-
    (	T>40
    ->	format(user_error,"~w msec for ~w~n",[T,Module:Fun/Ar])
    ;	true
    ).

ftime(Debug,Call):-
    (	Debug >= 2
    ->	fsa_util:parse_call(Call,Module,Fun,Ar),
#swi#     statistics(cputime,T00),
#swi#     T0 is integer(1000*T00),
#!swi#     statistics(runtime,[T0,_]),
	call_c(Call,C),
#swi#     statistics(cputime,T10),
#swi#     T1 is integer(1000*T10),
#!swi#     statistics(runtime,[T1,_]),
	T is T1-T0,
	format_time(T,Module,Fun,Ar),
	call(C)
    ;	call(Call)
    ).

call_c(Call,C) :-
    (	call(Call)
    ->	C=true
    ;	C=fail
    ).

prepare_reachable(efree,Fsa,d(Finals,TT,JT),Starts,Sigma):-
    fsa_data:fsa_components(Sigma,Size,Starts,Finals,_,Jumps,Fsa),
    fsa_determinizer:transitive_closure(Jumps,JT,Size),
    fsa_data:index_transitions(Fsa,TT).

is_final_reachable(efree,d(Finals,_,JT),State,Bool):-
    fsa_array:fsa_array_access(State,Sources0,[],JT),
    sort([State|Sources0],Sources),
    fsa_util:ord_intersect(Sources,Finals,Bool).

trans_reachables(efree,State,d(_,TT,JT),New):-
    fsa_array:fsa_array_access(State,Sources,[],JT),
    all_transitions0([State|Sources],State,TT,New,[]).

all_transitions0([],_,_,Trn,Trn).
all_transitions0([S|Ss],Source,TT,Trn0,Trn):-
    fsa_array:fsa_array_access(S,Trans,[],TT),
    all_transitions00(Trans,Source,Trn0,Trn1),
    all_transitions0(Ss,Source,TT,Trn1,Trn).

all_transitions00([],_,T,T).
all_transitions00([Sym-Targets|Tail],Source,Tr0,Tr):-
    all_transitions000(Targets,Source,Sym,Tr0,Tr1),
    all_transitions00(Tail,Source,Tr1,Tr).

all_transitions000([],_,_,Tr,Tr).
all_transitions000([H|T],St,Sym,[trans(St,Sym,H)|Trn0],Trn):-
    all_transitions000(T,St,Sym,Trn0,Trn).

prepare_co_reachable(efree,Fa,d(Starts,TransTable),Finals,Sigma):-
    fsa_data:fsa_final_states(Fa,Finals),
    fsa_data:fsa_symbols_decl(Fa,Sigma),
    fsa_data:fsa_start_states(Fa,Starts0),
    fsa_data:fsa_states_number(Fa,Size),
    fsa_data:fsa_jumps(Fa,Jumps),
    fsa_determinizer:transitive_closure(Jumps,Table,Size),
    fsa_hash:fsa_hash_new(Memo),
    fsa_determinizer:lookup_each_and_union(Starts0,Table,Starts1,Memo),
    fsa_data:starts_to_thing(Starts1,Starts),
    fsa_m_array:fsa_m_array_new(TransTable,Size),
    extend_trans_to_uthing(Fa,Table,TransTable,Memo).

is_start_co_reachable(efree,d(Starts,_),P,Bool):-
    fsa_data:indexed_start(P,Starts,Bool).

trans_co_reachables(efree,P,d(_,Trans),List):-
    fsa_m_array:fsa_m_array_get(P,List,[],Trans).

extend_trans_to_uthing(Fa,JumpTable,Table,M) :-
    fsa_determinizer:transitions_to_graph(Fa,Graph),
    graph_to_trans_with_jumps(Graph,JumpTable,M,Table).

graph_to_trans_with_jumps([],_,_,_).
graph_to_trans_with_jumps([Source-G0|Gs],Table,M,T):-
    graph_to_trans_with_jumps_(G0,Source,Table,M,T),
    graph_to_trans_with_jumps(Gs,Table,M,T).

graph_to_trans_with_jumps_([],_,_,_,_).
graph_to_trans_with_jumps_([Sym-List|Tail],Source,Table,Memo,T) :-
    fsa_determinizer:lookup_each_and_union(List,Table,Union,Memo),
    list_to_transitions(Union,Source,Sym,T),
    graph_to_trans_with_jumps_(Tail,Source,Table,Memo,T).

list_to_transitions([],_,_,_).
list_to_transitions([H|T],Source,Sym,Tr) :-
    fsa_m_array:fsa_m_array_put(H,List,[],[trans(Source,Sym,H)|List],Tr),
    list_to_transitions(T,Source,Sym,Tr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%  generation of random automata %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
fsa_random(Fa) :-
    fsa_random(15,15,0.04,0,1,Fa).

fsa_random(States,Syms,Density,Jdens,FDens,FaOut):-
    set_random,
    fsa_random0(States,Syms,Density,Jdens,Transitions,Jumps),
    States1 is States-1,
    findall(State,fsa_util:between(0,States1,State),AllStates),
    create_finals(AllStates,Finals,FDens),
    fsa_data:fsa_construct_rename_states(
          r(fsa_frozen),[0],Finals,Transitions,Jumps,FaOut).

fsa_random0(States,Syms,Density,Jdens,Trans,Jumps) :-
    States0 is States-1,
    findall(St,fsa_util:between(0,States0,St),StatesList),
    connected_states(StatesList,Trans0,Syms),
    findall(Trans,leslie_trans(States,Syms,Density,Trans),Trans,Trans0),
    findall(Jump,leslie_jump(States,Jdens,Jump),Jumps).

leslie_jump(States,Density,jump(P,Q)):-
    States0 is States-1,
    fsa_util:between(0,States0,P),
    fsa_util:between(0,States0,Q),
#sic#     random(Random),
#yap#     Random is random,
#swi#     Random is random(100000)/100000,
    Random < Density.

leslie_trans(States,Syms,Density,trans(P,Sym,Q)) :-
    States0 is States-1,
    Syms0 is Syms-1,
    fsa_util:between(0,States0,P),
    fsa_util:between(0,States0,Q),
    fsa_util:between(0,Syms0,Sym),
#sic#     random(Random),
#yap#     Random is random,
#swi#     Random is random(100000)/100000,
    Random <Density-(1/(States*Syms)).
%% adjust for states already introduced by connected.

connected_states([State|States],Trans,Syms) :-
    connected_states(States,[State],Syms,Trans,[]).

%% connected_states(NotConnected,Connected,Tr0,Tr)
connected_states([],_,_,Tr,Tr).

connected_states([NotConn0H|NotConn0T],Conn,Syms,[trans(P,Sym,Q)|Tr0],Tr) :-
#sic# 	    random(0,Syms,Sym),
#yap#       Sym is integer(Syms*random),
#swi#       Sym is integer(random(Syms)),
    random_element(P,Conn),
    random_select(Q,[NotConn0H|NotConn0T],NotConn),
    connected_states(NotConn,[Q|Conn],Syms,Tr0,Tr).

random_element(P,Conn) :-
    length(Conn,Length),
#sic# 	    random(0,Length,Nth0),
#yap#       Nth0 is integer(Length*random),
#swi#       Nth0 is integer(random(Length)),
    nth0(Nth0,Conn,P).

random_select(P,L0,L) :-
    length(L0,Length),
#sic# 	    random(0,Length,Nth0),
#yap#       Nth0 is integer(Length*random),
#swi#       Nth0 is integer(random(Length)),
    nth0(Nth0,L0,P,L).

%% deterministic, complete, efree
det_random(Fa):-
    det_random(15,15,0.04,0.2,Fa).

det_random(NoStates,NoSymbols,Tdens,Fdens,Fa):-
    set_random,
    NoStates0 is NoStates-1,
    findall(State,fsa_util:between(0,NoStates0,State),States),
    create_transitions(States,NoSymbols,Trans,NoStates,Tdens),
    create_finals(States,FinalStates,Fdens),
    fsa_data:fsa_components(r(fsa_frozen),NoStates,[0],FinalStates,Trans,[],Fa).

create_transitions([],_,[],_,_).
create_transitions([Source|Srcs],NoSymbols,Trans,NoStates,Dens):-
    create_transition(0,NoSymbols,Source,Trans,Trans0,NoStates,Dens),
    create_transitions(Srcs,NoSymbols,Trans0,NoStates,Dens).

create_transition(Nr0,Nr,Source,Trans0,Trans,NrStates,Dens):-
    (	Nr0 =:= Nr
    ->	Trans=Trans0
    ;   
#sic#         random(R),
#yap#         R is random,
#swi#         R is random(100000)/100000,
	(   R < Dens
	->  Trans0=[trans(Source,Nr0,Target)|Trans1],
#sic# 	    random(0,NrStates,Target)
#yap#       Target is integer(NrStates*random)
#swi#       Target is integer(random(NrStates))
	;   Trans0=Trans1
	),	    
	Nr1 is Nr0+1,
	create_transition(Nr1,Nr,Source,Trans1,Trans,NrStates,Dens)
    ).

create_finals([],[],_).
create_finals([H|T],Fs,Fdens):-
#sic#     random(R),
#yap#     R is random,
#swi#     R is random(100000)/100000,
    (	R < Fdens
    ->  Fs=[H|Fs0]
    ;   Fs=Fs0
    ),
    create_finals(T,Fs0,Fdens).


set_random :-
    fsa_globals:fsa_global_get(set_random,Off),
    (   Off==off
    ->  true                   %% don't set it
    ;	user:bb_get(fsa_regex:set_random,_)   %% already set
    ->	true
    ;	
#sic#   process_id(Pid0),
#sic# 	Pid is Pid0 mod 30268, %% scale since args of rand/3 are limited
#sic# 	setrand(random(Pid,Pid,Pid,Pid)),
#yap#   statistics(walltime,I),
#yap#   srandom(I),
#swi#   %% in swi, always setrand'ed
	user:bb_put(fsa_regex:set_random,[])  %% don't set it next-time
    ).

#swi# nth0(N,List,Element,Rest) :-
#swi#     (	integer(N)
#swi#     ->	N >= 0,
#swi# 	nth0i(N,List,Element,Rest)
#swi#     ;   nth0v(List,Element,0,N,Rest)
#swi#     ).
#swi# 
#swi# nth0v([Element|Tail],Element,Index,Index,Tail).
#swi# nth0v([Head|Tail],Element,M,Index,[Head|Rest]) :-
#swi#     N is M + 1,
#swi#     nth0v(Tail,Element,N,Index,Rest).
#swi# 
#swi# nth0i(N,List,Element,Result):-
#swi#     (	N =:= 0
#swi#     ->	List = [Element|Result]
#swi#     ;   List = [Head|Tail],
#swi# 	Result=[Head|Rest],
#swi# 	M is N - 1,
#swi# 	nth0i(M,Tail,Element,Rest)
#swi#     ).
#swi# 

domain_sigma(r(M),M).
domain_sigma(r(M,_),M).
domain_sigma(t(M,_),M).
domain_sigma(t(M,_,_),M).


%%
prepare_reachable(fsa_regex_aux:Op,Fa,Data,Finals,Sigma) :-
    fsa_regex_aux:prepare_reachable(Op,Fa,Data,Finals,Sigma).
prepare_reachable(fsa_regex_aux:Op,Fa0,Fa1,Data,Finals,Sigma) :-
    fsa_regex_aux:prepare_reachable(Op,Fa0,Fa1,Data,Finals,Sigma).
is_final_reachable(fsa_regex_aux:Op,Data,State,Bool) :-
    fsa_regex_aux:is_final_reachable(Op,Data,State,Bool).
trans_reachables(fsa_regex_aux:Op,P,Data,List):-
    fsa_regex_aux:trans_reachables(Op,P,Data,List).
prepare_co_reachable(fsa_regex_aux:Op,Fa,Data,Finals,Sigma) :-
    fsa_regex_aux:prepare_co_reachable(Op,Fa,Data,Finals,Sigma).
is_start_co_reachable(fsa_regex_aux:Op,Data,P,Bool) :-
    fsa_regex_aux:is_start_co_reachable(Op,Data,P,Bool).
trans_co_reachables(fsa_regex_aux:Op,P,Data,List) :-
    fsa_regex_aux:trans_co_reachables(Op,P,Data,List).




warn_if_variable_symbol(Symbol,Pred) :-
    (	var(Symbol)
    ->	format(user_error,
          "WARNING: conversion of predicate ~w yields variable!~n",
          [Pred])
    ;	true
    ).

